
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * FitsManager.cpp
 *
 *  Created on: 09/09/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "FitsManager.hpp"

int FitsManager::buffer_size = 1;
omp_lock_t FitsManager::mutex;

/**
 * @brief Creates a manager to deal with all input and output data.
 *
 * @param _n_zernikes The number of Zernikes in use.
 * @param _img_size The image width.
 * @param _fft_size The FFT width.
 * @param obj_filename The FITS file name for the Object.
 * @param _path_in The path for the multiple 2D Image FITS files Or the full path for the 3D Image FITS file.
 * @param _path_out The output path where the output will be stored.
 */
FitsManager::FitsManager(int _n_zernikes, int _img_size, int _fft_size, char *obj_filename, char *_path_in, char *_path_out){
	path_out = _path_out;
	if(isFolder(_path_in)){
		path_in = _path_in;
		image_filename = NULL;

		startup(_n_zernikes, _img_size, _fft_size, obj_filename);

		multi_files = true;
		listFolder(path_in, files, "fits");
	}
	else{
		path_in = NULL;
		image_filename = new string(_path_in);

		startup(_n_zernikes, _img_size, _fft_size, obj_filename);

		multi_files = false;
		saved = false;
		fits_size = fitsSize((char*)image_filename->c_str());
		fptr_in.fptr = fptr_out_psf.fptr = fptr_out_psfinv.fptr = fptr_out_phase.fptr = fptr_out_conobj.fptr =
				fptr_out_imgdiff.fptr = NULL;
		fptr_in.first_elem = fptr_out_psf.first_elem = fptr_out_psfinv.first_elem = fptr_out_phase.first_elem =
				fptr_out_conobj.first_elem = fptr_out_imgdiff.first_elem = 1;
		psf_buffer.pos = psfinv_buffer.pos = phase_buffer.pos = conobj_buffer.pos = imgdiff_buffer.pos = coefs_buffer.pos = 0;
		// These buffers must be double because their values will be saved to the output FITS files.
		psfinv_buffer.buffer = new double[fits_size * img_area];
		psf_buffer.buffer = new double[fits_size * img_area];
		phase_buffer.buffer = new double[fits_size * img_area];
		conobj_buffer.buffer = new double[fits_size * img_area];
		imgdiff_buffer.buffer = new double[fits_size * img_area];
		// Coefficients must be WORD.
		coefs_buffer.buffer = new WORD[fits_size * n_zernikes];
	}
}

/**
 * @brief If a 3D Image FITS file was used as input, it will be saved Only at object destruction time.
 */
FitsManager::~FitsManager(){
	int i;
	for (i=0; i < buffer_size; i++){
		delete buffer[i]->image;
		delete buffer[i]->coefs;
		delete buffer[i]->psf_generated;
		delete buffer[i]->psf_inverted;
		delete buffer[i]->phase_generated;
		delete buffer[i]->conobj;
		delete buffer[i]->img_diff;
		delete buffer[i];
	}
	delete[] buffer;
	delete object;

	if(multi_files){
		// Delete the filenames allocated.
		for(i=files.size()-1; i >= 0; i--){
			delete files[i];
		}
	}
	else{
		if(saved){
			// Save the results to files.
			char path[1024], filename[100];
			getFileName((char*)image_filename->c_str(), filename);
			sprintf(path, "%s/%s_psf.fits", path_out, filename);
			writeFitsFile(path, &fptr_out_psf, true, psf_buffer.buffer, img_size, img_size, fits_size);
			sprintf(path, "%s/%s_psfinv.fits", path_out, filename);
			writeFitsFile(path, &fptr_out_psfinv, true, psfinv_buffer.buffer, img_size, img_size, fits_size);
			sprintf(path, "%s/%s_phase.fits", path_out, filename);
			writeFitsFile(path, &fptr_out_phase, true, phase_buffer.buffer, img_size, img_size, fits_size);
			sprintf(path, "%s/%s_conobj.fits", path_out, filename);
			writeFitsFile(path, &fptr_out_conobj, true, conobj_buffer.buffer, img_size, img_size, fits_size);
			sprintf(path, "%s/%s_imgdiff.fits", path_out, filename);
			writeFitsFile(path, &fptr_out_imgdiff, true, imgdiff_buffer.buffer, img_size, img_size, fits_size);
			sprintf(path, "%s/%s_coefs.txt", path_out, filename);
			save_coefs(path, coefs_buffer.buffer, fits_size, n_zernikes);
		}

		delete image_filename;
		delete psfinv_buffer.buffer;
		delete psf_buffer.buffer;
		delete phase_buffer.buffer;
		delete conobj_buffer.buffer;
		delete imgdiff_buffer.buffer;
		delete coefs_buffer.buffer;
	}

	omp_destroy_lock(&mutex);
}

/**
 * @brief Initialize the FitsManager.
 *
 * @param _bufer_size The buffer size to be used. The default is 1.
 */
void FitsManager::startup(int _bufer_size){
	buffer_size = _bufer_size;
}

/**
 * @brief Initializes the manager.
 *
 * @param _n_zernikes The number of Zernikes in use.
 * @param _img_size The Image width.
 * @param _fft_size The FFT width.
 * @param obj_filepath The FITS file name for the Object.
 */
void FitsManager::startup(int _n_zernikes, int _img_size, int _fft_size, char *obj_filepath){
	omp_init_lock(&mutex);

	n_zernikes = _n_zernikes;
	img_size = _img_size;
	fft_size = _fft_size;
	img_area = img_size * img_size;
	phase_area = fft_size * fft_size;
	seq_read = seq_write = 0;

	object = new double[img_area];
	if(obj_filepath != NULL){
		readFitsFile(obj_filepath, object, img_area);
	}

	buffer = new fits_result*[buffer_size];
	for (int i=0; i < buffer_size; i++){
		buffer[i] = new fits_result;
		// Set the same object for all the entries in the buffer.
		buffer[i]->object = object;
		// Expect the FITS input image to be in double format.
		buffer[i]->image = new double[img_area];
		// These are the values directly calculated by the PSF estimation, thus they must be WORD.
		buffer[i]->coefs = new WORD[n_zernikes];
		buffer[i]->psf_generated = new WORD[img_area];
		buffer[i]->psf_inverted = new WORD[img_area];
		buffer[i]->phase_generated = new WORD[phase_area];
		buffer[i]->conobj = new WORD[img_area];
		buffer[i]->img_diff = new WORD[img_area];
		buffer[i]->img_filename = image_filename;
		available_buffer.push_back(buffer[i]);
	}
}

/**
 * @brief Return an entry in the buffer containing the next Image to be processed.
 *
 * @return The next Image and all the buffers necessary to store the results.
 */
fits_result* FitsManager::nextImage(){
	omp_set_lock(&mutex);

	if (multi_files){
		if(files.size() == 0){
			return NULL;
		}
	}
	else if(seq_read == fits_size){
		return NULL;
	}

	while(available_buffer.size() == 0){
		omp_unset_lock(&mutex);
		usleep(1000);
		omp_set_lock(&mutex);
	}
	fits_result *result = available_buffer[0];
	available_buffer.erase(available_buffer.begin());
	omp_unset_lock(&mutex);

	char path[1024];
	if (multi_files){
		result->sequence = 0;
		result->img_filename = files[0];
		files.erase(files.begin());
		sprintf(path, "%s/%s", path_in, result->img_filename->c_str());
		readFitsFile(path, result->image, img_area);
	}
	else{
		result->sequence = seq_read++;
		readFitsFile((char*)result->img_filename->c_str(), &fptr_in, false, result->image, img_area);
	}
	memset(result->coefs, 0, n_zernikes * sizeof(WORD));
	memset(result->psf_generated, 0, img_area * sizeof(WORD));
	memset(result->psf_inverted, 0, img_area * sizeof(WORD));
	memset(result->phase_generated, 0, phase_area * sizeof(WORD));
	memset(result->conobj, 0, img_area * sizeof(WORD));
	memset(result->img_diff, 0, img_area * sizeof(WORD));
	return result;
}

/**
 * @brief Save the results of the PSF estimation.
 *
 * If it's dealing with a 2D Image FITS file, the results are saved directly to the disk.
 * If it's a 3D Image FITS file, the results are stored in a secondary buffer, which
 * will be saved Only at destruction moment.
 *
 * @param result The buffer entry containing the processing results.
 */
void FitsManager::saveImage(fits_result *result){
	char path[1024];
	char *filename;
	if (multi_files){
		// Save directly to the disk.
		filename = (char*)result->img_filename->c_str();
		sprintf(path, "%s/%s_psf.fits", path_out, filename);
		writeFitsFile(path, result->psf_generated, img_size, img_size);
		sprintf(path, "%s/%s_psfinv.fits", path_out, filename);
		writeFitsFile(path, result->psf_inverted, img_size, img_size);
		sprintf(path, "%s/%s_phase.fits", path_out, filename);
		writeFitsFile(path, result->phase_generated, img_size, img_size);
		sprintf(path, "%s/%s_conobj.fits", path_out, filename);
		writeFitsFile(path, result->conobj, img_size, img_size);
		sprintf(path, "%s/%s_imgdiff.fits", path_out, filename);
		writeFitsFile(path, result->img_diff, img_size, img_size);
		sprintf(path, "%s/%s_coefs.txt", path_out, filename);
		save_coefs(path, result->coefs, 1, n_zernikes);

		// Put the entry back to the buffer.
		omp_set_lock(&mutex);
		available_buffer.push_back(result);
	}
	else{
		omp_set_lock(&mutex);

		fits_result *to_write;
		// Put the result in a separate list that is used to organize the output sequence.
		awaiting_buffer.push_back(result);
		// Guarantee that the output sequence is exactly the same as the 3D input file.
		for(int i=awaiting_buffer.size()-1; i >= 0; i--){
			// If this entry is the next to be written, save to the results in a secondary buffer.
			if(awaiting_buffer[i]->sequence == seq_write){
				seq_write++;
				to_write = awaiting_buffer[i];
				// Remove the entry from the separate list.
				awaiting_buffer.erase(awaiting_buffer.begin() + i);

				// Save the results to the secondary buffer.
				holdData(result);

				// Put the entry back to the main buffer.
				available_buffer.push_back(to_write);

				// Reinitialize the sequencing, since the next due entry can be anywhere.
				i = awaiting_buffer.size();
			}
		}
		saved = true;
	}
	omp_unset_lock(&mutex);
}

/**
 * @brief Return the number of Images available for this FITS manager.
 *
 * @return The number of images available.
 */
int FitsManager::size(){
	if (multi_files){
		return files.size();
	}
	else{
		return fits_size;
	}
}

/**
 * @brief Return a reference to the Object data.
 *
 * @return A reference to the Object data
 */
double *FitsManager::getObject(){
	return object;
}

/**
 * @brief Store the results to a secondary buffer, converting the values from float to double if necessary.
 *
 * @param result The results to be stored.
 */
void FitsManager::holdData(fits_result *result){
	int i;
	// Depending on the WORD, this makes any necessary conversion from float to double
	// which is necessary to save to the output FITS file.
	for(i=0; i < img_area; i++, psf_buffer.pos++) psf_buffer.buffer[psf_buffer.pos] = result->psf_generated[i];
	for(i=0; i < img_area; i++, psfinv_buffer.pos++) psfinv_buffer.buffer[psfinv_buffer.pos] = result->psf_inverted[i];
	for(i=0; i < img_area; i++, phase_buffer.pos++) phase_buffer.buffer[phase_buffer.pos] = result->phase_generated[i];
	for(i=0; i < img_area; i++, conobj_buffer.pos++) conobj_buffer.buffer[conobj_buffer.pos] = result->conobj[i];
	for(i=0; i < img_area; i++, imgdiff_buffer.pos++) imgdiff_buffer.buffer[imgdiff_buffer.pos] = result->img_diff[i];
	for(i=0; i < n_zernikes; i++, coefs_buffer.pos++) coefs_buffer.buffer[coefs_buffer.pos] = result->coefs[i];
}

/**
 * @brief Read the FITS file.
 *
 * @param filename The FITS file name.
 * @param image A pointer where the data will be stored. It must be previously allocated.
 * @param size The size of the image.
 */
void FitsManager::readFitsFile(char* filename, double* image, int size) {
	fits_ptr ptr = {NULL, 1};
	return readFitsFile(filename, &ptr, true, image, size);
}

/**
 * @brief Read the FITS file.
 *
 * @param filename The FITS file name.
 * @param ptr The pointer to the FITS file. If not provided, one will be created.
 * @param closeFits If close or leave the file open.
 * @param image A pointer where the data will be stored. It must be previously allocated.
 * @param size The size of the image.
 */
void FitsManager::readFitsFile(char* filename, fits_ptr *ptr, bool closeFits, double* image, int size) {
	omp_set_lock(&mutex);
	int status = 0;
	// Open the FITS file.
	if (ptr->fptr == NULL && fits_open_file(&ptr->fptr, filename, READONLY, &status)){
		printerror(status);
		exit(1);
	}

	// Read the FITS file at the correct sequence, based on the FITS pointer provided.
	int anynull;
	double nullval = 0;
	fits_read_img(ptr->fptr, TDOUBLE, ptr->first_elem, size, &nullval, image, &anynull, &status);

	// Close the file, if necessary.
	if ((status || closeFits)){
		if (fits_close_file(ptr->fptr, &status)){
			printerror(status);
			exit(1);
		}
	}
	else{
		// If don't close the file, increase the sequence in the FITS pointer.
		ptr->first_elem += size;
	}
	omp_unset_lock(&mutex);
}

/**
 * @brief Get the number of images available in the FITS file.
 *
 * If the file is 2D, it always returns 1.
 *
 * @param filename The FITS file name.
 * @return The number of images available in the FITS file.
 */
int FitsManager::fitsSize(char* filename) {
	omp_set_lock(&mutex);
	fitsfile *fptr;
	int status = 0;
	if (fits_open_file(&fptr, filename, READONLY, &status)){
		printerror(status);
		exit(1);
	}

	int size;
	fits_read_key(fptr, TLONG, "NAXIS3", &size, NULL, &status);
	// A 2D file doesn't contain the key NAXIS3.
	if(status){
		size = 1;
		status = 0;
	}

	if (fits_close_file(fptr, &status)){
		printerror(status);
		exit(1);
	}
	omp_unset_lock(&mutex);

	return size;
}

/**
 * @brief Write the image the a FITS file.
 *
 * It converts the data to double, if necessary.
 *
 * @param filename The FITS file name to save the results.
 * @param array The image data.
 * @param sx The width of the image.
 * @param sy The height of the image.
 */
void FitsManager::writeFitsFile(char* filename, WORD* array, int sx, int sy) {
	// Ensure that the output file data type is always double, regardless the data type currently in use.
	int size = sx * sy;
	double _array[size];
	// Convert the data from float to double, if necessary.
	for(int i=0; i < size; i++) _array[i] = array[i];

	fits_ptr ptr = {NULL, 1};
	writeFitsFile(filename, &ptr, true, _array, sx, sy, 1);
}

/**
 * @brief Write the image the a FITS file.
 *
 * If the FITS pointer is NULL, a new pointer will be created and the
 * FITS file, if exists, will be deleted.
 *
 * @param filename The FITS file name to save the results.
 * @param ptr The FITS pointer.
 * @param closeFits If the file will be closed.
 * @param array The image data.
 * @param sx The width of the image.
 * @param sy The height of the image.
 * @param sz The number of images to be saved. If this is a 2D image, 1 must be provided.
 */
void FitsManager::writeFitsFile(char* filename, fits_ptr *ptr, bool closeFits, double* array, int sx, int sy, int sz) {
	omp_set_lock(&mutex);
	int status = 0; // initialize status before calling fitsio routines
	if (ptr->fptr == NULL){
		// Delete old file if it already exists
		remove(filename);

		// create new FITS files
		if (fits_create_file(&(ptr->fptr), filename, &status)){
			printerror(status);
			exit(1);
		}
	}

    int  bitpix   =  -64;
    long naxis = (sz > 1) ? 3 : 2; // Specify the dimension of the image.
	long naxes[naxis];
	naxes[0] = sx; naxes[1] = sy;
    if(sz > 1){
    	// Set the number of images in a 3D file.
    	naxes[2] = sz;
    }

    // Create an entry to the image in the FITS file.
    if (fits_create_img(ptr->fptr, bitpix, naxis, naxes, &status)){
		printerror(status);
	}

	// write the array to the FITS file
	long size = sx * sy * sz;
	if (fits_write_img(ptr->fptr, TDOUBLE, ptr->first_elem, size, array, &status)){
		printerror(status);
	}

	// Close the file, if necessary.
	if ((status || closeFits)){
		// close the file
		if (fits_close_file(ptr->fptr, &status)){
			printerror(status);
			exit(1);
		}
	}
	else{
		// If don't close the file, increase the sequence in the FITS pointer.
		ptr->first_elem += size;
	}
	omp_unset_lock(&mutex);
}

/**
 * @brief Save the coefficients to file in text format.
 *
 * @param file_name The output file name.
 * @param coefs A pointer to the coefficients.
 * @param n_rows The number of rows to be saved.
 * @param n_coefs The number of coefficients per each row.
 */
void FitsManager::save_coefs(char *file_name, WORD *coefs, int n_rows, int n_coefs){
	if(coefs == NULL || n_rows < 1 || n_coefs < 1 || file_name == NULL) return;

	// Allow a buffer with 30 chars + ", " per number, with a '\n' per line.
	char s_coefs[n_rows * ((n_coefs * (30 + 2)) + 1)];
	s_coefs[0] = '\0';
	for(int j, i=0; i < n_rows; i++){
		for(j=0; j < n_coefs; j++){
			sprintf(s_coefs, "%s%f, ", s_coefs, coefs[(i*n_coefs)+j]);
		}
		s_coefs[strlen(s_coefs)-1] = '\n';
	}

	FILE *fp = fopen(file_name, "w");
	if (!fp) {
		cerr << "Failed to save coefs [" << file_name << "]." << std::endl;
		return;
	}
	int sz1 = fwrite(s_coefs, 1, strlen(s_coefs), fp);
	if (sz1 < 1) {
		cerr << "Failed to save coefs [" << file_name << "]." << std::endl;
	}
	fclose(fp);
}

/**
 * @brief Print the FITS error to the stderr.
 *
 * @param status The error code.
 */
void FitsManager::printerror(int status) {
	if (status) {
		fits_report_error(stderr, status); /* print error report */
		exit(status); /* terminate program, return error status */
	}
}

/**
 * @brief Return a list of files that match the provided file extension.
 *
 * Every result returned must be deleted manually.
 *
 * @param folder The folder to be listed.
 * @param files A vector where to save the results.
 * @param extension The extension to be filtered.
 */
void FitsManager::listFolder(char *folder, vector<string*> &files, const char *extension) {
	DIR *dp;
	struct dirent *dirp;
	if ((dp = opendir(folder)) == NULL) {
		cout << "Error opening " << folder << endl;
		exit(1);
	}

	char ext[30];
	while ((dirp = readdir(dp)) != NULL) {
		getFileExtension(dirp->d_name, ext);
		if(strcmp(ext, extension) == 0){
			files.push_back(new string(dirp->d_name));
		}
	}
	closedir(dp);
}

/**
 * @brief Tell if the path provided is a folder.
 *
 * @param folder The path to verify.
 * @return true if the path is a folder, false otherwise.
 */
bool FitsManager::isFolder(char *folder) {
	DIR *dp;
	if ((dp = opendir(folder)) == NULL) {
		return false;
	}

	closedir(dp);
	return true;
}

/**
 * @brief Return only the file name from a full path.
 *
 * @param fullPath The full path of a file.
 * @param fileName A pointer where the file name part will be stored. This pointer must be previously allocated.
 */
void FitsManager::getFileName(char *fullPath, char *fileName){
	int i, j, pos;
	for(i=0; fullPath[i] != '\0'; i++){
		if(fullPath[i] == '/' || fullPath[i] == '\\'){
			pos = i;
		}
	}
	for(j=0, pos++; pos <= i; pos++, j++){
		fileName[j] = fullPath[pos];
	}
}

/**
 * @brief Return the file extension from a file name.
 *
 * @param fullPath The full path of a file.
 * @param fileExtension The file extension.
 */
void FitsManager::getFileExtension(char *fullPath, char *fileExtension){
	int i, j, pos;
	for(i=0; fullPath[i] != '\0'; i++){
		if(fullPath[i] == '.'){
			pos = i;
		}
	}
	for(pos++, j=0; pos <= i; pos++, j++){
		fileExtension[j] = fullPath[pos];
	}
}
