
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * FitsManager.hpp
 *
 *  Created on: 09/09/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#ifndef FITSMANAGER_HPP_
#define FITSMANAGER_HPP_

#include <dirent.h>
#include <iostream>
#include <string.h>
#include <vector>
#include <fitsio.h>
#include "config.hpp"
#include "util.hpp"

using namespace std;

/**
 * @brief This struct contains all the basic data necessary to estimate the PSF,
 * and contains the buffers to store the results.
 */
typedef struct{
	int sequence;
	string *img_filename;
	double *image;
	double *object;
	WORD *coefs;
	WORD *psf_generated;
	WORD *psf_inverted;
	WORD *phase_generated;
	WORD *conobj;
	WORD *img_diff;
} fits_result;

/**
 * @brief This struct is a pointer to navigate into a FITS file.
 */
struct fits_ptr{
	fitsfile *fptr;
	long first_elem;
};

/**
 * @brief This struct is a buffer to store 3D results in double format.
 */
struct fits_out_buffer{
	double *buffer;
	int pos;
};

/**
 * @brief This struct is a buffer to store 3D results in WORD format.
 */
struct word_out_buffer{
	WORD *buffer;
	int pos;
};

/**
 * @brief This class is responsible for managing the input and output FITS files.
 *
 * It accepts as input either a folder containing multiple 2D FITS files Or
 * a single 3D FITS file. The output correspond to the same format as provided
 * by the input, i.e., either multiple 2D outputs Or a single 3D output.
 * For every input file processed, it generates as output 4 files with same
 * file name + a suffix, where each suffix is describe below:
 * <ul>
 * <li>_psf.fits: The PSF calculated.</li>
 * <li>_phase.fits: The phase calculated.</li>
 * <li>_conobj.fits: The convolved object calculated.</li>
 * <li>_coefs.txt: The Zernike coefficients calculated.</li>
 * </ul>
 */
class FitsManager{
	static int buffer_size;
	/**
	 * All calls must be synchronized between threads, since no parallel I/O can be executed for FITS files.
	 */
	static omp_lock_t mutex;

	double *object;
	fits_result **buffer;
	fits_out_buffer psfinv_buffer, psf_buffer, phase_buffer, conobj_buffer, imgdiff_buffer;
	word_out_buffer coefs_buffer;
	int fits_size, seq_read, seq_write;
	int n_zernikes, img_size, fft_size, img_area, phase_area;
	char *path_in;
	char *path_out;
	string *image_filename;
	vector<fits_result*> awaiting_buffer;
	vector<fits_result*> available_buffer;
	vector<string*> files;
	fits_ptr fptr_in, fptr_out_psf, fptr_out_psfinv, fptr_out_phase, fptr_out_conobj, fptr_out_imgdiff;

	bool multi_files, saved;

	void startup(int _n_zernikes, int _img_area, int _size_fft, char *obj_filepath);

	void readFitsFile(char* filename, double* image, int size);

	void readFitsFile(char* filename, fits_ptr *fptr, bool closeFits, double* image, int size);

	int fitsSize(char* filename);

	void save_coefs(char *file_name, WORD *coefs, int n_rows, int n_coefs);

	void holdData(fits_result *result);

	void printerror(int status);

	void listFolder(char *folder, vector<string*> &files, const char *extension);

	void getFileName(char *fullPath, char *fileName);

	bool isFolder(char *folder);

	void getFileExtension(char *fullPath, char *fileExtension);

public:
	FitsManager(int _n_zernikes, int _img_area, int _size_fft, char *obj_filename, char *_path_in, char *_path_out);

	~FitsManager();

	static void startup(int _bufer_size);

	void writeFitsFile(char* filename, WORD* array, int sx, int sy);

	void writeFitsFile(char* filename, fits_ptr *ptr, bool closeFits, double* array, int sx, int sy, int sz);

	fits_result* nextImage();

	void saveImage(fits_result *result);

	double *getObject();

	int size();
};

#endif /* FITSMANAGER_HPP_ */
