
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * TimeTracker.cpp
 *
 *  Created on: 28/09/2011
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include <iostream>
#include <sstream>
#include "TimeTracker.hpp"

using namespace std;

/**
 * @brief Constructor.
 *
 * @param _name The name of this tracker.
 */
TimeTracker::TimeTracker(string _name){
	name.append(_name.c_str());
	resetAll();
}

/**
 * @brief Starts this tracker.
 */
void TimeTracker::start(){
	if(done){
		resetAll();
	}
	else{
		paused = false;
	}
	gettimeofday(&startTime, NULL);
	startClock = clock();
	processing = true;
}

/**
 * @brief Pause the tracking temporarily.
 */
void TimeTracker::pause(){
	if(paused){
		return;
	}
	gettimeofday(&endTime, NULL);
	endClock = clock();
	currElapsedTime += (endTime.tv_sec - startTime.tv_sec) +
			(endTime.tv_usec - startTime.tv_usec)/1000000.0l;
//	currWorkSpent += (endClock - startClock) * 1000.0l / CLOCKS_PER_SEC;
	currWorkSpent += ((endClock - startClock) * 1.0) / CLOCKS_PER_SEC;
	resetCounters();
	processing = false;
	paused = true;
}

/**
 * @brief Resume the tracking.
 */
void TimeTracker::resume(){
	start();
}

/**
 * @brief Finish this tracker.
 *
 * @return The total elapsed time.
 */
long double TimeTracker::end(){
	if(!done){
		pause();
		done = true;
		processing = false;
		paused = false;
	}
	return elapsedTime();
}

/**
 * @brief Return the elapsed time based on the system time.
 *
 * The elapsed time is calculated as the difference between the
 * tracker start and: either the last call of
 * {@link #pause} method or the call of the {@link #end} method.
 *
 * @return Current elapsed time.
 */
long double TimeTracker::elapsedTime(){
	return currElapsedTime;
}

/**
 * @brief Return the work spent based on the CPU clock.
 *
 * @return Current work spent.
 */
long double TimeTracker::workSpent(){
	return currWorkSpent;
}

/**
 * @brief Reset all counters and the tracker state.
 */
void TimeTracker::reset(){
	resetAll();
}

/**
 * @brief Reset all counters and the tracker state.
 */
void TimeTracker::resetAll(){
	resetCounters();
	currElapsedTime = currWorkSpent = 0;
	done = false;
	processing = false;
	paused = false;
}

/**
 * @brief Reset all counters.
 */
void TimeTracker::resetCounters(){
	startTime.tv_sec = startTime.tv_usec =
		endTime.tv_sec = endTime.tv_usec =
		startClock = endClock = 0;
}

/**
 * @brief Sum the current work with the parameter's work.
 *
 * @param tt The tracker instance to be added to the current work.
 */
void TimeTracker::sumWork(TimeTracker *tt){
	currWorkSpent += tt->currWorkSpent;
}

/**
 * @brief Sum the current elapsed time with the parameter's elapsed time.
 *
 * @param tt The tracker instance to be added to the current elapsed time.
 */
void TimeTracker::sumTime(TimeTracker *tt){
	currElapsedTime += tt->currElapsedTime;
}

/**
 * @brief Sum this tracker with the parameter.
 *
 * @param tt The tracker instance to be added to the current tracker.
 */
void TimeTracker::sum(TimeTracker *tt){
	sumTime(tt);
	sumWork(tt);
}

/**
 * @brief Print the tracker to the output stream.
 */
void TimeTracker::print(){
	cout << "Tracker: " << getName();
	cout << "\nElapsed Time: " << elapsedTime();
	cout << "\nWork Spent: " << workSpent() << '\n';
}

/**
 * @brief Print the tracker to the stream provided.
 *
 * @param out The stream to be used for the printing.
 */
void TimeTracker::print(ostringstream *out){
	*out << "Tracker: " << getName();
	*out << "\nElapsed Time: " << elapsedTime();
	*out << "\nWork Spent (clock): " << workSpent() << '\n';
}
