
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * cl_factory.cpp
 *
 *  Created on: 16/07/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */
#include "cl_factory.hpp"

cl_context *clFactory::context;
omp_lock_t clFactory::mutex;
omp_lock_t clFactory::mutex_queue;
int clFactory::n_contexts;
int clFactory::n_queues;
int clFactory::curr_queue;
int clFactory::curr_context;
vector<clQueue*> clFactory::queues;
map<cl_context, context_t*> clFactory::contexts;
map<cl_device_id, omp_lock_t*> clFactory::devices;

/**
 * @brief Initialize the GPU environment.
 *
 * It is responsible for reading and compiling the OpenCl source code,
 * creating the contexts and one command queue per context, and initializing
 * the GPU FFT environment.
 *
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 */
void clFactory::initialize_gpu(vector<int> *_devices){
	omp_init_lock(&mutex);
	omp_init_lock(&mutex_queue);

	// Read the .cl file.
	char *src = readCl();

	// Create the OpenCL contexts.
	createContext(_devices);

	// Create the command queues.
	char cBuffer[1024];
	int i, j;
	context_t *_context;
	cl_program program;
	kernel_set *kernels;
	omp_lock_t *_mutex;
	n_queues = 0;
	map<cl_context, context_t*>::iterator iter;
	for(iter=contexts.begin(); iter != contexts.end(); iter++){
		_context = iter->second;
		createCommandQueue(_context);
		if((int)queues.size() == n_queues){
			continue;
		}

		// Create the OpenCL program from .cl kernel source for every device available inside the context.
		if ((program=createProgram(_context->context, _context->devices[0], src)) == NULL) {
			// If cannot be compiled in any device, stop the program.
			for(i=0; i < n_queues; i++){
				clReleaseCommandQueue(queues[i]->getCommandQueue());
			}
			for(i=0; i < n_contexts; i++){
				clReleaseContext(context[i]);
			}
			for(iter=contexts.begin(); iter != contexts.end(); iter++){
				delete iter->second->devices;
				delete iter->second;
			}
			delete context;
			delete src;
			exit(1);
		}
		// Initialize the basic cl environment (like kernels).
		kernels = initialize_cl(_context->context, program);
		// Set the correct kernels and mutex for the newly created queues.
		for(j=n_queues; j < (int)queues.size(); j++){
			_mutex = devices[queues[j]->getDevice()];
			queues[j]->kernels = clone_kernels(kernels, _mutex);
		}
		n_queues = queues.size();
	}
	delete src;

	if (n_queues == 0) {
		for(i=0; i < n_contexts; i++){
			clReleaseContext(context[i]);
		}
		for(iter=contexts.begin(); iter != contexts.end(); iter++){
			delete iter->second->devices;
			delete iter->second;
		}
		delete context;
		printf("No device found that matches the provided parameters!\n");
		exit(1);
	}

	// Print the device information.
	// Note that only 1 command queue is created for every device during startup.
	for(i=0; i < n_queues; i++){
		getDeviceName(queues[i], cBuffer, 1024);
		printf("Queues created on Device [%s]\n", cBuffer);
	}

	// Initialize the FFT context.
	// Use the first Context and its command queue for the FFT.
	viennacl::ocl::setup_context(0, queues[0]->getContext(), queues[0]->getDevice(), queues[0]->getCommandQueue());
	// Retain one additional reference for the first Context and for its command queue,
	// to avoid finalization errors.
	clRetainContext(queues[0]->getContext());
	clRetainCommandQueue(queues[0]->getCommandQueue());
}


/**
 * @brief Create an OpenCL context on all available platforms using
 * either a GPU or CPU, depending on what is available for this process.
 *
 * Only 1 single context is created per platform.
 * For example: if there are 2 platforms available, one for each vendor (like NVidia and AMD),
 * with every platform containing 2 distinct devices, every context will contain the
 * 2 devices related to the same vendor.
 *
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 */
void clFactory::createContext(vector<int> *_devices) {
	cl_int errNum;
	cl_uint numPlatforms;
	cl_platform_id platformIds[CL_PLATFORMS_TO_TRY];

	// First, select an OpenCL platform to run on.
	// This code will try up to CL_PLATFORMS_TO_TRY available platforms.
	errNum = clGetPlatformIDs(CL_PLATFORMS_TO_TRY, platformIds, &numPlatforms);
	if (errNum != CL_SUCCESS || numPlatforms <= 0) {
		std::cerr << "Failed to find any OpenCL platforms." << std::endl;
		exit(1);
	}
	if(numPlatforms > CL_PLATFORMS_TO_TRY) numPlatforms = CL_PLATFORMS_TO_TRY;

	// Create OpenCL contexts on the platform for the primary device type (GPU or CPU).
	errNum = CL_DEVICE_NOT_FOUND;
	cl_context _cl_context[numPlatforms];
	int n_found_contexts = 0;
	int i;
	for(i=0; i < (int)numPlatforms; i++){
		cl_context_properties contextProperties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties) platformIds[i], 0 };
		cl_context _context = clCreateContextFromType(contextProperties, PRIMARY_DEVICE_TYPE, NULL, NULL, &errNum);
		if(errNum == CL_SUCCESS){
			_cl_context[n_found_contexts++] = _context;
		}
	}
	// If all primary devices have failed, try to create it on the secondary device type.
	if(n_found_contexts == 0){
		for(i=0; i < (int)numPlatforms && errNum != CL_SUCCESS; i++){
			cl_context_properties contextProperties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties) platformIds[i], 0 };
			cl_context _context = clCreateContextFromType(contextProperties, SECONDARY_DEVICE_TYPE, NULL, NULL, &errNum);
			if(errNum == CL_SUCCESS){
				_cl_context[n_found_contexts++] = _context;
			}
		}
	}
	if (n_found_contexts == 0) {
		std::cerr << "Failed to create an OpenCL GPU or CPU context." << std::endl;
		exit(1);
	}

	// Determine the devices that will compose each context.
	context_t *_context;
	int j, k, new_count, curr_device = 1;
	for(i=0; i < n_found_contexts; i++){
		_context = new context_t;
		getDevices(_cl_context[i], _context);
		// Only store the context if it contains at least 1 device available.
		if(_context->n_devices > 0){
			vector<cl_device_id> saved_devices;
			// If filter the devices.
			if(_devices != NULL){
				// For all devices within this context.
				for(j=0; j < _context->n_devices; j++, curr_device++){
					// For all devices to be filtered.
					for(k=0; k < (int)_devices->size(); k++){
						if(curr_device == _devices->at(k)){
							saved_devices.push_back(_context->devices[j]);
						}
					}
				}
				new_count = saved_devices.size();
				// If any device within this context shall be used.
				if(new_count > 0){
					// If any device was discarded, rebuild the pointer.
					if(new_count != _context->n_devices){
						_context->n_devices = new_count;
						delete _context->devices;
						_context->devices = new cl_device_id[_context->n_devices];
						for(j=0; j < _context->n_devices; j++){
							_context->devices[j] = saved_devices[j];
						}
					}
				}
				else{
					// Ignore this context.
					delete _context->devices;
					delete _context;
					continue;
				}
			}
			n_contexts++;
			contexts[_cl_context[i]] = _context;
		}
		else{
			delete _context;
		}
	}
	// Store the cl_context's separately for easier manipulation.
	context = new cl_context[n_contexts];
	map<cl_context, context_t*>::iterator iter;
	for(i=0, iter=contexts.begin(); iter != contexts.end(); iter++, i++){
		context[i] = iter->first;
	}
}

/**
 * @brief Create one command queue for every device available for the context,
 * adding every new queue to the pool.
 *
 * @param _context The context.
 */
void clFactory::createCommandQueue(context_t *_context) {
	cl_int errNum;
	omp_lock_t *_mutex;
	cl_command_queue commandQueue = NULL;
	cl_command_queue_properties prop = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE;
#ifdef _PROFILING_
	prop |= CL_QUEUE_PROFILING_ENABLE;
#endif

	for(int i=0; i < _context->n_devices; i++){
		commandQueue = clCreateCommandQueue(_context->context, _context->devices[i], prop, &errNum);
		if (errNum != CL_SUCCESS) {
			std::cerr << "Failed to create command queue, errno=" << errNum << "." << endl;
		}
		else if (commandQueue != NULL) {
			// Add the command queue to the pool of queues.
			queues.push_back(new clQueue(_context->context, _context->devices[i], commandQueue));
			// Create one single mutex per device.
			if((_mutex = devices[_context->devices[i]]) == NULL){
				_mutex = new omp_lock_t;
				omp_init_lock(_mutex);
				devices[_context->devices[i]] = _mutex;
			}
		}
	}
}

/**
 * @brief Read the OpenCl source code.
 *
 * @return A char pointer to the source code. This pointer must be deleted later, after the use.
 */
char* clFactory::readCl(){
	FILE *fp = fopen("src/kippe_aux.cl", "r");
	if (!fp) {
		cerr << "Failed to open .cl for reading." << std::endl;
		return NULL;
	}
	fseek(fp, 0, SEEK_END);
	int sz1 = ftell(fp);
	rewind(fp);
	char *src = new char[sz1];
	int sz2 = fread(src, 1, sz1, fp);
	fclose(fp);
	if (sz1 != sz2) {
		cerr << "Fail reading .cl." << std::endl;
		delete src;
		return NULL;
	}
	return src;
}

/**
 * @brief Create an OpenCL program from the OpenCl source file and compile it.
 *
 * @param _context The context where the program will be compiled.
 * @param _device Any device from the context, only to read any compilation error.
 * @param src A char pointer to the source code.
 *
 * @return The compiled program.
 */
cl_program clFactory::createProgram(cl_context _context, cl_device_id _device, char *src) {
	cl_int errNum;
	// Create a cl_program into the context, meaning that every device present inside the context will have the compiled kernels.
	cl_program program = clCreateProgramWithSource(_context, 1, (const char**) &src, NULL, &errNum);
	if (program == NULL || errNum != CL_SUCCESS) {
		cerr << "Failed to create CL program from source (" << errNum << ")." << endl;
		return NULL;
	}

	//Passing parameters to the cl file.
	ostringstream oss;
#ifdef _DOUBLE_WORD_
	oss << "-D_DOUBLE_WORD_ -DWORD=double -DFFT_TYPE=double2 ";
#else
	oss << "-DWORD=float -DFFT_TYPE=float2 ";
#endif
	oss << "-DREDUCTION_BLOCKSZ=" << REDUCTION_BLOCKSZ << " -DREDUCTION_NBLOCKS=" << REDUCTION_NBLOCKS
			<< " -DPSF_CORNER=" << PSF_CORNER << " -DPSF_CENTER=" << PSF_CENTER << " -DPSF_EXTRACT=" << PSF_EXTRACT
			<< " -DWORD_MAX=" << WORD_MAX << " -DINT_MAX=" << INT_MAX << " -DRAND_MAX=" << RAND_MAX;

	// Build the program inside the devices under the context.
	errNum = clBuildProgram(program, 0, NULL, oss.str().c_str(), NULL, NULL);
	if (errNum != CL_SUCCESS) {
		size_t ret_val_size;
		// Get the error message from inside the device.
		//----------------------------------------------
		// Check the error message size.
		clGetProgramBuildInfo(program, _device, CL_PROGRAM_BUILD_LOG, 0, NULL, &ret_val_size);
		// Alocates only the memory needed.
		char *build_log = (char*)malloc(ret_val_size+1);
		// Read the error text.
		clGetProgramBuildInfo(program, _device, CL_PROGRAM_BUILD_LOG, ret_val_size, build_log, NULL);

		// The \0 is not mandatory in the OpenCl specification, thus we add it here for safety.
		build_log[ret_val_size] = '\0';

		printf("BUILD LOG: \n%s\n", build_log);
		free(build_log);

		return NULL;
	}
	return program;
}

/**
 * @brief Create a OpenCl kernel for a program.
 *
 * @param program The OpenCl program instance.
 * @param kernel_name The kernel name.
 *
 * @return The kernel instance.
 */
cl_kernel clFactory::createKernel(cl_program program, const char *kernel_name) {
	cl_kernel kernel = clCreateKernel(program, kernel_name, NULL);
	if (kernel == NULL){
		std::cerr << "Failed to create kernel" << kernel_name << endl;
		finalize_cl();
		exit(1);
	}
	return kernel;
}

/**
 * @brief Initialize the OpenCl instances into a given context.
 * The instances that are initialized are the kernel instances.
 * Also their mutexes are initialized at this method.
 *
 * @param context The context to have the instances initialized.
 * @param program The program instance related to this context.
 *
 * @return A {@link #kernel_set} instance containing the kernels,
 * the mutexes and the program.
 */
kernel_set* clFactory::initialize_cl(cl_context context, cl_program program) {
	kernel_set *kernels = new kernel_set;
	kernels->program = program;
	kernels->cl_generate_phase.kernel_instance = createKernel(program, "cl_generate_phase");
	kernels->cl_generate_phase.mutex = &kernels->mutex;  // For now, let's keep one single kernel mutex for the entire context.
	kernels->cl_power_spec.kernel_instance = createKernel(program, "cl_power_spec");
	kernels->cl_power_spec.mutex = &kernels->mutex;
	kernels->cl_reduce.kernel_instance = createKernel(program, "cl_reduce");
	kernels->cl_reduce.mutex = &kernels->mutex;
	kernels->cl_multiply_complexarr.kernel_instance = createKernel(program, "cl_multiply_complexarr");
	kernels->cl_multiply_complexarr.mutex = &kernels->mutex;
	kernels->cl_resize_psf.kernel_instance = createKernel(program, "cl_resize_psf");
	kernels->cl_resize_psf.mutex = &kernels->mutex;
	kernels->cl_multiply_doublearr.kernel_instance = createKernel(program, "cl_multiply_doublearr");
	kernels->cl_multiply_doublearr.mutex = &kernels->mutex;
	kernels->cl_multiply_fftw_complex_arrays.kernel_instance = createKernel(program, "cl_multiply_fftw_complex_arrays");
	kernels->cl_multiply_fftw_complex_arrays.mutex = &kernels->mutex;
	kernels->cl_calc_cost.kernel_instance = createKernel(program, "cl_calc_cost");
	kernels->cl_calc_cost.mutex = &kernels->mutex;
	kernels->cl_calc_mismatch.kernel_instance = createKernel(program, "cl_calc_mismatch");
	kernels->cl_calc_mismatch.mutex = &kernels->mutex;
	kernels->cl_adjust_fft.kernel_instance = createKernel(program, "cl_adjust_fft");
	kernels->cl_adjust_fft.mutex = &kernels->mutex;
	kernels->cl_real2complex.kernel_instance = createKernel(program, "cl_real2complex");
	kernels->cl_real2complex.mutex = &kernels->mutex;
	kernels->cl_complex2real.kernel_instance = createKernel(program, "cl_complex2real");
	kernels->cl_complex2real.mutex = &kernels->mutex;
	kernels->cl_cpso.kernel_instance = createKernel(program, "cl_cpso");
	kernels->cl_cpso.mutex = &kernels->mutex;
	kernels->cl_real.kernel_instance = createKernel(program, "cl_real");
	kernels->cl_real.mutex = &kernels->mutex;

	// Store the kernels separately so they can be finalized at the end.
	contexts[context]->kernels = kernels;
	return kernels;
}

/**
 * @brief Create a new copy of the provided kernel_set.
 *
 * The new copy is basically a copy of pointers, so there's no duplication of data on device.
 * The returned must be deleted manually later.
 *
 * @param kernels The kernel_set to be copied.
 * @param _mutex The mutex to be used for the new copy.
 * @return The new copy.
 */
kernel_set* clFactory::clone_kernels(kernel_set *kernels, omp_lock_t *_mutex) {
	kernel_set *new_kernels = new kernel_set;
	new_kernels->mutex = *_mutex;
	new_kernels->program = kernels->program;
	new_kernels->cl_generate_phase.kernel_instance = kernels->cl_generate_phase.kernel_instance;
	new_kernels->cl_generate_phase.mutex = _mutex;
	new_kernels->cl_power_spec.kernel_instance = kernels->cl_power_spec.kernel_instance;
	new_kernels->cl_power_spec.mutex = _mutex;
	new_kernels->cl_reduce.kernel_instance = kernels->cl_reduce.kernel_instance;
	new_kernels->cl_reduce.mutex = _mutex;
	new_kernels->cl_multiply_complexarr.kernel_instance = kernels->cl_multiply_complexarr.kernel_instance;
	new_kernels->cl_multiply_complexarr.mutex = _mutex;
	new_kernels->cl_resize_psf.kernel_instance = kernels->cl_resize_psf.kernel_instance;
	new_kernels->cl_resize_psf.mutex = _mutex;
	new_kernels->cl_multiply_doublearr.kernel_instance = kernels->cl_multiply_doublearr.kernel_instance;
	new_kernels->cl_multiply_doublearr.mutex = _mutex;
	new_kernels->cl_multiply_fftw_complex_arrays.kernel_instance = kernels->cl_multiply_fftw_complex_arrays.kernel_instance;
	new_kernels->cl_multiply_fftw_complex_arrays.mutex = _mutex;
	new_kernels->cl_calc_cost.kernel_instance = kernels->cl_calc_cost.kernel_instance;
	new_kernels->cl_calc_cost.mutex = _mutex;
	new_kernels->cl_calc_mismatch.kernel_instance = kernels->cl_calc_mismatch.kernel_instance;
	new_kernels->cl_calc_mismatch.mutex = _mutex;
	new_kernels->cl_adjust_fft.kernel_instance = kernels->cl_adjust_fft.kernel_instance;
	new_kernels->cl_adjust_fft.mutex = _mutex;
	new_kernels->cl_real2complex.kernel_instance = kernels->cl_real2complex.kernel_instance;
	new_kernels->cl_real2complex.mutex = _mutex;
	new_kernels->cl_complex2real.kernel_instance = kernels->cl_complex2real.kernel_instance;
	new_kernels->cl_complex2real.mutex = _mutex;
	new_kernels->cl_cpso.kernel_instance = kernels->cl_cpso.kernel_instance;
	new_kernels->cl_cpso.mutex = _mutex;
	new_kernels->cl_real.kernel_instance = kernels->cl_real.kernel_instance;
	new_kernels->cl_real.mutex = _mutex;

	return new_kernels;
}

/**
 * @brief Finalize the OpenCl instances.
 */
void clFactory::finalize_cl() {
	int i;
	for(i=0; i < n_queues; i++){
		clReleaseCommandQueue(queues[i]->getCommandQueue());
	}

	context_t *_context;
	kernel_set *kernels;
	map<cl_context, context_t*>::iterator iter;
	for(iter=contexts.begin(); iter != contexts.end(); iter++){
		// Release the kernels.
		_context = iter->second;
		kernels = _context->kernels;
		clReleaseKernel(kernels->cl_generate_phase.kernel_instance);
		clReleaseKernel(kernels->cl_power_spec.kernel_instance);
		clReleaseKernel(kernels->cl_reduce.kernel_instance);
		clReleaseKernel(kernels->cl_multiply_complexarr.kernel_instance);
		clReleaseKernel(kernels->cl_resize_psf.kernel_instance);
		clReleaseKernel(kernels->cl_multiply_doublearr.kernel_instance);
		clReleaseKernel(kernels->cl_multiply_fftw_complex_arrays.kernel_instance);
		clReleaseKernel(kernels->cl_calc_cost.kernel_instance);
		clReleaseKernel(kernels->cl_adjust_fft.kernel_instance);
		clReleaseKernel(kernels->cl_real2complex.kernel_instance);
		clReleaseKernel(kernels->cl_complex2real.kernel_instance);
		clReleaseKernel(kernels->cl_cpso.kernel_instance);
		clReleaseKernel(kernels->cl_real.kernel_instance);

		// Release the program.
		clReleaseProgram(kernels->program);

		delete _context->devices;
		delete _context->kernels;

		// Release the context.
		clReleaseContext(_context->context);
		omp_destroy_lock(&kernels->mutex);

		delete _context;
	}

	omp_lock_t *_mutex;
	map<cl_device_id, omp_lock_t*>::iterator iter2;
	for(iter2=devices.begin(); iter2 != devices.end(); iter2++){
		_mutex = iter2->second;
		omp_destroy_lock(_mutex);
		delete _mutex;
	}
}

/**
 * @brief Prepare this factory for the use.
 *
 * It's mandatory to call this method Before calling any other method
 * of this factory.
 */
void clFactory::startup(){
	startup(NULL);
}

/**
 * @brief Prepare this factory for the use, filtering only the devices specified by the vector.
 *
 * It's mandatory to call this method Before calling any other method of this factory.
 *
 * @param _devices The list of devices to be used.
 */
void clFactory::startup(vector<int> *_devices){
	omp_set_lock(&mutex);
	n_queues = 0;
	n_contexts = 0;
	initialize_gpu(_devices);
	curr_queue = 0; // The first queue is exclusive for FFT use, so curr_queue is already 0 (meaning the 0 position is in use).
	curr_context = 0;
	omp_unset_lock(&mutex);
}

/**
 * @brief Finalize this factory.
 *
 * It's mandatory to call this method after all the work is done!
 */
void clFactory::shutdown(){
	omp_set_lock(&mutex);
	finalize_cl();
	delete context;
	for(int i=0; i < n_queues; i++){
		delete queues[i];
	}
	omp_unset_lock(&mutex);

	omp_destroy_lock(&mutex);
	omp_destroy_lock(&mutex_queue);
}

/**
 * @brief Get the next command queue available for the use.
 * If none is available on the pool, just create one on the next
 * available context.
 *
 * @return The {@link #clQueue} instance containing the command queue
 * and all its configuration.
 */
clQueue* clFactory::getQueue(){
	// If the contexts were not yet initialized.
	if(n_contexts == 0){
		startup(); // No device filter.
		if(n_contexts == 0){
			return NULL;
		}
	}
	int _curr_queue;

	omp_set_lock(&mutex_queue);
	// If it's necessary to create more queues.
	if(++curr_queue == n_queues){
		do{
			// Round-robin over the contexts.
			if(++curr_context == n_contexts){
				curr_context = 0;
			}
			// Create one additional queue for every device present in the next context.
			createCommandQueue(contexts[context[curr_context]]);
		}while(n_queues == (int)queues.size());
		// Set the correct kernels for the newly created queues.
		kernel_set *kernels = contexts[context[curr_context]]->kernels;
		omp_lock_t *_mutex;
		for(int i=n_queues; i < (int)queues.size(); i++){
			// Set the correct mutex based on the device where the queue was created.
			_mutex = devices[queues[i]->getDevice()];
			queues[i]->kernels = clone_kernels(kernels, _mutex);
		}
		n_queues = queues.size();
	}
	_curr_queue = curr_queue;
	omp_unset_lock(&mutex_queue);

	return queues[_curr_queue];
}

/**
 * @brief Dispose the clQueue instance, putting it back to the pool.
 *
 * @param clqueue The clQueue instance to be released.
 */
void clFactory::disposeQueue(clQueue *clqueue){
	omp_set_lock(&mutex_queue);
	for(int i=0; i < n_queues; ++i) {
		if(queues[i] == clqueue){
			// Move the queue to the last position in the pool, recycling it.
			queues.erase(queues.begin() + i);
			queues.push_back(clqueue);
			curr_queue--;
			break;
		}
	}
	omp_unset_lock(&mutex_queue);
}

/**
 * @brief Get the device name from the queue provided.
 *
 * @param queue The queue.
 * @param deviceName A pointer where to put the device name.
 * @param nameSize The memory size where the pointer points to.
 */
void clFactory::getDeviceName(clQueue* queue, char *deviceName, int nameSize){
	omp_set_lock(&mutex_queue);
	size_t sz;
	clGetDeviceInfo(queue->getDevice(), CL_DEVICE_NAME, nameSize, deviceName, &sz);
	deviceName[sz] = '\0';
	omp_unset_lock(&mutex_queue);
}

/**
 * @brief Return the number of contexts available for use.
 *
 * @return The number of contexts available for use.
 */
int clFactory::getNumAvailableContexts(){
	return n_contexts;
}

/**
 * @brief Return the number of devices available for use.
 *
 * @return The number of devices available for use.
 */
int clFactory::getNumAvailableDevices(){
	return devices.size();
}

/**
 * @brief Read directly from the hardware the number of devices available for use.
 *
 * @return The number of devices available for use, without consulting the internal cached data.
 */
int clFactory::countDevicesFromHardware() {
	cl_int errNum;
	size_t deviceBufferSize = -1;
	int n_devices = 0;

	for(int i=0; i < n_contexts; i++){
		// First get the size of the devices buffer
		errNum = clGetContextInfo(context[i], CL_CONTEXT_DEVICES, 0, NULL, &deviceBufferSize);
		if (errNum == CL_SUCCESS) {
			n_devices += deviceBufferSize / sizeof(cl_device_id);
		}
	}

	return n_devices;
}

/**
 * @brief Get all devices available within a context.
 *
 * @param _context The context where to look for devices.
 * @param context_data The context_t pointer where to set the data. This pointer must be previously allocated.
 */
void clFactory::getDevices(cl_context _context, context_t *context_data){
	context_data->context = _context;

	cl_int errNum;
	size_t deviceBufferSize = -1;

	// First get the size of the devices buffer
	errNum = clGetContextInfo(_context, CL_CONTEXT_DEVICES, 0, NULL, &deviceBufferSize);
	if (errNum != CL_SUCCESS || deviceBufferSize <= 0) {
		context_data->n_devices = 0;
		return;
	}

	// Allocate memory for the devices buffer
	context_data->n_devices = deviceBufferSize / sizeof(cl_device_id);
	context_data->devices = new cl_device_id[context_data->n_devices];
	// Get the device IDs available in the context (notice that such ids are more like internal pointers but not sequential numbers).
	errNum = clGetContextInfo(_context, CL_CONTEXT_DEVICES, deviceBufferSize, context_data->devices, NULL);
	if (errNum != CL_SUCCESS) {
		context_data->n_devices = 0;
		delete context_data->devices;
	}
}
