
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * cl_factory.hpp
 *
 *  Created on: 15/07/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#ifndef CL_FACTORY_HPP_
#define CL_FACTORY_HPP_

#include <iostream>
#include <map>
#include <CL/cl.h>
#include <CL/opencl.h>
#include <CL/cl_platform.h>
#include "viennacl/vector.hpp"
#include "config.hpp"
#include "util.hpp"

using namespace std;

/**
 * @brief This struct data type contains the OpenCl cl_kernel instance
 * and its mutex to synchronize the calls.
 */
typedef struct {
	cl_kernel kernel_instance;
	omp_lock_t *mutex;
} kernel_t;

/**
 * @brief This struct data type contains the complete kernel configuration
 * and its mutex to synchronize the calls.
 */
typedef struct {
	cl_program program;
	kernel_t cl_generate_phase, cl_power_spec, cl_reduce, cl_multiply_complexarr, cl_resize_psf,
		cl_multiply_doublearr, cl_multiply_fftw_complex_arrays, cl_calc_cost, cl_adjust_fft,
		cl_calc_mismatch, cl_real2complex, cl_complex2real, cl_cpso, cl_real;
	omp_lock_t mutex;
} kernel_set;

typedef struct {
	cl_context context;
	cl_device_id *devices;
	int n_devices;
	kernel_set *kernels;
} context_t;

/**
 * @brief This class contains an OpenCl environment ready for the use.
 */
class clQueue{
	cl_context context;
	cl_device_id device;
	cl_command_queue command_queue;

public:
	kernel_set *kernels;

	clQueue(cl_context _context, cl_device_id _device, cl_command_queue _command_queue){
		context = _context;
		device = _device;
		command_queue = _command_queue;
	}

	cl_context getContext() {return context;}

	cl_command_queue getCommandQueue() {return command_queue;}

	cl_program getProgram() {return (kernels!=NULL) ? kernels->program : NULL;}

	cl_device_id getDevice() {return device;}
};

/**
 * @brief The class is a factory for the OpenCl contexts, command queues and kernels.
 * It is also a pool manager for the queues created.
 *
 * It manages all the initialization and finalization procedures, providing a {@link #clQueue}
 * instance entirely ready for the use. Every queue is created on a distinct device (a context
 * here uses only a single device) and the devices are selected in round-robin sequence.
 */
class clFactory{
	static cl_context *context;
	static map<cl_context, context_t*> contexts;
	static map<cl_device_id, omp_lock_t*> devices;

	static omp_lock_t mutex, mutex_queue;
	static int n_contexts;
	static int n_queues;
	static int curr_queue;
	static int curr_context;
	static vector<clQueue*> queues;

	static void initialize_gpu(vector<int> *_devices);

	static void createContext(vector<int> *_devices);

	static void  createCommandQueue(context_t *_context);

	static char* readCl();

	static cl_program createProgram(cl_context _context, cl_device_id _device, char *src);

	static cl_kernel createKernel(cl_program program, const char *kernel_name);

	static kernel_set* initialize_cl(cl_context context, cl_program program);

	static kernel_set* clone_kernels(kernel_set *kernels, omp_lock_t *_mutex);

	static void finalize_cl();

	static int countDevicesFromHardware();

	static void getDevices(cl_context _context, context_t *context_data);

public:
	static void startup();

	static void startup(vector<int> *devices);

	static void shutdown();

	static clQueue* getQueue();

	static void disposeQueue(clQueue *clqueue);

	static void getDeviceName(clQueue* queues, char *deviceName, int nameSize);

	static int getNumAvailableContexts();

	static int getNumAvailableDevices();

};

#endif /* CL_FACTORY_HPP_ */
