
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * kippe_aux.cpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "cpso.hpp"

using namespace std;

map<cl_context, cl_mem*> CPSO::static_references;
int *CPSO::diffraction_mask;
omp_lock_t CPSO::mutex;
omp_lock_t CPSO::mutex_fft;
omp_lock_t Debug::mutex1;
omp_lock_t Debug::mutex2;
bool CPSO::lock_initialized = false;
bool Debug::lock_initialized = false;
char Debug::buffer[DEFAULT_BUFFER_SIZE];
int CPSO::seq = 0;

/**
 * @brief The Contructor.
 *
 * It is responsible for requesting a queue for the OpenCl factory and
 * for allocating all the static and non-static data.
 *
 * @param _zernikes The complete Zernike matrix.
 * @param _phase_mask The phase mask.
 * @param _diffraction_mask The diffraction mask.
 * @param _phase_size The phase mask.
 * @param _image_size The image width.
 * @param _n_zernikes The number of Zernike terms used for the calculations.
 * @param _psf_range The Zernike coefficient range.
 * @param _w The parameter w for the CPSO search.
 * @param c1 The parameter c1 for the CPSO search.
 * @param c2 The parameter c2 for the CPSO search.
 * @param _reset_at The particle reset factor for the CPSO search.
 * @param _n_particles The number of particles for the CPSO search.
 * @param _n_swarms The number of swarms for the CPSO search.
 */
CPSO::CPSO(double* _zernikes, int *_phase_mask, int *_diffraction_mask, int _phase_size, int _image_size, int _n_zernikes,
		int _psf_range, WORD _w, WORD c1, WORD c2, WORD _reset_at, int _n_particles, int _n_swarms){
	startup_locks();

	omp_set_lock(&mutex);
	UID = seq++;
	omp_unset_lock(&mutex);

	buffer = new char[DEFAULT_BUFFER_SIZE];
	queue = clFactory::getQueue();
	context = queue->getContext();
	command_queue = queue->getCommandQueue();
	kernels = queue->kernels;
	has_startup_coefs = false;
	store_static_data(context, _n_zernikes, _phase_size, _image_size, _zernikes, _phase_mask, _diffraction_mask);
	allocate_data(_phase_size, _image_size, _n_zernikes, _psf_range, _w, c1, c2, _reset_at, _n_particles, _n_swarms);
}

/**
 * @brief The destructor.
 */
CPSO::~CPSO(){
	// Release the internal buffer.
	delete buffer;
}

/**
 * @brief This method is responsible for storing the static data.
 *
 * It stores the static data Only Once for every available context.
 * It's the developer's responsibility to call the static method {@link #clear_static_data}
 * after all the processing in all available CPSOs is done.
 * This method can be called several times during the application life time.
 *
 * @param _context The OpenCl context.
 * @param _n_zernikes The number of Zernike terms used for the calculations.
 * @param _phase_size The phase width.
 * @param _image_size The image width.
 * @param zernikes The complete Zernike matrix.
 * @param phase_mask The phase mask.
 * @param _diffraction_mask The diffraction mask.
 */
void CPSO::store_static_data(cl_context _context, int _n_zernikes, int _phase_size, int _image_size, double* zernikes,
		int *phase_mask, int *_diffraction_mask){
	// If the data was already stored into this context, leave this method.
	if(static_references.count(_context) != 0){
		return;
	}

	int _img_area = _image_size * _image_size;
	int _size_fft = _phase_size * _phase_size;
	int _z_size = _n_zernikes * _size_fft;

	WORD *_zernikes = new WORD[_z_size];
	for(int i = 0; i < _z_size; i++) _zernikes[i] = zernikes[i];
	cl_mem _cl_zernikes, _cl_phase_mask, _cl_diffraction_mask;
	CREATE_BUFFER(_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, _z_size * SIZEOF_WORD, _zernikes, _cl_zernikes);
	CREATE_BUFFER(_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, _size_fft * sizeof(int), phase_mask, _cl_phase_mask);
	CREATE_BUFFER(_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, _img_area * sizeof(int), _diffraction_mask, _cl_diffraction_mask);//img_areah

	delete _zernikes;

	cl_mem *static_data = new cl_mem[3];
	static_data[0] = _cl_zernikes;
	static_data[1] = _cl_phase_mask;
	static_data[2] = _cl_diffraction_mask;
	static_references[_context] = static_data;

	diffraction_mask = new int[_img_area];
	memcpy(diffraction_mask, _diffraction_mask, _img_area * sizeof(int));
}

/**
 * @brief Clear all the static data.
 *
 * This method Must be called only Once during the application life time!
 */
void CPSO::clear_static_data(){
	cl_mem *static_data;
	map<cl_context, cl_mem*>::iterator iter;
	for(iter=static_references.begin(); iter != static_references.end(); iter++){
		static_data = iter->second;
		clReleaseMemObject(static_data[0]);
		clReleaseMemObject(static_data[1]);
		clReleaseMemObject(static_data[2]);
		delete static_data;
	}
	destroy_locks();
	delete diffraction_mask;
}

/**
 * @brief Startup the locks for parallel executions.
 */
void CPSO::startup_locks(){
	if(!lock_initialized){
		omp_init_lock(&mutex);
		omp_init_lock(&mutex_fft);
		Debug::startup_locks();
		lock_initialized = true;
	}
}

/**
 * @brief Destroy all locks for parallel execution.
 */
void CPSO::destroy_locks(){
	if(lock_initialized){
		omp_destroy_lock(&mutex);
		omp_destroy_lock(&mutex_fft);
		Debug::destroy_locks();
		lock_initialized = false;
	}
}

/**
 * @brief This method allocates all non-static data for the search.
 *
 * @param _phase_size The phase width.
 * @param image_size The image width.
 * @param _n_zernikes The number of Zernike terms used for the calculations.
 * @param _psf_range The Zernike coefficient range.
 * @param _w The parameter w for the CPSO search.
 * @param c1 The parameter c1 for the CPSO search.
 * @param c2 The parameter c2 for the CPSO search.
 * @param _reset_at The particle reset factor for the CPSO search.
 * @param _n_particles The number of particles for the CPSO search.
 * @param _n_swarms The number of swarms for the CPSO search.
 */
void CPSO::allocate_data(int _phase_size, int image_size, int _n_zernikes, int _psf_range,
		WORD _w, WORD c1, WORD c2, WORD _reset_at, int _n_particles, int _n_swarms) {

	// Set the simple Host variables.
	threads = DEFAULT_BLOCKSZ;
	psf_range = _psf_range;
	img_size = image_size;
	phase_size = _phase_size;
	n_zernikes = _n_zernikes;
	img_sizeh = (img_size / 2) + 1;
	img_area = img_size * img_size;
	img_areah = img_size * img_sizeh;
	size_fft = phase_size * phase_size;
	z_size = n_zernikes * size_fft;

	n_particles = _n_particles;
	n_swarms = _n_swarms;
	int swarm_sizes[n_swarms];
	float swarm_dim = n_zernikes / (float)n_swarms;
	for(int i=0; i < n_swarms; i++){
		swarm_sizes[i] = (i + 1) * swarm_dim - 1;
	}
	reset_at = _reset_at;
	w = _w;

	startup_coefs = new WORD[n_zernikes];
	int n_psfs = n_particles * n_swarms;

	// Allocate memory on device and set its contents.
	// For the PSF.
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * size_fft * SIZEOF_FFTTYPE, NULL, cl_pupil);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * size_fft * SIZEOF_WORD, NULL, cl_phase);// Only necessary for validation purpose.
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * img_area * SIZEOF_FFTTYPE, NULL, cl_fft_psfe);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * img_area * SIZEOF_FFTTYPE, NULL, cl_fft_conobj);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * REDUCTION_NBLOCKS * SIZEOF_WORD, NULL, cl_sum);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * img_area * SIZEOF_WORD, NULL, cl_cost);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_WORD, NULL, cl_mismatch);

	// For the CPSO.
	CREATE_BUFFER(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, n_swarms * sizeof(int), swarm_sizes, cl_swarm_dim);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * n_zernikes * SIZEOF_WORD, NULL, cl_coefs);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_particles * n_zernikes * SIZEOF_WORD, NULL, cl_speed);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * n_zernikes * SIZEOF_WORD, NULL, cl_pbest);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_zernikes * SIZEOF_WORD, NULL, cl_gbest);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * SIZEOF_WORD, NULL, cl_pbest_value);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_swarms * SIZEOF_WORD, NULL, cl_gbest_value);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * sizeof(int), NULL, cl_reset_search);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * sizeof(uint4), NULL, cl_rand_ctx);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * SIZEOF_WORD, NULL, cl_w);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * SIZEOF_WORD, NULL, cl_c1);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_psfs * SIZEOF_WORD, NULL, cl_c2);

	// Remaining objects.
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_WORD, NULL, cl_best_conobj);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, size_fft * SIZEOF_WORD, NULL, cl_best_phase);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, size_fft * SIZEOF_WORD, NULL, cl_best_psf);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_WORD, NULL, cl_best_psfe);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, n_zernikes * SIZEOF_WORD, NULL, cl_best_coefs);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_best_fft_psfe);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_best_fft_conobj);
	CREATE_BUFFER(context, CL_MEM_READ_ONLY, img_area * SIZEOF_FFTTYPE, NULL, cl_object);
	CREATE_BUFFER(context, CL_MEM_READ_ONLY, img_area * SIZEOF_FFTTYPE, NULL, cl_image);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_fft_object);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_fft_image);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_fft_original_psf);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, z_size * SIZEOF_FFTTYPE, NULL, cl_debug_info);

	// Get the references for the static values.
	cl_mem *static_data = static_references[context];
	cl_zernikes = static_data[0];
	cl_phase_mask = static_data[1];
	cl_diffraction_mask = static_data[2];

	DEBUG(int, command_queue, "phase-mask", cl_phase_mask, phase_size, phase_size);
	DEBUG(int, command_queue, "difraction-mask", cl_diffraction_mask, img_size, img_size);//img_size, img_sizeh

	// Initialize the FFT contexts.
	vn_object = new viennacl::vector<WORD>(cl_object, img_area*2);
	vn_image = new viennacl::vector<WORD>(cl_image, img_area*2);
	vn_fft_object = new viennacl::vector<WORD>(cl_fft_object, img_area*2);
	vn_fft_image = new viennacl::vector<WORD>(cl_fft_image, img_area*2);
	vn_fft_original_psf = new viennacl::vector<WORD>(cl_fft_original_psf, img_area*2);
	vn_fft_conobj = new viennacl::vector<WORD>(cl_fft_conobj, img_area*2);

	// Each particle will have its own PSF instance.
	psf = new PSF*[n_psfs];
	for(int i=0; i < n_psfs; i++){
		psf[i] = new PSF(context, command_queue, img_size, phase_size, n_zernikes);
		// Set the PSF references into the major matrixes.
		psf[i]->setCoefsPosition(cl_coefs, i * n_zernikes);
		psf[i]->setPupilPosition(cl_pupil, i * size_fft);
	}
	original_psf_fft = new FFT_TYPE[img_area];

	// Set CPSO search parameters.
	setW(w);
	setC1(c1);
	setC2(c2);

	SYNC_QUEUE
}

/**
 * @brief This method makes the final preparation for the search.
 *
 * It must be called to set the coefficients, the object and the image.
 *
 * @param object The object that will be used as starting point to obtain the image.
 * @param image The image that must be obtained by the search.
 */
void CPSO::set_images(double *object, double *image) {
	has_psf_original = false;

	// Create an array with values zero to clear the device memory.
	FFT_TYPE *zero = (FFT_TYPE*)buffer;
	memset(zero, 0, z_size * SIZEOF_FFTTYPE);

	int i;
	WORD worst_value[n_swarms];
	// Worst possible value is the maximum possible value (minimization problem).
	for(i=0; i < n_swarms; i++) worst_value[i] = WORD_MAX;

	int n_coefs = n_particles * n_zernikes;
	WORD coefs[n_coefs * n_swarms];
	if(has_startup_coefs){
		// NOTE: The PSO results CANNOT be directly compared to the Simulated Annealing results!
		//       This happens because the SA calculates Only 1 Zernike at a time and has many controls to accept
		//       the coefficient or not, what makes it very difficult to force it to accept manual coefs.
		//       At same time, the PSO pupil is totally recalculated every time, whereas the SA one has
		//       the intermediate values stored, what cause slight differences that will be cascade all
		//       over the computation, resulting in a near but Not exact costs (PSO cost x SA cost).
		// Initialize 1 particle with the static startup coefficients provided.
		memcpy(coefs, startup_coefs, n_zernikes * SIZEOF_WORD);
		// Replicate these same coefs to all particles into the first swarm.
		for(i=1; i < n_particles; i++) memcpy(&coefs[i * n_zernikes], coefs, n_zernikes * SIZEOF_WORD);
	}
	else{
		//generateRandomCoefs(coefs, n_coefs, psf_range);
		generateNormalDistrRandomCoefs(coefs, n_coefs, psf_range);
        // Generate random coefficients for every dimension in every particle into the first swarm.
        //for(i=0; i < n_coefs; i++) coefs[i] = getRandCoef(psf_range);
	}
	has_startup_coefs = false;

	// Replicate the particle coefficients to all remaining swarms.
	for(i=1; i < n_swarms; i++) memcpy(&coefs[i * n_coefs], coefs, n_coefs * SIZEOF_WORD);

	// Generate random speeds for every dimension in every particle.
	// Speeds should not be replicated for every swarm, since it will be used Only to update
	// the next position for the own swarm's dimension interval, thus the remaining intervals
	// would be useless.
	WORD speed[n_coefs];
	// Max.speed is half search space.
	for(i=0; i < n_coefs; i++) speed[i] = getRandCoef(psf_range / 2);

	// Create a random seed for every particle.
	int n_psfs = n_particles * n_swarms;
	uint4 rand_ctx[n_psfs];
	for(i=0; i < n_psfs; i++) {rand_ctx[i].x = rand(); rand_ctx[i].y = rand(); rand_ctx[i].z = rand(); rand_ctx[i].w = rand(); }

	//Regardless of the FITS data type (double or float), convert it to the type currently in use.
	FFT_TYPE *_image = new FFT_TYPE[img_area];
	for(i = 0; i < img_area; i++) {_image[i].x = image[i]; _image[i].y = 0;}
	FFT_TYPE *_object = new FFT_TYPE[img_area];
	for(i = 0; i < img_area; i++) {_object[i].x = object[i]; _object[i].y = 0;}

	// Copy the values to the device.
	// For the CPSO.
	clMemcpyHostToDevice(command_queue, cl_coefs, coefs, n_coefs * n_swarms * SIZEOF_WORD);
	clMemcpyHostToDevice(command_queue, cl_speed, speed, n_coefs * SIZEOF_WORD);
	// Set the personal best position as the initial position.
	clMemcpyHostToDevice(command_queue, cl_pbest, coefs, n_coefs * n_swarms * SIZEOF_WORD);
	// Set the global best position as zero, cause it must be calculated before used (thus there's no initial value).
	clMemcpyHostToDevice(command_queue, cl_gbest, zero, n_zernikes * SIZEOF_WORD);
	// Set the global best value as the worst possible value.
	clMemcpyHostToDevice(command_queue, cl_gbest_value, worst_value, n_swarms * SIZEOF_WORD);
	clMemcpyHostToDevice(command_queue, cl_reset_search, zero, n_psfs * sizeof(int));
	clMemcpyHostToDevice(command_queue, cl_rand_ctx, rand_ctx, n_psfs * sizeof(uint4));

	// For the remaining objects.
	clMemcpyHostToDevice(command_queue, cl_debug_info, zero, z_size * SIZEOF_FFTTYPE);
	clMemcpyHostToDevice(command_queue, cl_object, _object, img_area * SIZEOF_FFTTYPE);
	clMemcpyHostToDevice(command_queue, cl_image, _image, img_area * SIZEOF_FFTTYPE);
	clMemcpyHostToDevice(command_queue, cl_fft_image, zero, img_area * SIZEOF_FFTTYPE);

	delete _image;
	delete _object;

	// Reset the number of evaluations.
	n_psf_evals = 0;
	max_evals = -1;
	convergence_stable_cycle = -1;

	SYNC_QUEUE

	// Calculate the Object's and the Image's FFTs.
	omp_set_lock(&mutex_fft);
	viennacl::fft(*vn_object, *vn_fft_object);
	viennacl::fft(*vn_image, *vn_fft_image);
	omp_unset_lock(&mutex_fft);

	DEBUG_2D(WORD, command_queue, "coefs", cl_coefs, n_particles, n_zernikes);
	DEBUG_COMPLEX(command_queue, "object", cl_object, img_size, img_size);
	DEBUG_COMPLEX(command_queue, "fft-object", cl_fft_object, img_size, img_size);
	DEBUG_COMPLEX(command_queue, "image", cl_image, img_size, img_size);
	DEBUG_COMPLEX(command_queue, "fft-image", cl_fft_image, img_size, img_size);
}

/**
 * @brief This method makes the CPSO startup, a mandatory step for this search method.
 *
 * @param trackers The trackers.
 */
void CPSO::startup(TimeTracker **trackers) {
	// Run the PSF calculation once for every particle,
	// obtaining a starting point for the search technique.
	runPsf(trackers, n_particles);

	TRACK(if(trackers!=NULL)trackers[1]->resume())
	int i, pos;
	WORD cost[n_particles * n_swarms];
	WORD partial[n_particles * REDUCTION_NBLOCKS];

	// Finalize the reduction.
	clMemcpyDeviceToHost(command_queue, partial, cl_sum, n_particles * REDUCTION_NBLOCKS * SIZEOF_WORD);

	// The CPSO startup calculation is composed of 'n_particles' PSFs,
	// whose results are replicated for all remaining swarms.
	for(i=0, pos=0; i < n_particles; i++){
		cost[i] = 0;
		// Finalizes the reduction for every particle calculated.
		for(int j=0; j < REDUCTION_NBLOCKS; j++){
			cost[i] += partial[pos++];
		}
	}

	// Repeat the cost for every swarm.
	for(i=1; i < n_swarms; i++) memcpy(&cost[i * n_particles], cost, n_particles * SIZEOF_WORD);
	clMemcpyHostToDevice(command_queue, cl_pbest_value, cost, n_swarms * n_particles * SIZEOF_WORD);

	// Repeat the partial costs for every swarm, since it's required by the CPSO kernel.
	for(i=0; i < n_swarms; i++){
		clMemcpyHostToDeviceOffset(command_queue, cl_sum, i * n_particles * REDUCTION_NBLOCKS * SIZEOF_WORD, partial, n_particles * REDUCTION_NBLOCKS * SIZEOF_WORD);
	}
	TRACK(if(trackers!=NULL)trackers[1]->pause())
}

/**
 * @brief Finalize all internal objects related to this CPSO instance.
 */
void CPSO::finalize_cl() {
	// Release OpenCl instances.
	clReleaseMemObject(cl_coefs);
	clReleaseMemObject(cl_pupil);
	clReleaseMemObject(cl_phase);
	clReleaseMemObject(cl_cost);
	clReleaseMemObject(cl_mismatch);
	clReleaseMemObject(cl_sum);
	clReleaseMemObject(cl_fft_conobj);
	clReleaseMemObject(cl_fft_psfe);

	clReleaseMemObject(cl_swarm_dim);
	clReleaseMemObject(cl_speed);
	clReleaseMemObject(cl_pbest);
	clReleaseMemObject(cl_gbest);
	clReleaseMemObject(cl_pbest_value);
	clReleaseMemObject(cl_gbest_value);
	clReleaseMemObject(cl_reset_search);
	clReleaseMemObject(cl_rand_ctx);
	clReleaseMemObject(cl_w);
	clReleaseMemObject(cl_c1);
	clReleaseMemObject(cl_c2);

	clReleaseMemObject(cl_object);
	clReleaseMemObject(cl_image);
	clReleaseMemObject(cl_debug_info);
	clReleaseMemObject(cl_fft_object);
	clReleaseMemObject(cl_fft_image);
	clReleaseMemObject(cl_best_phase);
	clReleaseMemObject(cl_best_psf);
	clReleaseMemObject(cl_best_psfe);
	clReleaseMemObject(cl_best_fft_psfe);
	clReleaseMemObject(cl_best_conobj);
	clReleaseMemObject(cl_fft_original_psf);
	clReleaseMemObject(cl_best_fft_conobj);
	clReleaseMemObject(cl_best_coefs);

	// Dispose the command queue.
	clFactory::disposeQueue(queue);

	// Delete the Host pointers.
	int n_psfs = n_particles * n_swarms;
	for(int i=0; i < n_psfs; i++){
		delete psf[i];
	}
	delete[] psf;

	delete startup_coefs;
	delete original_psf_fft;

	// Release the FFT contexts.
	delete vn_object;
	delete vn_image;
	delete vn_fft_object;
	delete vn_fft_image;
	delete vn_fft_original_psf;
	delete vn_fft_conobj;
}

/**
 * @brief This method runs the search of the object over the image.
 *
 * It must be called to start the CPSO search.
 *
 * @param trackers The tracker.
 * @param n_cycles The number of search cycles to be executed.
 */
void CPSO::run(TimeTracker **trackers, int n_cycles){
	if(n_cycles <= 0){
		return;
	}

	// Initializes the CPSO calculation.
	startup(trackers);

	// Begins the calculations.
	int j;
	ostringstream oss;
	WORD bkp_coefs[n_zernikes];
	int n_psfs = n_swarms * n_particles;
	uint4 rand_ctx[n_psfs];
	double all_costs[n_cycles], all_psf_diff[n_cycles];
	int tmp, i = 0;
	do{
		do{
			if(i > 0){
				// Calculates the current PSF.
				runPsf(trackers);
			}
			// Runs the CPSO to obtain the new coefficients.
			runCPSO(trackers, n_cycles);

			TRACK(if(trackers!=NULL)trackers[1]->resume())

			all_costs[i] = getMinCost();
			// Repeat the PSF difference is the cost has not reduced, because it's still the same PSF calculated.
			all_psf_diff[i] = (i==0 || all_costs[i]<all_costs[i-1]) ? calcPsfDifferences() : all_psf_diff[i-1];

			oss << "PSO#" << UID << " Cycle#" << i << " -> (cost " << all_costs[i] << ", psf_diff " << all_psf_diff[i] << ", " << getNPsfEvals() << " evals) [";
			clMemcpyDeviceToHostOffset(command_queue, bkp_coefs, cl_pbest, getBestPsfPos() * n_zernikes * SIZEOF_WORD, n_zernikes * SIZEOF_WORD);
			for(j=0; j < n_zernikes; j++) oss << bkp_coefs[j] << ", ";
			oss << "]";
			Debug::debug(oss.str().c_str());
			oss.str("");

			// Refresh the particle's seeds to avoid the problem of
			// exhausting the random numbers for the seed, keeping a good randomization.
			for(j=0; j < n_psfs; j++) {rand_ctx[j].x = rand(); rand_ctx[j].y = rand(); rand_ctx[j].z = rand(); rand_ctx[j].w = rand(); }
			clMemcpyHostToDevice(command_queue, cl_rand_ctx, rand_ctx, n_psfs * sizeof(uint4));

			i++;
			SYNC_QUEUE;

			TRACK(if(trackers!=NULL)trackers[1]->pause())
		}while(i % n_cycles != 0);

		TRACK(if(trackers!=NULL)trackers[1]->resume())

		// Make the backup for the initial coefficients.
		clMemcpyDeviceToHost(command_queue, bkp_coefs, cl_coefs, n_zernikes * SIZEOF_WORD);

		// Calculate the best search result.
		//----------------------------------
		clMemcpyDeviceToDeviceOffset(command_queue, cl_coefs, 0, cl_pbest, getBestPsfPos() * n_zernikes * SIZEOF_WORD, n_zernikes * SIZEOF_WORD, true);
		tmp = n_psf_evals;

		TRACK(if(trackers!=NULL)trackers[1]->pause())

		runPsf(trackers, 1);

		TRACK(if(trackers!=NULL)trackers[1]->resume())
		n_psf_evals = tmp;
		saveFirstResult();
		TRACK(if(trackers!=NULL)trackers[1]->pause())

		TRACK(if(trackers!=NULL)trackers[5]->resume())
		WORD _gbest_cost;
		DEBUG(WORD, command_queue, "sum-cost-gbest",  cl_sum, 1, REDUCTION_NBLOCKS);
		FINAL_REDUCE(command_queue,  cl_sum, _gbest_cost);
		CHECKSUM(WORD, command_queue,  cl_cost, img_area, _gbest_cost);//img_areah
		gbest_cost = _gbest_cost;
		TRACK(if(trackers!=NULL)trackers[5]->pause())
		//----------------------------------

		TRACK(if(trackers!=NULL)trackers[1]->resume())
		oss << "Final GBest Cost: = " << gbest_cost;
		Debug::debug(oss.str().c_str());
		oss.str("");

		// Finds the cycle where the convergence occurred.
		double mean = calc_mean(all_costs, n_cycles);
		double stddev = calc_stddev(all_costs, n_cycles);
		for(j=0; j < n_cycles; j++){
			if(all_costs[j] <= (mean + stddev)){
				setStableCycle(j + 1);
				break;
			}
		}

		// Restore the backup of the initial coefficients.
		clMemcpyHostToDevice(command_queue, cl_coefs, bkp_coefs, n_zernikes * SIZEOF_WORD);

		// Restore the w to be able to run the CPSO again.
		setW(w);
		TRACK(if(trackers!=NULL)trackers[1]->pause())

	}while(getNPsfEvals() < max_evals);
}

/**
 * @brief Runs 'number of particles * number of swarms' PSFs,
 * that were configured previously for this CPSO instance.
 *
 * @param trackers The trackers.
 */
void CPSO::runPsf(TimeTracker **trackers){
	runPsf(trackers, n_particles * n_swarms);
}

/**
 * @brief Runs 'n_psfs' PSFs.
 *
 * @param trackers The trackers.
 * @param n_psfs The number of PSFs to be calculated.
 */
void CPSO::runPsf(TimeTracker **trackers, int n_psfs){
	generatePhase(trackers, n_psfs);
	makePsf(trackers, n_psfs);
	convolveObj(trackers, n_psfs);
	calcCost(trackers, n_psfs);
	n_psf_evals += n_psfs;
}

/**
 * @brief Run the CPSO search.
 *
 * @param tracker The tracker.
 * @param n_cycles The number of cycles that this search was initially configured to run.
 */
void CPSO::runCPSO(TimeTracker **tracker, int n_cycles){
	int max_threads = n_particles * n_swarms;

	TRACK(if(tracker!=NULL)tracker[1]->resume())
	CALL_KERNEL(command_queue, kernels->cl_cpso, max_threads, n_particles, 20,
		{sizeof(cl_mem), (void*)&cl_coefs},
		{sizeof(cl_mem), (void*)&cl_speed},
		{sizeof(cl_mem), (void*)&cl_pbest},
		{sizeof(cl_mem), (void*)&cl_gbest},
		{n_particles * SIZEOF_WORD, NULL},
		{sizeof(cl_mem), (void*)&cl_pbest_value},
		{sizeof(cl_mem), (void*)&cl_gbest_value},
		{sizeof(cl_mem), (void*)&cl_reset_search},
		{sizeof(cl_mem), (void*)&cl_swarm_dim},
		{sizeof(cl_mem), (void*)&cl_sum},
		{sizeof(cl_mem), (void*)&cl_rand_ctx},
		{sizeof(cl_mem), (void*)&cl_w},
		{sizeof(CL_WORD), (void*)&w},
		{sizeof(cl_mem), (void*)&cl_c1},
		{sizeof(cl_mem), (void*)&cl_c2},
		{sizeof(cl_int), (void*)&n_particles},
		{sizeof(cl_int), (void*)&n_swarms},
		{sizeof(CL_WORD), (void*)&psf_range},
		{sizeof(CL_WORD), (void*)&reset_at},
		{sizeof(cl_int), (void*)&n_cycles}
    );

	SYNC_QUEUE

	TRACK(if(tracker!=NULL)tracker[1]->pause())
}

/**
 * @brief This method calculates the phase and the pupil.
 *
 * @param tracker The tracker.
 * @param n_psfs The number of PSFs to be calculated.
 */
void CPSO::generatePhase(TimeTracker **tracker, int n_psfs) {
	int max_threads = n_psfs * phase_size;

	TRACK(if(tracker!=NULL)tracker[2]->resume())
	CALL_KERNEL(command_queue, kernels->cl_generate_phase, max_threads, phase_size, 9,
		{sizeof(cl_mem), (void*)&cl_zernikes},
		{sizeof(cl_mem), (void*)&cl_coefs},
		{sizeof(cl_mem), (void*)&cl_phase_mask},
		{sizeof(cl_mem), (void*)&cl_pupil},
		{sizeof(cl_mem), (void*)&cl_phase}, // This parameter, Only for validation purposes.
		{SIZEOF_WORD*n_zernikes, NULL},
		{sizeof(cl_int), (void*)&phase_size},
		{sizeof(cl_int), (void*)&n_zernikes},
		{sizeof(cl_int), (void*)&max_threads}
    );
	SYNC_QUEUE
	DEBUG_COMPLEX3D(command_queue, "pupil", cl_pupil, phase_size, phase_size, n_psfs);

	// Move the pupil information into every separate PSF instance.
	for(int i=0; i < n_psfs; i++){
		psf[i]->refreshPupil(false);
	}
	SYNC_QUEUE

	TRACK(if(tracker!=NULL)tracker[2]->pause())
}

/**
 * @brief Calculate the basic PSF information.
 *
 * @param tracker The tracker.
 * @param n_psfs The number of PSFs to be calculated.
 */
void CPSO::makePsf(TimeTracker **tracker, int n_psfs) {
	int blocks_fft = ceil(size_fft / (float) MATRIX_OP_BLOCKSZ);
	int blocks_img = ceil(img_area / (float) MATRIX_OP_BLOCKSZ);
	int i;

	TRACK(if(tracker!=NULL)tracker[3]->resume());
	// Calculate the FFT of the pupils.
	for(i=0; i < n_psfs; i++){
		omp_set_lock(&mutex_fft);
		psf[i]->pupilToFocusFft();
		omp_unset_lock(&mutex_fft);
	}

	for(i=0; i < n_psfs; i++){
		DEBUG_COMPLEX(command_queue, "focus", psf[i]->cl_focus, phase_size, phase_size);
	}
	TRACK(if(tracker!=NULL)tracker[3]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Get the actual PSF values.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_power_spec, blocks_fft*MATRIX_OP_BLOCKSZ, MATRIX_OP_BLOCKSZ, 3,
			{sizeof(cl_mem), (void*)&psf[i]->cl_focus},
			{sizeof(cl_mem), (void*)&psf[i]->cl_psf},
			{sizeof(cl_int), (void*)&size_fft}
		);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "psf", psf[i]->cl_psf, phase_size, phase_size);
	}
	TRACK(if(tracker!=NULL)tracker[4]->pause())

	TRACK(if(tracker!=NULL)tracker[5]->resume())
	// Sum PSF values.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_reduce, REDUCTION_NBLOCKS*REDUCTION_BLOCKSZ, REDUCTION_BLOCKSZ, 4,
			{sizeof(cl_mem), (void*)&psf[i]->cl_psf},
			{sizeof(cl_mem), (void*)&psf[i]->cl_sum},
			{REDUCTION_BLOCKSZ * SIZEOF_WORD, NULL},
			{sizeof(cl_int), (void*)&size_fft}
		);
	}
	SYNC_QUEUE
	WORD scale[n_psfs];
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "sum-psf", psf[i]->cl_sum, 1, REDUCTION_NBLOCKS);
		FINAL_REDUCE(command_queue, psf[i]->cl_sum, scale[i]);
		CHECKSUM(WORD, command_queue, psf[i]->cl_psf, size_fft, scale[i]);
		scale[i] = 1 / (WORD)scale[i];
	}
	TRACK(if(tracker!=NULL)tracker[5]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Scale the pupils with the PSFs sums.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_multiply_complexarr, blocks_fft*MATRIX_OP_BLOCKSZ, MATRIX_OP_BLOCKSZ, 4,
			{sizeof(cl_mem), (void*)&psf[i]->cl_pupil},
			{sizeof(cl_mem), (void*)&psf[i]->cl_pupil},
			{sizeof(cl_float), (void*)&scale[i]},
			{sizeof(cl_int), (void*)&size_fft}
		);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG_COMPLEX(command_queue, "norm-pupil", psf[i]->cl_pupil, phase_size, phase_size);
	}
	TRACK(if(tracker!=NULL)tracker[4]->pause())

	TRACK(if(tracker!=NULL)tracker[3]->resume())
	// Calculate the FFT of the scaled pupils.
	for(i=0; i < n_psfs; i++){
		omp_set_lock(&mutex_fft);
		psf[i]->pupilToFocusFft();
		omp_unset_lock(&mutex_fft);
	}
	for(i=0; i < n_psfs; i++){
		DEBUG_COMPLEX(command_queue, "norm-focus", psf[i]->cl_focus, phase_size, phase_size);
	}
	TRACK(if(tracker!=NULL)tracker[3]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Get the actual PSF scaled values.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_power_spec, blocks_fft*MATRIX_OP_BLOCKSZ, MATRIX_OP_BLOCKSZ, 3,
			{sizeof(cl_mem), (void*)&psf[i]->cl_focus},
			{sizeof(cl_mem), (void*)&psf[i]->cl_psf},
			{sizeof(cl_int), (void*)&size_fft}
		);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "norm-psf", psf[i]->cl_psf, phase_size, phase_size);
	}

	// Extract the PSFs adjusted to the image width.
	for(i=0; i < n_psfs; i++){
		if(img_size < phase_size){
			// Extract the values according to the PSF_EXTRACT constant.
			CALL_KERNEL(command_queue, kernels->cl_resize_psf, blocks_img*MATRIX_OP_BLOCKSZ, MATRIX_OP_BLOCKSZ, 5,
				{sizeof(cl_mem), (void*)&psf[i]->cl_psf},
				{sizeof(cl_mem), (void*)&psf[i]->cl_psfe},
				{sizeof(cl_int), (void*)&phase_size},
				{sizeof(cl_int), (void*)&img_size},
				{sizeof(cl_int), (void*)&img_area}
			);
		}
		else{
			// Just copy.
			clMemcpyDeviceToDevice(command_queue, psf[i]->cl_psfe, psf[i]->cl_psf, img_area * SIZEOF_WORD, false);
		}
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "psfe", psf[i]->cl_psfe, img_size, img_size);
	}
	TRACK(if(tracker!=NULL)tracker[4]->pause())

	TRACK(if(tracker!=NULL)tracker[5]->resume())
	// Sum the extracted PSF values.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_reduce, REDUCTION_NBLOCKS*REDUCTION_BLOCKSZ, REDUCTION_BLOCKSZ, 4,
			{sizeof(cl_mem), (void*)&psf[i]->cl_psfe},
			{sizeof(cl_mem), (void*)&psf[i]->cl_sum},
			{REDUCTION_BLOCKSZ * SIZEOF_WORD, NULL},
			{sizeof(cl_int), (void*)&img_area}
		);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "sum-psfe",  psf[i]->cl_sum, 1, REDUCTION_NBLOCKS);
		FINAL_REDUCE(command_queue,  psf[i]->cl_sum, scale[i]);
		CHECKSUM(WORD, command_queue,  psf[i]->cl_psfe, img_area, scale[i]);
		scale[i] = 1.0 / scale[i];
	}
	TRACK(if(tracker!=NULL)tracker[5]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Scale the extracted PSFs with their own sums.
	for(i=0; i < n_psfs; i++){
		CALL_KERNEL(command_queue, kernels->cl_multiply_doublearr, blocks_img*MATRIX_OP_BLOCKSZ, MATRIX_OP_BLOCKSZ, 4,
			{sizeof(cl_mem), (void*)& psf[i]->cl_psfe},
			{sizeof(cl_mem), (void*)& psf[i]->cl_psfe},
			{sizeof(cl_float), (void*)&scale[i]},
			{sizeof(cl_int), (void*)&img_area}
		);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG(WORD, command_queue, "norm-psfe",  psf[i]->cl_psfe, img_size, img_size);
	}
	TRACK(if(tracker!=NULL)tracker[4]->pause())
}

/**
 * @brief Convolve the object with the PSFs calculated, generating several corrupted images.
 *
 * @param tracker The tracker.
 * @param n_psfs The number of PSFs that will generate the same number of convolved objects.
 */
void CPSO::convolveObj(TimeTracker **tracker, int n_psfs) {
	int max_threads = n_psfs * img_size;
	int i;

	TRACK(if(tracker!=NULL)tracker[3]->resume())
	// Move the real numbers to the FFT_TYPE format.
	for(i=0; i < n_psfs; i++){
		REAL2COMPLEX_GPU_VIA_GPU(command_queue, psf[i]->cl_psfe, psf[i]->cl_fft_psfe, img_area);
	}
	SYNC_QUEUE
	for(i=0; i < n_psfs; i++){
		DEBUG_COMPLEX(command_queue, "fft-psfe", psf[i]->cl_fft_psfe, img_size, img_size); //img_sizeh
	}

	// Calculate the FFT of the PSFs.
	for(i=0; i < n_psfs; i++){
		omp_set_lock(&mutex_fft);
		psf[i]->psfeFft();
		omp_unset_lock(&mutex_fft);
	}

	// Move the FFTs of the PSFs (calculated previously) back to the major array.
	for(i=0; i < n_psfs; i++){
		clMemcpyDeviceToDeviceOffset(command_queue, cl_fft_psfe, i * img_area * SIZEOF_FFTTYPE, psf[i]->cl_fft_psfe, 0, img_area * SIZEOF_FFTTYPE, false);
	}
	SYNC_QUEUE
	DEBUG_COMPLEX3D(command_queue, "fft-psfe", cl_fft_psfe, img_size, img_size, n_psfs); //img_sizeh
	TRACK(if(tracker!=NULL)tracker[3]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Convolve the object with the PSFs, generating 'n_psfs' convolved objects.
	CALL_KERNEL(command_queue, kernels->cl_multiply_fftw_complex_arrays, max_threads, img_size, 5,
		{sizeof(cl_mem), (void*)&cl_fft_psfe},
		{sizeof(cl_mem), (void*)&cl_fft_object},
		{sizeof(cl_mem), (void*)&cl_fft_conobj},
		{sizeof(cl_int), (void*)&img_size},
		{sizeof(cl_int), (void*)&max_threads}
	);
	SYNC_QUEUE
	DEBUG_COMPLEX3D(command_queue, "fft-conobj", cl_fft_conobj, img_size, img_size, n_psfs);
	TRACK(if(tracker!=NULL)tracker[4]->pause())
}

/**
 * @brief Calculate the cost of the convolved objects.
 *
 * @param tracker The tracker.
 * @param n_psfs The number of PSFs that were used on this search.
 */
void CPSO::calcCost(TimeTracker **tracker, int n_psfs) {
	int max_threads = n_psfs * img_size;

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Calculate the search cost as the difference between the convolved object and the image.
	// The result for every point is stored into cl_cost.
	WORD scale = img_area;
	CALL_KERNEL(command_queue, kernels->cl_calc_cost, max_threads, img_size, 7,
		{sizeof(cl_mem), (void*)&cl_fft_image},
		{sizeof(cl_mem), (void*)&cl_fft_conobj},
		{sizeof(cl_mem), (void*)&cl_cost},
		{sizeof(cl_mem), (void*)&cl_diffraction_mask},
		{sizeof(cl_int), (void*)&scale},
		{sizeof(cl_int), (void*)&img_size},
		{sizeof(cl_int), (void*)&max_threads}
	);
	SYNC_QUEUE
	DEBUG_3D(WORD, command_queue, "cost", cl_cost, img_size, img_size, n_psfs);
	TRACK(if(tracker!=NULL)tracker[4]->pause())

	TRACK(if(tracker!=NULL)tracker[5]->resume())
	// Sum cl_cost to obtain the final cost (one cost per convolved object).
	CALL_KERNEL2D(command_queue, kernels->cl_reduce, REDUCTION_NBLOCKS*REDUCTION_BLOCKSZ, n_psfs, REDUCTION_BLOCKSZ, 1, 4,
		{sizeof(cl_mem), (void*)&cl_cost},
		{sizeof(cl_mem), (void*)&cl_sum},
		{REDUCTION_BLOCKSZ * SIZEOF_WORD, NULL},
		{sizeof(cl_int), (void*)&img_area}
	);
	SYNC_QUEUE
	WORD chk_sum;
	DEBUG_2D(WORD, command_queue, "sum-cost",  cl_sum, n_psfs, REDUCTION_NBLOCKS);
	FINAL_REDUCTIONS(command_queue,  cl_sum, chk_sum, n_psfs);
	CHECKSUM(WORD, command_queue, cl_cost, img_area*n_psfs, chk_sum);//img_sizeh

	TRACK(if(tracker!=NULL)tracker[5]->pause())
}

FFT_TYPE CPSO::calcDifference(TimeTracker **tracker, WORD *img, WORD *img_diff){
	FFT_TYPE *fft_conobj = (FFT_TYPE*)buffer;
	for(int i = 0; i < img_area; i++) {fft_conobj[i].x = img[i]; fft_conobj[i].y = 0;}
	clMemcpyHostToDevice(command_queue, cl_fft_conobj, fft_conobj, img_area * SIZEOF_FFTTYPE);

	TRACK(if(tracker!=NULL)tracker[3]->resume())
	omp_set_lock(&mutex_fft);
	viennacl::inplace_fft(*vn_fft_conobj); // Inplace transform.
	omp_unset_lock(&mutex_fft);
	DEBUG_COMPLEX(command_queue, "fft-conobj", cl_fft_conobj, img_size, img_size);
	TRACK(if(tracker!=NULL)tracker[3]->pause())

	FFT_TYPE result = calcMismatch(tracker, cl_fft_conobj);
	clMemcpyDeviceToHost(command_queue, img_diff, cl_cost, img_area * SIZEOF_WORD);

	return result;
}

/**
 * @brief Calculate the differences between one convolved object and the image.
 *
 * The differences calculated are the cost (image - convolved object)
 * and the comparison (image != convolved object), returning 1 result per image point.
 *
 * @param tracker The tracker.
 * @param _cl_fft_conobj A pointer to the FFT of the convoved object
 *                    to calculate the difference from the FFT of the image.
 *
 * @return The mismatch as a {@link #FFT_TYPE}: {mismatch.x, mismatch.y} = {sum cost, sum comparison}.
 */
FFT_TYPE CPSO::calcMismatch(TimeTracker **tracker, cl_mem _cl_fft_conobj) {
	int max_threads = img_size;

	if(_cl_fft_conobj != cl_fft_conobj){
		clMemcpyDeviceToDevice(command_queue, cl_fft_conobj, _cl_fft_conobj, img_area * SIZEOF_FFTTYPE, true);
		DEBUG_COMPLEX(command_queue, "fft-conobj", cl_fft_conobj, img_size, img_size);
	}

	// Backtransform the first convolved object Only.
	TRACK(if(tracker!=NULL)tracker[3]->resume())
	omp_set_lock(&mutex_fft);
	viennacl::inplace_ifft(*vn_fft_conobj); // Inplace inverse transform.
	omp_unset_lock(&mutex_fft);
	DEBUG_COMPLEX(command_queue, "i-fft-conobj", cl_fft_conobj, img_size, img_size);
	TRACK(if(tracker!=NULL)tracker[3]->pause())

	TRACK(if(tracker!=NULL)tracker[4]->resume())
	// Calculate the mismatch and save the best convolved object.
	CALL_KERNEL(command_queue, kernels->cl_calc_mismatch, max_threads, img_size, 7,
		{sizeof(cl_mem), (void*)&cl_image},
		{sizeof(cl_mem), (void*)&cl_fft_conobj},
		{sizeof(cl_mem), (void*)&cl_best_conobj},
		{sizeof(cl_mem), (void*)&cl_cost},
		{sizeof(cl_mem), (void*)&cl_mismatch},
		{sizeof(cl_int), (void*)&img_size},
		{sizeof(cl_int), (void*)&max_threads}
	);
	SYNC_QUEUE
	DEBUG_3D(WORD, command_queue, "cost", cl_cost, img_size, img_size, 1);
	TRACK(if(tracker!=NULL)tracker[4]->pause())

	// Make the sums.
	FFT_TYPE mismatch;
	reduce_squares(tracker, 1, img_area, cl_cost, cl_sum, &mismatch.x);
	reduce_squares(tracker, 1, img_area, cl_mismatch, cl_sum, &mismatch.y);

	return mismatch;
}

/**
 * @brief This method makes the reductions for 'n_reductions' matrixes.
 *
 * Every summed matrix will return a WORD value inside the diff parameter,
 * thus this parameter must have allocated enough space for the results.
 *
 * @param tracker The tracker
 * @param n_reductions The number of matrixes to be reduced.
 * @param reduction_width The square matrix width.
 * @param square The matrix data.
 * @param sum The OpenCl pointer where the intermediate sum will be stored.
 * @param result The pointer where every final resulting will be stored.
 */
void CPSO::reduce_squares(TimeTracker **tracker, int n_reductions, int reduction_width, cl_mem square, cl_mem sum, WORD* result){
	TRACK(if(tracker!=NULL)tracker[5]->resume());
	CALL_KERNEL2D(command_queue, kernels->cl_reduce, REDUCTION_NBLOCKS*REDUCTION_BLOCKSZ, n_reductions, REDUCTION_BLOCKSZ, 1, 4,
		{sizeof(cl_mem), (void*)&square},
		{sizeof(cl_mem), (void*)&sum},
		{REDUCTION_BLOCKSZ * SIZEOF_WORD, NULL},
		{sizeof(cl_int), (void*)&reduction_width}
	);
	SYNC_QUEUE

	for(int i=0; i < n_reductions; i++){
		DEBUG(WORD, command_queue, "sum-square", sum, 1, REDUCTION_NBLOCKS);
		FINAL_REDUCE(command_queue, sum, result[i]);
		CHECKSUM(WORD, command_queue, square, reduction_width, result[i]);
	}
	TRACK(if(tracker!=NULL)tracker[5]->pause());
}

/**
 * @brief Validate an external PSF.
 *
 * The validation is made by:
 * <ul>
 * <li> Convolutioning the external PSF with the internal object;</li>
 * <li> Calculating the mismatch between the convolutioned object
 * and the internal image.
 * </ul>
 *
 * @param psf_data The external PSF.
 *
 * @return The mismatch calculated.
 */
FFT_TYPE CPSO::validatePsf(double *psf_data){
	// Move the PSF data to the PSF instance.
	PSF psf(context, command_queue, img_size, phase_size, 1);
	FFT_TYPE *_psf = (FFT_TYPE*)buffer;
	for(int i = 0; i < img_area; i++) {_psf[i].x = psf_data[i]; _psf[i].y = 0;}
	clMemcpyHostToDevice(command_queue, psf.cl_fft_psfe, _psf, img_area * SIZEOF_FFTTYPE);
	DEBUG_COMPLEX(command_queue, "psf-validate", psf.cl_fft_psfe, img_size, img_size);
	// Calculate the FFT of the PSF.
	psf.psfeFft();
	DEBUG_COMPLEX(command_queue, "fft-psf-validate", psf.cl_fft_psfe, img_size, img_size);

	// Convolve the internal object with the external PSF.
	CALL_KERNEL(command_queue, kernels->cl_multiply_fftw_complex_arrays, img_size, img_size, 5,
		{sizeof(cl_mem), (void*)&psf.cl_fft_psfe},
		{sizeof(cl_mem), (void*)&cl_fft_object},
		{sizeof(cl_mem), (void*)&cl_fft_conobj},
		{sizeof(cl_int), (void*)&img_size},
		{sizeof(cl_int), (void*)&img_size}
	);
	SYNC_QUEUE
	DEBUG_COMPLEX(command_queue, "fft-conobj", cl_fft_conobj, img_size, img_size);

	// Return the mismatch between the internal object
	// convolved with the external PSF and the internal image.
	return calcMismatch(NULL, cl_fft_conobj);
}

/**
 * @brief Get a random coefficient within a specified range.
 *
 * @param range The range to generate the random value.
 *
 * @return The random coefficient.
 */
WORD CPSO::getRandCoef(double range) {
	WORD randn = ((rand() / (WORD)RAND_MAX) - 0.5) * 2 * range;
	if (randn > range) {
		randn = -2.0 * range + randn;
	}
	else if (randn < -range) {
		randn = 2 * range + randn;
	}
	HALT(randn > range || randn < -range);

	return randn;
}

/**
 * @brief Set the Zernike coefficients for the startup.
 *
 * It must be called before the search has begun.
 *
 * @param coefs The Zernike coefficients for the startup.
 */
void CPSO::setStartupCoefs(WORD *coefs){
	has_startup_coefs = true;
	for(int i=0; i < n_zernikes; i++) startup_coefs[i] = coefs[i];
}

/**
 * @brief Save the results of the first Particle / PSF as the best value found so far.
 */
void CPSO::saveFirstResult(){
	// Save the convolved objects's FFT before inverting it.
	clMemcpyDeviceToDevice(command_queue, cl_best_fft_conobj, cl_fft_conobj, img_area * SIZEOF_FFTTYPE, false);

	// Invert the FFT of the first the convolved object.
	omp_set_lock(&mutex_fft);
	viennacl::inplace_ifft(*vn_fft_conobj); // Inplace transform.
	omp_unset_lock(&mutex_fft);

	// Extract the real part of the fist convolved object and save it.
	CALL_KERNEL(command_queue, kernels->cl_real, img_area, img_size, 3,
		{sizeof(cl_mem), (void*)&cl_fft_conobj},
		{sizeof(cl_mem), (void*)&cl_best_conobj}, // Save it here as the best convolved object.
		{sizeof(cl_int), (void*)&img_area}
	);
	SYNC_QUEUE

	// Save the others values of the first PSF as the best values.
	clMemcpyDeviceToDevice(command_queue, cl_best_phase, cl_phase, size_fft * SIZEOF_WORD, false);
	clMemcpyDeviceToDevice(command_queue, cl_best_coefs, cl_coefs, n_zernikes * SIZEOF_WORD, false);
	clMemcpyDeviceToDevice(command_queue, cl_best_psf, psf[0]->cl_psf, img_area * SIZEOF_WORD, false);
	clMemcpyDeviceToDevice(command_queue, cl_best_psfe, psf[0]->cl_psfe, img_area * SIZEOF_WORD, false);
	clMemcpyDeviceToDevice(command_queue, cl_best_fft_psfe, psf[0]->cl_fft_psfe, img_area * SIZEOF_FFTTYPE, false);
}

/**
 * @brief Get the minimum cost obtained from all the swarms available.
 *
 * @return The minimum cost obtained during the search.
 */
WORD CPSO::getMinCost(){
	WORD ret = WORD_MAX;
	WORD *cost = (WORD*)buffer;
	clMemcpyDeviceToHost(command_queue, cost, cl_gbest_value, n_swarms * SIZEOF_WORD);
	for(int i=0; i < n_swarms; i++){
		if(cost[i] < ret){
			ret = cost[i];
		}
	}

	return ret;
}

/**
 * @brief Get the position between all the particles of the best PSF found.
 *
 * @return The particle position of the best PSF found.
 */
int CPSO::getBestPsfPos(){
	int n_psfs = n_swarms * n_particles;
	WORD val = WORD_MAX, pos;
	WORD *cost = (WORD*)buffer;
	clMemcpyDeviceToHost(command_queue, cost, cl_pbest_value, n_psfs * SIZEOF_WORD);
	for(int i=0; i < n_psfs; i++){
		if(cost[i] < val){
			val = cost[i];
			pos = i;
		}
	}
	return pos;
}

/**
 * @brief Get the best coefficients.
 *
 * @param _coefs The WORD pointer where to copy the coefficients.
 */
void CPSO::getBestCoefs(WORD* _coefs){
	clMemcpyDeviceToHost(command_queue, _coefs, cl_best_coefs, n_zernikes * SIZEOF_WORD);
}

/**
 * @brief Returns the best cost found by the search.
 * Should be called Only after the {@link run} method has finished.
 */
WORD CPSO::getGBestCost(){
	return gbest_cost;
}

/**
 * Not in use.
 */
void CPSO::commitBestValues() {
//	clMemcpyDeviceToDevice(cl_best_psf, cl_psf, size_fft * SIZEOF_WORD);
//	clMemcpyDeviceToDevice(cl_best_psfe, cl_psfe, img_area * SIZEOF_WORD);
//	clMemcpyDeviceToDevice(cl_best_conobj, cl_conobj, img_area * SIZEOF_WORD);
}

/**
 * @brief Get the best phase.
 *
 * @param phase The Host pointer where to copy the best phase.
 */
void CPSO::getBestPhase(WORD *phase) {
	clMemcpyDeviceToHost(command_queue, phase, cl_best_phase, size_fft * SIZEOF_WORD);
}

/**
 * @brief Get the best PSF.
 *
 * @param psf The Host pointer where to copy the best PSF.
 */
void CPSO::getBestPsf(WORD *psf) {
	clMemcpyDeviceToHost(command_queue, psf, cl_best_psf, size_fft * SIZEOF_WORD);
}

/**
 * @brief Get the best extracted (final) PSF.
 *
 * @param psfe The Host pointer where to copy the best final PSF.
 */
void CPSO::getBestPsfe(WORD *psfe) {
	clMemcpyDeviceToHost(command_queue, psfe, cl_best_psfe, img_area * SIZEOF_WORD);
}

/**
 * @brief Get the best convolved object.
 *
 * @param conobj The Host pointer where to copy the best convolved object.
 */
void CPSO::getBestConvolvedObject(WORD *conobj) {
	clMemcpyDeviceToHost(command_queue, conobj, cl_best_conobj, img_area * SIZEOF_WORD);
}

/**
 * @brief Get the FFT of the best convolved object.
 *
 * @param conobj_fft The Host pointer where to copy the FFT of the best convolved object.
 */
void CPSO::getBestConvolvedObjectFFT(FFT_TYPE *conobj_fft) {
	clMemcpyDeviceToHost(command_queue, conobj_fft, cl_best_fft_conobj, img_area * SIZEOF_FFTTYPE);
}

/**
 * @brief Get the FFT of the best extracted PSF.
 *
 * @param psfe_fft The Host pointer where to copy the FFT of the best extracted PSF.
 */
void CPSO::getBestPsfeFFT(FFT_TYPE *psfe_fft) {
	clMemcpyDeviceToHost(command_queue, psfe_fft, cl_best_fft_psfe, img_area * SIZEOF_FFTTYPE);
}

/**
 * @brief Get the FFT of the object.
 *
 * @param obj_fft The Host pointer where to copy the FFT of the object.
 */
void CPSO::getObjectFFT(FFT_TYPE *obj_fft) {
	clMemcpyDeviceToHost(command_queue, obj_fft, cl_fft_object, img_area * SIZEOF_FFTTYPE);
}

/**
 * @brief Get the FFT of the image.
 *
 * @param img_fft The Host pointer where to copy the FFT of the image.
 */
void CPSO::getImageFFT(FFT_TYPE *img_fft) {
	clMemcpyDeviceToHost(command_queue, img_fft, cl_fft_image, img_area * SIZEOF_FFTTYPE);
}

/**
 * @brief Set the search parameter 'w' for all the particles.
 *
 * @param _w The search parameter w.
 */
void CPSO::setW(WORD _w){
	replicateValue(_w, n_particles * n_swarms, cl_w);
}

/**
 * @brief Set the search parameter 'c1' for all the particles.
 *
 * @param _c1 The search parameter c1.
 */
void CPSO::setC1(WORD _c1){
	replicateValue(_c1, n_particles * n_swarms, cl_c1);
}

/**
 * @brief Set the search parameter 'c2' for all the particles.
 *
 * @param _c2 The search parameter c2.
 */
void CPSO::setC2(WORD _c2){
	replicateValue(_c2, n_particles * n_swarms, cl_c2);
}

/**
 * @brief Set the exact original PSF that caused the corruption over the image.
 *
 * @param original_psf the original PSF.
 */
void CPSO::setOriginalPsf(WORD *original_psf){
	FFT_TYPE *arr = (FFT_TYPE*)buffer;
	for(int i=0; i < img_area; i++) arr[i].x = original_psf[i];
	clMemcpyHostToDevice(command_queue, cl_fft_original_psf, arr, img_area * SIZEOF_FFTTYPE);

	// Invert the FFT of the first the original PSF.
	omp_set_lock(&mutex_fft);
	viennacl::inplace_fft(*vn_fft_original_psf); // Inplace transform.
	omp_unset_lock(&mutex_fft);

	clMemcpyDeviceToHost(command_queue, original_psf_fft, cl_fft_original_psf, img_area * SIZEOF_FFTTYPE);
	has_psf_original = true;
}

/**
 * @brief Calculate the difference from the Original PSF to the currently calculated PSF.
 *
 * @return the difference between the PSFs.
 */
double CPSO::calcPsfDifferences(){
	if(!has_psf_original){
		return -1;
	}

	FFT_TYPE *arr = (FFT_TYPE*)buffer;
	clMemcpyDeviceToHost(command_queue, arr, psf[getBestPsfPos()]->cl_fft_psfe, img_area * SIZEOF_FFTTYPE);
	WORD x, y;
	double diff = 0;
	for(int i=0; i < img_area; i++){
		x = original_psf_fft[i].x - arr[i].x;
		y = original_psf_fft[i].y - arr[i].y;
		diff += (diffraction_mask[i]) * (sqrt(x * x + y * y) / img_area);
	}
	return diff;
}

/**
 * @brief Replicate one value 'sz' times into a OpenCl pointer.
 *
 * @param value The value to be replicated
 * @param sz The number of times the values will be replicated.
 * @param cl_ref The OpenCl pointer where to replicate the value.
 */
void CPSO::replicateValue(WORD value, int sz, cl_mem cl_ref){
	WORD *arr = (WORD*)buffer;
	for(int i=0; i < sz; i++) arr[i] = value;
	clMemcpyHostToDevice(command_queue, cl_ref, arr, sz * SIZEOF_WORD);
}

/**
 * @brief Lock this CPSO instance.
 */
void CPSO::lock(){
	in_use = true;
}

/**
 * @brief Release this CPSO instance.
 */
void CPSO::release() {
	in_use = false;
}

/**
 * @brief Inform if this CPSO instance is in use.
 */
bool CPSO::isInUse() {
	return in_use;
}

/**
 * Force the values to be copied from host to device as floats (not in use).
 */
void CPSO::copyToDeviceAsFloat(cl_command_queue command_queue, cl_mem dest, double *values, int size) {
	float *floatValue = (float*)buffer;
	for (int i = 0; i < size; i++) {
		floatValue[i] = values[i];
	}
	clMemcpyHostToDevice(command_queue, dest, floatValue, size * sizeof(float));
}

/**
 * Force the values to be copied from device to host as doubles (not in use).
 */
void CPSO::copyToHostAsDouble(cl_command_queue command_queue, cl_mem src, double *values, int size) {
	float *floatValue = (float*)buffer;
	clMemcpyDeviceToHost(command_queue, floatValue, src, size * sizeof(float));
	for (int i = 0; i < size; i++) {
		values[i] = floatValue[i];
	}
}

/**
 * @brief Calculate the mean of a set of values.
 *
 * @param values The values to calculate the mean.
 * @param size The array size.
 * @return The calculated mean.
 */
double CPSO::calc_mean(double *values, int size){
	double mean = 0;
	for(int i=0; i < size; i++){
		mean += values[i];
	}
	return mean / size;
}

/**
 * @brief Calculate the variance of a set of values.
 *
 * @param values The values to calculate the variance.
 * @param size The array size.
 * @return The calculated variance.
 */
double CPSO::calc_variance(double *values, int size){
	if(size == 1) return 0;

	double mean = calc_mean(values, size);
	double var = 0;
	for(int i=0; i < size; i++){
		var += (values[i] - mean) * (values[i] - mean);
	}
	return var / (size - 1);
}

/**
 * @brief Calculate the standard deviation of a set of values.
 *
 * @param values The values to calculate the standard deviation.
 * @param size The array size.
 * @return The calculated standard deviation.
 */
double CPSO::calc_stddev(double *values, int size){
	return sqrt(calc_variance(values, size));
}

/**
 * @brief Generate random coefficients in a specific range.
 *
 * @param coefs A pointer where the the coefficients will be generated.
 * @param n_zernikes The number of coefficients to be generated.
 * @param range The coefficients range.
 */
void CPSO::generateRandomCoefs(WORD *coefs, int n_zernikes, double range){
	// The number of Zernike terms to be distorted is randomized
	// to simulate a real environment.
	int n_randcoefs = (rand() / (WORD)RAND_MAX) * n_zernikes;
	if(n_randcoefs == 0) n_randcoefs = 1;
	memset(coefs, 0, n_zernikes * SIZEOF_WORD);
	WORD val;
	for(int i=0; i < n_randcoefs; i++){
		do{
			val = randNormalDistribution(0, range / 4);
		}while(val > range);
		// The range used for the random coefficients is randomized to avoid
		// too drastic image distortions and to simulate a real environment.
		coefs[(int)((rand() / (WORD)RAND_MAX) * n_zernikes)] = val;
	}
}

/**
 * @brief Generate random coefficients in a normal distribution.
 *
 * @param coefs A pointer where the the coefficients will be generated.
 * @param n_zernikes The number of coefficients to be generated.
 * @param range The coefficients range.
 */
void CPSO::generateNormalDistrRandomCoefs(WORD *coefs, int n_zernikes, double range){
	WORD val;
	for(int i=0; i < n_zernikes; i++){
		do{
			val = randNormalDistribution(0, range / 4);
		}while(val > range);
		coefs[i] = val;
	}
}

/**
 * @brief Generate a double value in a normal distribution.
 *
 * @param mean The mean of the normal distribution.
 * @param std_dev The standard deviation of the normal distribution.
 * @return The normal distribution number generated.
 */
double CPSO::randNormalDistribution(double mean, double std_dev){
	return (mean + (rand()%2 ? -1.0 : 1.0) *
			std_dev * pow(-log(0.99999*((double)rand()/RAND_MAX)), 0.5)
	);
}
