
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * cpso.hpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */


#ifndef _CPSO_HPP_
#define _CPSO_HPP_

#include <CL/cl.h>
#include <CL/opencl.h>
#include <CL/cl_platform.h>
#include <CL/cl_ext.h>
#include <vector_types.h>
#include <string.h>
#include <cmath>
#include <stdio.h>
#include <stdlib.h>
#include <complex>
#include <iostream>
#include <map>
#include "viennacl/vector.hpp"
#include "viennacl/fft.hpp"
#include "util.hpp"
#include "cl_factory.hpp"
#include "psf.hpp"
#include "TimeTracker.hpp"
#include "Profiling.hpp"

using namespace std;

/**
 * @brief This class represents the Search method Cooperative Particle Swarm Optimization (CPSO).
 *
 * All processing related to this search method is included into this class.
 */
class CPSO {
private:
	static map<cl_context, cl_mem*> static_references;
	static omp_lock_t mutex, mutex_fft;
	static int seq;
	static bool lock_initialized;
	static int *diffraction_mask;

	/**
	 * The Unique ID for the CPSO instance.
	 */
	int UID;

	// OpenCl section.
	//----------------
	// Variables.
	cl_mem cl_coefs, cl_sum, cl_conobj, cl_cost, cl_mismatch;
	cl_mem cl_speed, cl_pbest, cl_gbest, cl_pbest_value, cl_gbest_value, cl_reset_search, cl_swarm_dim,
		cl_rand_ctx, cl_w, cl_c1, cl_c2, cl_best_coefs;
	cl_mem cl_fft_psfe, cl_fft_conobj, cl_pupil, cl_phase, cl_focus;
	cl_mem cl_object, cl_image, cl_best_phase, cl_best_psf, cl_best_psfe, cl_best_conobj, cl_best_fft_conobj,
		cl_best_fft_psfe, cl_fft_object, cl_fft_image, cl_debug_info, cl_fft_original_psf;

	// Constant values.
	cl_mem cl_phase_mask, cl_zernikes, cl_diffraction_mask;

	// FFT.
	viennacl::vector<WORD> *vn_object, *vn_image, *vn_fft_object, *vn_fft_image, *vn_fft_conobj,
		*vn_fft_original_psf;

	// Environment control.
	cl_int status;
	cl_context context;
	cl_command_queue command_queue;
	cl_program program;
	cl_device_id device;
	clQueue *queue;
	kernel_set *kernels;

	// Host section.
	//--------------
	bool in_use, has_startup_coefs, has_psf_original;
	int threads, phase_size, img_size, img_sizeh, img_area, img_areah, size_fft, z_size, n_zernikes,
		n_particles, n_swarms, n_psf_evals, max_evals;
	int convergence_stable_cycle;
	WORD w, psf_range, reset_at, gbest_cost;
	WORD *startup_coefs;
	FFT_TYPE *original_psf_fft; // To be used during Validation process only, for comparison with the calculated one.

	PSF **psf;

	char *buffer;

	// Methods.
	//---------
	cl_context createContext();

	cl_command_queue createCommandQueue(cl_context context, cl_device_id *device);

	cl_program createProgram(cl_context context, cl_device_id device);

	void initialize_cl();

	void allocate_data(int _phase_size, int image_size, int _n_zernikes, int _psf_range,
			WORD _w, WORD c1, WORD c2, WORD _reset_at, int _n_particles, int _n_swarms);

	void startup(TimeTracker **trackers);

	static void startup_locks();

	static void destroy_locks();

	void setStableCycle(int cycle) { convergence_stable_cycle = cycle; }

	FFT_TYPE calcMismatch(TimeTracker **tracker, cl_mem _cl_fft_conobj);

public:
	CPSO(double* _zernikes, int *_phase_mask, int *_diffraction_mask, int _phase_size, int _image_size, int _n_zernikes,
			int _psf_range, WORD _w, WORD c1, WORD c2, WORD _reset_at, int _n_particles, int _n_swarms);

	~CPSO();

	void set_images(double *object, double *image);

	static void store_static_data(cl_context _context, int _n_zernikes, int _phase_size, int _image_size, double* zernikes,
			int *phase_mask, int *_diffraction_mask);

	static void clear_static_data();

	void initialize_gpu();

	void finalize_cl();

	int getStableCycle(){ return convergence_stable_cycle; }

	PSF** getPsfs(){ return psf; }

	WORD getMinCost();

	int getBestPsfPos();

	void getBestCoefs(WORD* _coefs);

	void getBestPsf(WORD *psf);

	void getBestPsfe(WORD *psfe);

	void getBestPhase(WORD *phase);

	void getBestConvolvedObject(WORD *conobj);

	void getBestConvolvedObjectFFT(FFT_TYPE *conobj_fft);

	void getBestPsfeFFT(FFT_TYPE *psfe_fft);

	void getObjectFFT(FFT_TYPE *obj_fft);

	void getImageFFT(FFT_TYPE *img_fft);

	cl_context getContext(){ return context; }

	cl_command_queue getCommandQueue(){ return command_queue; }

	int getPhaseSize(){ return phase_size; }

	int getImageSize(){ return img_size; }

	int getFftSize(){ return size_fft; }

	int getNZernikes() { return n_zernikes; }

	int getPsfRange() { return psf_range; }

	int getNPsfEvals() { return n_psf_evals; }

	void setMaxEvaluations(int _max_evals) { max_evals = _max_evals; }

	void replicateValue(WORD value, int sz, cl_mem cl_ref);

	void setW(WORD _w);

	void setC1(WORD _c1);

	void setC2(WORD _c2);

	void setOriginalPsf(WORD *original_psf);

	double calcPsfDifferences();

	cl_mem getFftObject(){ return cl_fft_object; }

	cl_mem getFftImage() {return cl_fft_image; }

	void run(TimeTracker **trackers, int n_cycles);

	void runPsf(TimeTracker **trackers);

	void runPsf(TimeTracker **trackers, int n_psfs);

	void generatePhase(TimeTracker **tracker, int n_psfs);

	void makePsf(TimeTracker **tracker, int n_psfs);

	void convolveObj(TimeTracker **tracker, int n_psfs);

	void calcCost(TimeTracker **tracker, int n_psfs);

	FFT_TYPE calcDifference(TimeTracker **tracker, WORD *img, WORD *img_diff);

	void runCPSO(TimeTracker **tracker, int n_cycles);

	void reduce_squares(TimeTracker **tracker, int n_reductions, int reduction_width, cl_mem square, cl_mem sum, WORD* result);

	FFT_TYPE validatePsf(double *psf);

	void commitBestValues();

	void testFft(int sz, FFT_TYPE *values);

	void testFft(int sz);

	void testReduction(int sz);

	void testReduction(int sz, WORD *values);

	static WORD getRandCoef(double range);

	void copyToDeviceAsFloat(cl_command_queue command_queue, cl_mem dest, double *values, int size);

	void copyToHostAsDouble(cl_command_queue command_queue, cl_mem src, double *values, int size);

	const char* last_cuda_error();

	void lock();

	void release();

	bool isInUse();

	WORD getGBestCost();

	void setStartupCoefs(WORD *coefs);

	void saveFirstResult();

	static double calc_mean(double *image, int img_area);

	static double calc_variance(double *image, int img_area);

	static double calc_stddev(double *image, int img_area);

	static void generateRandomCoefs(WORD *coefs, int n_zernikes, double range);

	static void generateNormalDistrRandomCoefs(WORD *coefs, int n_zernikes, double range);

	static double randNormalDistribution(double mean, double std_dev);

	/**
	 * @brief Invert the PSF to put the bright spot in the center of the image.
	 *
	 * The inverted PSF is stored in the psf_inv pointer, which must be previously allocated.
	 *
	 * @param psf The PSF to be inverted.
	 * @param psf_inv The PSF inverted.
	 * @param psf_size The width of the PSF.
	 */
	template<class T>
	static void invertPsf(T *psf, T *psf_inv, int psf_size){
		int half = psf_size / 2;
		/* Original PSF
		 *    _________
		 *    | A | B |
		 *    |-------|
		 *    | C | D |
		 *    =========
		 */
		for(int c, l=0; l < half; l++){
			for(c=0; c < half; c++){
				/* PSF Inverted.
				 *    _________
				 *    | D | C |
				 *    |-------|
				 *    | B | A |
				 *    =========
				 */
				psf_inv[((l + half) * psf_size) + half + c] = psf[(l * psf_size) + c]; // A -> D
				psf_inv[((l + half) * psf_size) + c] = psf[(l * psf_size) + half + c]; // B -> C
				psf_inv[(l * psf_size) + half + c] = psf[((l + half) * psf_size) + c]; // C -> C
				psf_inv[(l * psf_size) + c] = psf[((l + half) * psf_size) + half + c]; // D  -> B
			}
		}
	}
};

#endif /* _CPSO_HPP_ */
