
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * kippe.cpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "kippef.hpp"

void show_usage();

/**
 * @brief The main method.
 */
int main(int argc, char *argv[])
{
	// Initialize the parameters.
	char *objectname = NULL;
	char *imgpath = NULL;
	char *outpath = NULL;
	int image_size = -1;
	int n_zernikes = -1;
	int fft_size = -1;
	int n_swarms = -1;
	int n_particles = -1;
	int n_cycles = -1;
	float w = -1, c1 = -1, c2 = -1, reset = -1;
	int n_validations = -1, rank_size = -1;
	int n_images_gen = -1;
	int poisson = 0;
	int parallel = 1;
	bool PSF_validation = false;
	bool gen_images = false;
	bool statistics = false;
	bool rms_contrast = false;
	bool sum = false;
	vector<int> *devices = NULL;
	int prange = 2;

	// Read the arguments passed on command line.
	for(int i=1; i < argc; i++){
		if(strcmp("-h", argv[i]) == 0){
			show_usage();
			return 0;
		}
		else if(strcmp("-testgpu", argv[i]) == 0){
			test_gpu();
			return 0;
		}
		else if(strcmp("-listgpu", argv[i]) == 0){
			list_gpu();
			return 0;
		}
		else if(strcmp("-rms_contrast", argv[i]) == 0){
			rms_contrast = true;
		}
		else if(strcmp("-sum", argv[i]) == 0){
			sum = true;
		}
		else if(strcmp("-obj", argv[i]) == 0){
			objectname = argv[++i];
		}
		else if(strcmp("-img", argv[i]) == 0){
			imgpath = argv[++i];
		}
		else if(strcmp("-out", argv[i]) == 0){
			outpath = argv[++i];
		}
		else if(strcmp("-fixedrand", argv[i]) == 0){
			srand(RAND_SEED);
		}
		else if(strcmp("-isz", argv[i]) == 0){
			image_size = atoi(argv[++i]);
		}
		else if(strcmp("-fsz", argv[i]) == 0){
			fft_size = atoi(argv[++i]);
		}
		else if(strcmp("-z", argv[i]) == 0){
			n_zernikes = atoi(argv[++i]);
		}
		else if(strcmp("-r", argv[i]) == 0){
			prange = atoi(argv[++i]);
		}
		else if(strcmp("-s", argv[i]) == 0){
			n_swarms = atoi(argv[++i]);
		}
		else if(strcmp("-p", argv[i]) == 0){
			n_particles = atoi(argv[++i]);
		}
		else if(strcmp("-c", argv[i]) == 0){
			n_cycles = atoi(argv[++i]);
		}
		else if(strcmp("-w", argv[i]) == 0){
			w = atof(argv[++i]);
		}
		else if(strcmp("-c1", argv[i]) == 0){
			c1 = atof(argv[++i]);
		}
		else if(strcmp("-c2", argv[i]) == 0){
			c2 = atof(argv[++i]);
		}
		else if(strcmp("-reset", argv[i]) == 0){
			reset = atof(argv[++i]);
		}
		else if(strcmp("-nval", argv[i]) == 0){
			n_validations = atoi(argv[++i]);
		}
		else if(strcmp("-ranksz", argv[i]) == 0){
			rank_size = atoi(argv[++i]);
		}
		else if(strcmp("-noise", argv[i]) == 0){
			poisson = atoi(argv[++i]);
		}
		else if(strcmp("-gen_imgs", argv[i]) == 0){
			gen_images = true;
			n_images_gen = atoi(argv[++i]);
		}
		else if(strcmp("-statistics", argv[i]) == 0){
			statistics = true;
		}
		else if(strcmp("-devices", argv[i]) == 0){
			char *ptr = strtok (argv[++i], ",");
			devices = new vector<int>;
			int device;
			while(ptr != NULL){
				if ((device = atoi(ptr)) == 0){
					delete devices;
					printf("Invalid Device ID!\n");
					show_usage();
					exit(1);
				}
				devices->push_back(device);
				ptr = strtok (NULL, ",");
			}
		}
		else if(strcmp("-validation", argv[i]) == 0){
			PSF_validation = true;
		}
		else if(strcmp("-parallel", argv[i]) == 0){
			parallel = atoi(argv[++i]);
		}
	}
	// Verify if any parameter is missing.
	if(objectname == NULL || image_size <= 0
			|| (!rms_contrast && !sum && (fft_size <= 0 || n_zernikes <= 0 || poisson < 0 || poisson > 700 || parallel < 1
					|| (!PSF_validation  && !gen_images  && (imgpath == NULL || outpath == NULL || n_particles <= 0 || n_swarms <= 0 || n_cycles <= 0 || w <= 0 || c1 <= 0  || c2 <= 0 || reset < 0))
					|| (PSF_validation && !gen_images && (n_validations <= 0 || rank_size <= 0))
					|| (!PSF_validation && gen_images && n_images_gen <= 0)
					|| (PSF_validation && gen_images)))){
		cout << "Error on parameters." << endl;
		show_usage();
		exit(1);
	}

	printf("\n\n\n   ------------------------------------------------\n");
	printf("   | PSF Estimation Based on CPSO Optimization for GPUs. |\n");
	printf("   ------------------------------------------------\n\n");

	int img_area = image_size * image_size;
	if(rms_contrast || sum){
		FitsManager::startup(fitsSize(objectname));
		FitsManager *fitsManager = new FitsManager(1, image_size, image_size, NULL, objectname, NULL);
		if(rms_contrast){
			int n_images = fitsManager->size();
			fits_result *result;
			for(int i=0; i < n_images; i++){
				if((result = fitsManager->nextImage()) == NULL){
					continue;
				}
				printf("[%i] RMS contrast = %f\n", i+1, rmsContrast(result->image, image_size));
			}
		}
		else{
			double sum[img_area];
			sum_3dfits_file(fitsManager, sum, img_area);
			ostringstream oss;
			oss << objectname << "-3dsum.fits";
			writeFitsFile((char*)oss.str().c_str(), sum, image_size, image_size, 1);
			printf("The sum was saved in : %s\n", (char*)oss.str().c_str());
		}
		delete fitsManager;
		return 0;
	}

	TimeTracker tracker("Main");
	tracker.start();

	// Set the following parameters based on the telescope characteristics.
	//---------------------------------------------------------------------
	double wavelength = 396E-9;
	double image_scale = 0.09;
	double tel_dia = 0.7;
	//---------------------------------------------------------------------

	int size_fft = fft_size * fft_size;
	double cut_freq = cmp_freq_cut(tel_dia, image_scale, wavelength, image_size);
	int size_phases = cmp_size_of_phase(tel_dia, image_scale, wavelength, fft_size) / 2;

	int *phase_mask = new int[size_fft];
	int* diff_mask = new int[img_area];
	double *object = new double[img_area];
	double *radial_values = new double[size_fft];
	double *theta_values = new double[size_fft];
	double *zernikes = new double[n_zernikes * size_fft];

	// compute radial values
	gen_radial_values(radial_values, size_phases, fft_size);

	// compute angular_values
	gen_angular_values(theta_values, fft_size);

	// generate mask
	gen_circ_mask(phase_mask, size_phases, fft_size / 2, fft_size / 2, fft_size);

	// compute zernikes
	gen_zernike_values(zernikes, radial_values, theta_values, fft_size, n_zernikes);
	norm_zernikes(zernikes, fft_size, n_zernikes);

	// create diffraction limit mask
	gen_freq_mask(diff_mask, cut_freq, image_size, image_size);

	if(PSF_validation || gen_images){
		// Read the Object.
		read_fits_file((char*)objectname, object, image_size, image_size);
		if(PSF_validation){
			// The validation of this method cannot have the Hamming window applied over the object and the images,
			// since it will be using the Object to create random artificial images as base for the PSF estimation method.
			// Besides, the use of a window is optional and will interfere with the validation of this estimation method.
			validate_psfcode(object, zernikes, diff_mask, n_zernikes, image_size, fft_size, prange, phase_mask,
					(char*)objectname, n_validations, rank_size, devices, poisson, parallel);
		}
		else{
			generate_random_images(object, zernikes, diff_mask, n_zernikes, image_size, fft_size, prange, phase_mask,
					(char*)objectname, n_images_gen, devices, poisson, parallel);
		}
	}
	else{
		// Set the buffer size to 2 time the number of threads.
		FitsManager::startup(parallel * 2);
		FitsManager *fitsManager = new FitsManager(n_zernikes, image_size, fft_size, objectname, imgpath, outpath);

		int n_images = fitsManager->size();
		fits_result *result;

		// Calculate the Hamming window.
		double *hamming = new double[img_area];
		frac_hamming(hamming, 0.3, image_size, image_size);

		// Apply the Hamming window over the Object.
		apply_window(fitsManager->getObject(), hamming, img_area);

		/* CPSO Parameters:
		 * - w: Never accelerating too much, since the search space is small. Too much speed reduction is also not good.
		 * - c1: Always more important than c2. A low value affects the convergence (c1<1) and a high value makes it stuck for long periods (c1>1.5).
		 * - c2: A good value would be c2<=c1, with c2~=c1/2 as usually the best choice.
		 * - Particle re-initialization factor: usually something between 0.05 and 0.25.
		 * - No.Particles: More particles is better (eg. 128 or 256), because it means more random starting points.
		 *                 However, on GPUs it seems to have better results when equals to the warp size (eg., 32).
		 * - No.Swarms: 3 swarms has returned much better values and has better convergence behavior than 2 swarms,
		 *              but for 1 swarm or n_swarms>=4 the convergence is often being stuck for much longer periods.
		 */
		PsfEstimator::startup(parallel, zernikes, phase_mask, diff_mask, fft_size,  image_size, n_zernikes, prange, w, c1, c2, reset, n_particles, n_swarms, devices);

		double all_costs[n_images], all_times[n_images], all_stable_convergence[n_images];

		omp_set_num_threads(parallel);
#pragma omp parallel for shared(fitsManager, object, zernikes, diff_mask, phase_mask, hamming, all_costs, all_times, all_stable_convergence) \
			firstprivate(n_zernikes, image_size, fft_size, prange, n_images, img_area, size_fft, n_cycles, statistics) \
			private(result) default(none)
		for (int i=0; i < n_images; i++) {
			if((result = fitsManager->nextImage()) == NULL){
				continue;
			}

			TimeTracker tracker_estimation("single-estimation");
			if(statistics) tracker_estimation.start();

			// Apply the Hamming window over the Image.
			apply_window(result->image, hamming, img_area);

			// Load the Object and the Image to the CPSO search mechanism.
			CPSO *cpso = PsfEstimator::getCPSO(result->object, result->image);
			// Compute the PSF.
			run_cpso(cpso, result, n_cycles);
			// Save the results.
			fitsManager->saveImage(result);

			if(statistics){
				tracker_estimation.end();
				all_costs[i] = cpso->getGBestCost();
				all_times[i] = tracker_estimation.elapsedTime();
				all_stable_convergence[i] = cpso->getStableCycle();
			}

			printf("File %i\n", i);

			cpso->release();
		}

		if(statistics){
			cout << "* Statistics for Estimations:" << endl;
			cout << "- Time: avg.time=" << CPSO::calc_mean(all_times, n_images) <<
					", std.dev.time=" << CPSO::calc_stddev(all_times, n_images) <<
					", variance.time=" << CPSO::calc_variance(all_times, n_images) << endl;
			cout << "- Fitness: avg.fitness=" << CPSO::calc_mean(all_costs, n_images) <<
					", std.dev.fitness=" << CPSO::calc_stddev(all_costs, n_images) <<
					", variance.fitness=" << CPSO::calc_variance(all_costs, n_images) << endl;
			cout << "- Stabilization: avg.cycle=" << CPSO::calc_mean(all_stable_convergence, n_images) <<
					", std.dev.cycle=" << CPSO::calc_stddev(all_stable_convergence, n_images) <<
					", variance.cycle=" << CPSO::calc_variance(all_stable_convergence, n_images) << endl;
		}

		delete fitsManager;
		delete hamming;
	}

	// Free up the memory.
	delete zernikes;
	delete object;
	delete theta_values;
	delete radial_values;
	delete diff_mask;
	delete phase_mask;
	if(devices != NULL){
		delete devices;
	}

	PsfEstimator::shutdown();

#ifdef _PROFILING_
	cout << "Total OpenCL Time (miliseconds): " << Profiling::getTotalProcessingTime() << endl;
#endif

	tracker.end();
	printf("Total time...\n");
	tracker.print();

	return 0;
}

/**
 * @brief Print the usage of this tool to the standard output.
 */
void show_usage(){
	cout <<    "Usage: kippe [-h] | [-testgpu] | [-listgpu] |" << endl
			<< "             [-rms_contrast -obj <image> -isz <image_width>] |" << endl
			<< "             [-sum -obj <image> -isz <image_width>] |" << endl
			<< "             [-obj <object> -isz <image_width> -fsz <fft_width> -z <num_zernikes>" << endl
			<< "               [[-img <image> -out <output_path> -s <num_swarms> -p <num_particles>" << endl
			<< "                 -c <num_cycles> -w <w_val> -c1 <c1_val> -c2 <c2_val> -reset <reset_factor>] |" << endl
			<< "               [-validation -nval <num_validations> -ranksz <rank_size> {-noise <noise-level>}] |" << endl
			<< "               [-gen_imgs <num_artificial_images>]]" << endl
			<< "              {-r <coef_range>} {-fixedrand} {-statistics} {-devices <device_id_to_use>}" << endl
			<< "              {-parallel <n_threads>}]" << endl
			<< "where: " << endl
			<< "\t-h            This help message." << endl
			<< "\t-testgpu      Test all GPUs available." << endl
			<< "\t-listgpu      Get the list of device ids available on the system." << endl
			<< "\t-sum          Sum a 3D FITS file and generate a 2D FITS file as output." << endl
			<< "\t-rms_contrast Calculate the image intensity." << endl
			<< "\t-obj          The path to the FITS file containing the Object." << endl
			<< "\t-img          The path to either a FITS file containing a 2D/3D Image" << endl
			<< "\t              or the path to a folder containing multiple 2D Image files." << endl
			<< "\t-out          The output folder where the data will be saved." << endl
			<< "\t-isz          The widh of the input image." << endl
			<< "\t-fsz          The width of the FFT to be calculated." << endl
			<< "\t-z            The number of Zernikes to be calculated." << endl
			<< "\t-r            The range for the Zernike coefficients. If not provided, the value 2 will be used." << endl
			<< "\t-s            The number of swarms to be created." << endl
			<< "\t-p            The number of particles to be used on every swarm." << endl
			<< "\t-c            The number of cycles to run the CPSO." << endl
			<< "\t-w            The Inertia weight W." << endl
			<< "\t-c1           The Acceleration coefficient C1 (cognitive)." << endl
			<< "\t-c2           The Acceleration coefficient C2 (social)." << endl
			<< "\t-reset        The factor for re-initialization of inactive particles [0-1]." << endl
			<< "\t-validation   If provided, the tool will run on Validation mode." << endl
			<< "\t-nval         The number of validations to be executed." << endl
			<< "\t-ranksz       The rank size to store the best and worst values." << endl
			<< "\t-noise        The Poisson noise level to be added to the artificial image: 0 (default), 1, .. 700." << endl
			<< "\t-gen_imgs     If provided, the tool will run on Generation Images mode." << endl
			<< "\t              Specify the number of artificial images to be generated in the 3D output FITS file." << endl
			<< "\t-fixedrand    If provided, the Random Number Generator seed will be fixed at program startup." << endl
			<< "\t-statistics   If provided, the statistics will be printed on standard output." << endl
			<< "\t-devices      The list of device ids (separated by comma and no space) that should be used." << endl
			<< "\t              To get the list of device ids, run the program with the option -listgpu." << endl
			<< "\t-parallel     Define the number of threads to be used. For basic image processing, it is also" << endl
			<< "\t              the number of CPSO's to be executed in parallel. The default is 1 thread." << endl;
}
