
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * kippe_aux.cl
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#define BARRIER {barrier(CLK_LOCAL_MEM_FENCE);}
#define BARRIER_GLOBAL {barrier(CLK_GLOBAL_MEM_FENCE);}

#ifdef _DOUBLE_WORD_
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif

__kernel void cl_generate_phase(__global const WORD *g_zernikes, __global WORD *g_coefs, __constant int *g_phase_mask,
		__global FFT_TYPE *g_pupil, __global WORD *g_phase, __local WORD *l_coefs, int phase_size, int n_zernikes,
		int max_threads)
{
	int lid = get_local_id(0);
	int gid = get_group_id(0);

	int i;
	for(i=lid; i < n_zernikes; i+=get_local_size(0)){
		l_coefs[i] = g_coefs[gid * n_zernikes + i];
	}

	if (get_global_id(0) < max_threads) {
		int z_pos, size_fft = phase_size * phase_size;
		WORD p_phase;
		FFT_TYPE l_pupil;
		for(i=0; i < phase_size; i++){
			p_phase = 0;
			l_pupil.x = 0;
			l_pupil.y = 0;
			z_pos = i * phase_size + lid;
			for (int j=0; j < n_zernikes; j++) {
				p_phase += l_coefs[j] * g_zernikes[j * size_fft + z_pos];
			}
			// Formula: cexp(p * i) = cos(p) + i * sin(p)
			if(g_phase_mask[i * phase_size + lid]){
				l_pupil.x = cos(p_phase);
				l_pupil.y = sin(p_phase);
			}

			g_phase[gid * size_fft + z_pos] = p_phase; // Only stored for validation purpose.
			g_pupil[gid * size_fft + z_pos] = l_pupil;
		}
	}
}

__kernel void cl_power_spec(__global FFT_TYPE *in, __global WORD *out, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		FFT_TYPE valIn = in[idx];
		WORD valAbs = sqrt((valIn.x * valIn.x) + (valIn.y * valIn.y));
		out[idx] = valAbs * valAbs;
	}
}

__kernel void cl_multiply_complexarr(__global FFT_TYPE *in, __global FFT_TYPE *out, WORD scale, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		FFT_TYPE valIn = in[idx];
		FFT_TYPE valOut = {valIn.x * scale, valIn.y * scale};
		out[idx] = valOut;
	}
}

__kernel void cl_multiply_fftw_complex_arrays(__global FFT_TYPE *var_arr, __global FFT_TYPE *fix_arr, __global FFT_TYPE *var_out, int img_size, int max_threads) {
	int lid = get_local_id(0);
	int gid = get_group_id(0);
	if (get_global_id(0) < max_threads) {
		int img_area = img_size * img_size;
		FFT_TYPE valIn1;
		FFT_TYPE valIn2;
		for(int i=0, v_pos, f_pos; i < img_size; i++){
			f_pos = i * img_size + lid;
			v_pos = gid * img_area + f_pos;
			valIn1 = var_arr[v_pos];
			valIn2 = fix_arr[f_pos];
			FFT_TYPE valOut = {
				(valIn1.x * valIn2.x) - (valIn1.y * valIn2.y),
				(valIn1.x * valIn2.y) + (valIn1.y * valIn2.x)
			};
			var_out[v_pos] = valOut;
		}
	}
}

__kernel void cl_multiply_doublearr(__global WORD *in, __global WORD *out, WORD scale, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		out[idx] = in[idx] * scale;
	}
}

__kernel void cl_real2complex(__global WORD *in, __global FFT_TYPE *out, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		FFT_TYPE valOut = {in[idx], 0};
		out[idx] = valOut;
	}
}

/*
 * Convert complex to real. 
 */
__kernel void cl_complex2real(__global FFT_TYPE *in, __global WORD *out, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		FFT_TYPE valIn = in[idx];
		out[idx] = sqrt((valIn.x * valIn.x) + (valIn.y * valIn.y));
	}
}

/*
 * Extract just the real part of the number. 
 */
__kernel void cl_real(__global FFT_TYPE *in, __global WORD *out, int max_threads) {
	int idx = get_global_id(0);
	if (idx < max_threads) {
		FFT_TYPE valIn = in[idx];
		out[idx] = valIn.x;
	}
}

/*
    This version adds multiple elements per thread sequentially.  This reduces the overall
    cost of the algorithm while keeping the work complexity O(n) and the step complexity O(log n).
    (Brent's Theorem optimization).
    This algorithm only works for multiple of 2 block sizes (32, 64, 128, 256 and 512), as long as
    the blocksize<=512.

    Note, this kernel needs a minimum of 64*sizeof(T) bytes of shared memory.
    In other words if blockSize <= 32, allocate 64*sizeof(T) bytes.
    If blockSize > 32, allocate blockSize*sizeof(T) bytes.
*/
__kernel void cl_reduce(__global WORD *g_idata, __global WORD *g_odata, __local WORD *sdata, unsigned int size)
{
    // perform first level of reduction,
    // reading from global memory, writing to shared memory
    unsigned int tid = get_local_id(0);
    unsigned int g_size = ( get_group_id(1) + 1 ) * size;
    unsigned int i = get_group_id(1) * size + get_group_id(0) * REDUCTION_BLOCKSZ * 2  +  tid;
    unsigned int gridSize = REDUCTION_BLOCKSZ * 2 * get_num_groups(0);

    WORD mySum = 0;

    // we reduce multiple elements per thread.  The number is determined by the
    // number of active thread blocks (via gridDim).  More blocks will result
    // in a larger gridSize and therefore fewer elements per thread
    while (i < g_size)
    {
        mySum += g_idata[i];
        if (i + REDUCTION_BLOCKSZ < g_size){
            mySum += g_idata[i+REDUCTION_BLOCKSZ];
		}
        i += gridSize;
    }

    // each thread puts its local sum into shared memory
    sdata[tid] = mySum;
    BARRIER

    // do reduction in shared mem
    if (REDUCTION_BLOCKSZ >= 512) { if (tid < 256) { sdata[tid] = mySum = mySum + sdata[tid + 256]; } BARRIER; }
    if (REDUCTION_BLOCKSZ >= 256) { if (tid < 128) { sdata[tid] = mySum = mySum + sdata[tid + 128]; } BARRIER; }
    if (REDUCTION_BLOCKSZ >= 128) { if (tid <  64) { sdata[tid] = mySum = mySum + sdata[tid +  64]; } BARRIER; }

    if (tid < 32) {
        if (REDUCTION_BLOCKSZ >=  64) { sdata[tid] = mySum = mySum + sdata[tid + 32]; } BARRIER;
        if (REDUCTION_BLOCKSZ >=  32) { sdata[tid] = mySum = mySum + sdata[tid + 16]; } BARRIER;
        if (REDUCTION_BLOCKSZ >=  16) { sdata[tid] = mySum = mySum + sdata[tid +  8]; } BARRIER;
        if (REDUCTION_BLOCKSZ >=   8) { sdata[tid] = mySum = mySum + sdata[tid +  4]; } BARRIER;
        if (REDUCTION_BLOCKSZ >=   4) { sdata[tid] = mySum = mySum + sdata[tid +  2]; } BARRIER;
        if (REDUCTION_BLOCKSZ >=   2) { sdata[tid] = mySum = mySum + sdata[tid +  1]; }
    }

    // write result for this block to global mem
    if (tid == 0){
        g_odata[get_group_id(1) * REDUCTION_NBLOCKS + get_group_id(0)] = sdata[0];
    }
}

__kernel void cl_reduce_fft(__global FFT_TYPE *g_idata, __global FFT_TYPE *g_odata, __local FFT_TYPE *sdata, unsigned int n)
{
    // perform first level of reduction,
    // reading from global memory, writing to shared memory
    unsigned int tid = get_local_id(0);
    unsigned int i = get_group_id(0) * REDUCTION_BLOCKSZ * 2  +  tid;
    unsigned int gridSize = REDUCTION_BLOCKSZ * 2 * get_num_groups(0);

    FFT_TYPE mySum = {0, 0};

    // we reduce multiple elements per thread.  The number is determined by the
    // number of active thread blocks (via gridDim).  More blocks will result
    // in a larger gridSize and therefore fewer elements per thread
    while (i < n)
    {
        mySum.x += g_idata[i].x;
        mySum.y += g_idata[i].y;
        if (i + REDUCTION_BLOCKSZ < n){
            mySum.x += g_idata[i+REDUCTION_BLOCKSZ].x;
            mySum.y += g_idata[i+REDUCTION_BLOCKSZ].y;
			}
        i += gridSize;
    }

    // each thread puts its local sum into shared memory
    sdata[tid].x = mySum.x;
    sdata[tid].y = mySum.y;
    BARRIER;

    // do reduction in shared mem
    if (REDUCTION_BLOCKSZ >= 512) { if (tid < 256) { sdata[tid].x = mySum.x = mySum.x + sdata[tid + 256].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid + 256].y;} BARRIER; }
    if (REDUCTION_BLOCKSZ >= 256) { if (tid < 128) { sdata[tid].x = mySum.x = mySum.x + sdata[tid + 128].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid + 128].y;} BARRIER; }
    if (REDUCTION_BLOCKSZ >= 128) { if (tid <  64) { sdata[tid].x = mySum.x = mySum.x + sdata[tid +  64].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid +  64].y;} BARRIER; }

    if (tid < 32)
    {
        if (REDUCTION_BLOCKSZ >=  64) { sdata[tid].x = mySum.x = mySum.x + sdata[tid + 32].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid + 32].y;} BARRIER;
        if (REDUCTION_BLOCKSZ >=  32) { sdata[tid].x = mySum.x = mySum.x + sdata[tid + 16].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid + 16].y;} BARRIER;
        if (REDUCTION_BLOCKSZ >=  16) { sdata[tid].x = mySum.x = mySum.x + sdata[tid +  8].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid +  8].y;} BARRIER;
        if (REDUCTION_BLOCKSZ >=   8) { sdata[tid].x = mySum.x = mySum.x + sdata[tid +  4].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid +  4].y;} BARRIER;
        if (REDUCTION_BLOCKSZ >=   4) { sdata[tid].x = mySum.x = mySum.x + sdata[tid +  2].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid +  2].y;} BARRIER;
        if (REDUCTION_BLOCKSZ >=   2) { sdata[tid].x = mySum.x = mySum.x + sdata[tid +  1].x; sdata[tid].y = mySum.y = mySum.y + sdata[tid +  1].y;}
    }

    // write result for this block to global mem
    if (tid == 0){
        g_odata[get_group_id(0)].x = sdata[0].x;
        g_odata[get_group_id(0)].y = sdata[0].y;
    }
}

__kernel void cl_resize_psf(__global WORD *psf, __global WORD *psf_out, int size_in, int size_out, int max_threads)
{
	int idx = get_global_id(0);
  if (idx < max_threads)
  {
		int x = idx / size_out;
		int y = idx - (x * size_out);
		if(PSF_EXTRACT == PSF_CORNER){
			int i = x + ( (x >= (size_out / 2)) * (size_in - size_out) );
			int j = y + ( (y >= (size_out / 2)) * (size_in - size_out) );
			psf_out[idx] = psf[i * size_in + j];
		}
		else if(PSF_EXTRACT == PSF_CENTER){
			int start = size_in / 2 - size_out / 2 + 1;
			psf_out[idx] = psf[(y + start) * size_in + (x + start)];
		}
	}
}

__kernel void cl_calc_cost_imgh(__global FFT_TYPE *fimage, __global FFT_TYPE *fconob, __global WORD *result, __global int *diffraction_mask, WORD scale, int img_size, int max_threads)
{
	int lid = get_local_id(0);
	int gid = get_group_id(0);
	if (get_global_id(0) < max_threads) {
		int img_area = img_size * img_size;
		int img_sizeh = (img_size / 2) + 1;
		int img_areah = img_sizeh * img_size;
		FFT_TYPE img, obj;
		WORD x, y;
		for(int row=0, img_pos, obj_pos, filter_pos, out_pos; row < img_size; row++){
			img_pos = row * img_size + lid;
			obj_pos = gid * img_area + img_pos;
			filter_pos = row * img_sizeh + lid;
			out_pos = gid * img_areah + filter_pos;
			img = fimage[img_pos];
			obj = fconob[obj_pos];
			x = (img.x - obj.x);
			y = (img.y - obj.y);
			result[out_pos] = (diffraction_mask[filter_pos]) * (sqrt(x * x + y * y) / scale);
		}
	}
}

__kernel void cl_calc_cost(__global FFT_TYPE *fimage, __global FFT_TYPE *fconob, __global WORD *result, __global int *diffraction_mask, WORD scale, int img_size, int max_threads)
{
	int lid = get_local_id(0);
	int gid = get_group_id(0);
	if (get_global_id(0) < max_threads) {
		int img_area = img_size * img_size;
		FFT_TYPE img, obj;
		WORD x, y;
		for(int row=0, img_pos, obj_pos; row < img_size; row++){
			img_pos = row * img_size + lid;
			obj_pos = gid * img_area + img_pos;
			img = fimage[img_pos];
			obj = fconob[obj_pos];
			x = (img.x - obj.x);
			y = (img.y - obj.y);
			result[obj_pos] = (diffraction_mask[img_pos]) * (sqrt(x * x + y * y) / scale);
		}
	}
}

__kernel void cl_calc_mismatch(__global FFT_TYPE *image, __global FFT_TYPE *conobj_fp2, __global FFT_TYPE *conobj,
		__global WORD *cost, __global WORD *mismatch, int img_size, int max_threads)
{
	int lid = get_local_id(0);
	int gid = get_group_id(0);
	if (get_global_id(0) < max_threads) {
		int img_area = img_size * img_size;
		FFT_TYPE img, obj;
		for(int row=0, img_pos, obj_pos; row < img_size; row++){
			img_pos = row * img_size + lid;
			obj_pos = gid * img_area + img_pos;
			img = image[img_pos];
			obj = conobj_fp2[obj_pos]; // bring the FFT_TYPE point from global memory. 
			conobj[obj_pos] = obj.x; // send the point as WORD back to global memory.
			cost[obj_pos] = fabs(img.x - obj.x);
			mismatch[obj_pos] = (WORD)(img.x != obj.x);
		}
	}
}

__kernel void cl_adjust_fft(__global FFT_TYPE *fft, int round, int max_threads)
{
	int idx = get_global_id(0);
	if(idx < max_threads){
		FFT_TYPE val = fft[idx];
		WORD tmp;
		if(isnan(val.x) || isinf(val.x)){
			val.x = 0;
		}
		else if(round) {
			if(fabs(val.x) <= 2e-05){
				val.x = 0;
			}
			else if((tmp=ceil(fabs(val.x))-fabs(val.x)) > 0 && tmp <= 2e-05){
				val.x = copysign(ceil(fabs(val.x)), val.x);
			}
			else if((tmp=fabs(val.x)-floor(fabs(val.x))) > 0 && tmp <= 2e-05){
				val.x = copysign(floor(fabs(val.x)), val.x);
			}
		}

		if(isnan(val.y) || isinf(val.y)){
			val.y = 0;
		}
		else if(round) {
			if(fabs(val.y) <= 2e-05){
				val.y = 0;
			}
			else if((tmp=ceil(fabs(val.y))-fabs(val.y)) > 0 && tmp <= 2e-05){
				val.y = copysign(ceil(fabs(val.y)), val.y);
			}
			else if((tmp=fabs(val.y)-floor(fabs(val.y))) > 0 && tmp <= 2e-05){
				val.y = copysign(floor(fabs(val.y)), val.y);
			}
		}

		fft[idx] = val;
	}
}

uint irand(uint *ctx) {
	uint t = ctx[0] ^ (ctx[0] << 11);
	ctx[0] = ctx[1];
	ctx[1] = ctx[2];
	ctx[2] = ctx[3];
	ctx[3] = ctx[3] ^ (ctx[3] >> 19) ^ (t ^ (t >> 8));

	return ctx[3];
}

float frand(uint *ctx) {
	return (WORD)irand(ctx) / (WORD)RAND_MAX; // * 2.3283064e-10;
}

__kernel void cl_cpso(__global WORD *x, __global WORD *v, __global WORD *p_best, __global WORD *g_best,
		__local WORD *l_pbest_value, __global WORD *pbest_value, __global WORD *gbest_value, __global int *reset_search,
		__global int *swarm_dim, __global WORD *cl_cost, __global uint4 *rand_ctx, __global WORD *w, WORD w_ini, __global WORD *c1,
		__global WORD *c2, int n_particles, int n_swarms, WORD range, WORD reset_at, int n_iterations)
{
	int particle_id = get_local_id(0);
	int swarm_id = get_group_id(0);
	int g_population_id = swarm_id * n_particles + particle_id;

	__local int start_swarm_dim;
	__local int end_swarm_dim;
	__local WORD l_gbest_value;
	__local int full_ndim; 
	if(particle_id == 0){
		// Put gbest value to local memory.
		l_gbest_value = gbest_value[swarm_id];

		full_ndim = swarm_dim[n_swarms - 1] + 1;
		if(swarm_id == 0){
			start_swarm_dim = 0;
		}
		else{
			start_swarm_dim = swarm_dim[swarm_id - 1] + 1;
		}
		if(swarm_id == n_swarms - 1){
			end_swarm_dim = full_ndim - 1;
		}
		else{
			end_swarm_dim = swarm_dim[swarm_id];
		}
	}
	BARRIER;

	// Finish the calculation of the cost.
	int i, pos1, pos2;
	WORD sum = 0;
	for(i=0; i < REDUCTION_NBLOCKS; i++){
		sum += cl_cost[g_population_id * REDUCTION_NBLOCKS + i];
	}

	// Put pbest value in local memory.
	l_pbest_value[particle_id] = pbest_value[g_population_id];

	// Check the PBest.
	if(sum < l_pbest_value[particle_id]){
		// Update pbest if new value is better.
		l_pbest_value[particle_id] = sum;
		pbest_value[g_population_id] = sum;
		// Replace the pbest positions directly to global memory.
		for(i=0; i < full_ndim; i++){
			pos1 = g_population_id * full_ndim + i;
			p_best[pos1] = x[pos1];
		}
	}
	// Barrier is required to verify the gbest.
	BARRIER;
	BARRIER_GLOBAL; // Global memory must be synchronized before continuing.

	// Check the GBest.
	// Only the first particle in the swarm does the verification.
	if(particle_id == 0){
		for(i=1, pos1=0; i < n_particles; i++){
			// Check from local memory what particle has the better pbest.
			if(l_pbest_value[i] < l_pbest_value[pos1]){
				pos1 = i;
			}
		}
		if(l_pbest_value[pos1] < gbest_value[swarm_id]){
			// Replace the gbest value directly to global memory.
			gbest_value[swarm_id] = l_pbest_value[pos1];
			// Replace the gbest positions directly to global memory.
			pos1 += swarm_id * n_particles;
			for(i=start_swarm_dim; i <= end_swarm_dim; i++){
				g_best[i] = p_best[pos1 * full_ndim + i];
			}
		}
	}
	// Barrier is required to synchronize the threads not in same warp as particle 0;
	BARRIER_GLOBAL;

	// Update the Velocity and Position.
	uint4 rand = rand_ctx[g_population_id];
	WORD l_w = w[g_population_id];
	WORD l_c1 = c1[g_population_id];
	WORD l_c2 = c2[g_population_id];
	WORD new_v, old_v, new_x, old_x, calc1, calc2, diff;
	int curr_reset = reset_search[g_population_id];
	int reset = (curr_reset == 5); // Requires 5 iterations under the limit to reset the particle.
	int reset_all = reset & (frand(&rand) < .5);
	curr_reset = (reset) ? 0 : curr_reset;

	diff = 0;
	// Time to time, replace the entire position of the first particle with the gbest.
	if(particle_id == 0 && frand(&rand) < .05){
		// Select a particle from the current swarm to have the position x replaced.
		pos1 = g_population_id * full_ndim;			
		for(i=0; i < full_ndim; i++, pos1++){
			new_x = g_best[i];
			// Only the swarms' dimensions count for the reinitialization mechanism.
			diff += (i >= start_swarm_dim && i <= end_swarm_dim) ? fdim(x[pos1], new_x) : 0;
			x[pos1] = new_x;
		}
	}
	else{
		// The entire dimension is required to calculate the PSF.
		for(i=0; i < full_ndim; i++){
			pos1 = particle_id * full_ndim + i;
			// Position exclusive for the position x, which contains the entire dimension.
			pos2 = g_population_id * full_ndim + i;
			// If the current dimension is part of this swarm,
			// does the update of the velocity and position.
			if(i >= start_swarm_dim && i <= end_swarm_dim){
				// If it was determined that this particle has to be reset.
				if(reset){
					if(reset_all){
						// Reset the current velocity and position.
						new_v = (frand(&rand) - 0.5) * 2 * range;
						new_v = ((new_v > range) ? -2.0 * range + new_v : ((new_v < -range) ? 2 * range + new_v: new_v)); 
						v[pos1] = new_v;
	
						new_x = (frand(&rand) - 0.5) * 2 * range;
						new_x = ((new_x > range) ? -2.0 * range + new_x : ((new_x < -range) ? 2 * range + new_x: new_x)); 
						diff += fdim(x[pos2], new_x);
						x[pos2] = new_x;
					}

					// Reset parameters w, c1, and c2:
					l_w = w_ini * frand(&rand); //  in the interval [0, w_ini].
					l_c2 = frand(&rand); // in the interval [2, 3].
					l_c1 = frand(&rand); // in the interval [2, 3].
				}
	
				// If not all values are being reset (or if no reset at all).
				if(!reset_all){
					old_v = ((old_v=v[pos1]) > range) ? range - fmod(old_v, range) : ((old_v < -range) ? (-range) - fmod(old_v, range) : old_v);
					old_x = x[pos2];
					calc1 = p_best[pos1] - x[pos2];
					calc2 = g_best[i] - p_best[pos1];
					new_v = l_w * old_v + l_c1 * frand(&rand) * calc1
										+ l_c2 * frand(&rand) * calc2;
	
					// Ensure new_v will be restricted to the interval [-range/2, range/2] by reducing its speed.
					calc1 = (range / 2.0) - old_v;
					calc2 = (-range / 2.0) - old_v;
					new_v = (new_v > range / 2.0) ? old_v + (calc1/new_v)*calc1 : ((new_v < -range / 2.0) ? old_v + (calc2/new_v)*calc2 : new_v);
					v[pos1] = new_v;
	
					new_x = (( new_x=old_x+new_v ) > range) ? 2.0 * range - new_x : ((new_x < -range) ? -2.0 * range - new_x : new_x);
					diff += fdim(old_x, new_x);
					x[pos2] = new_x;
				}
			}
			else{
				// If current dimension is not part of this swarm,
				// just copy the respective gbest.
				x[pos2] = g_best[i];
			}
			reset_all = 0;
		}
	}
	// Calculate a adapted Linear Decreasing Weight (LDW).
	l_w -= w_ini / n_iterations;
	w[g_population_id] = (l_w <= 0) ? w_ini : l_w;
	// If the parameters were reset, store the new values on global memory.
	if(reset){
		c1[g_population_id] = l_c1;
		c2[g_population_id] = l_c2;
	}
	// Update the rand seed.
	rand_ctx[g_population_id] = rand;
	// Check if it's necessary to reset the particle.
	reset_search[g_population_id] = (((diff / (end_swarm_dim-start_swarm_dim+1)) / range) < reset_at) ? curr_reset + 1 : 0;
}
