
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * kippef.cpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "kippef.hpp"

using namespace std;

/**
 * @brief Minimize a cost function by using the search method CPSO (Cooperative Particle Swarm Optimization).
 *
 * @param cpso The CPSO instance.
 * @param result A pointer where all the best values found will be stored. This pointer must be obtained through the {link FitsManager} class.
 * @param n_cycles The number of cycles to run the CPSO.
 */
void run_cpso(CPSO *cpso, fits_result *result, int n_cycles){
	TimeTracker **trackers = NULL;
#ifdef _TRACK_TIME_
	// Create the trackers.
	int num_trackers = 6;
	trackers = new TimeTracker*[num_trackers];
	trackers[0] = new TimeTracker("cpso");
	trackers[1] = new TimeTracker("cycles");
	trackers[2] = new TimeTracker("generatePhase");
	trackers[3] = new TimeTracker("fft");
	trackers[4] = new TimeTracker("matrix-operations");
	trackers[5] = new TimeTracker("reduction");
#endif

	TRACK(trackers[0]->resume())
	cpso->run(trackers, n_cycles);
	TRACK(trackers[0]->pause())

	// Collect the best values found.
	cpso->getBestCoefs(result->coefs);
	cpso->getBestPhase(result->phase_generated);
	cpso->getBestPsfe(result->psf_generated);
	cpso->getBestConvolvedObject(result->conobj);
	CPSO::invertPsf(result->psf_generated, result->psf_inverted, cpso->getImageSize());
	cpso->calcDifference(trackers, result->conobj, result->img_diff);

#ifdef _TRACK_TIME_
	// Print tracking information.
	cout << "Time spent per piece of code:\n";
	for (int i = 0; i < num_trackers; i++) {
		trackers[i]->end();
		trackers[i]->print();
		delete trackers[i];
	}
	delete[] trackers;
	cout << "Stabilization at cycle: " << cpso->getStableCycle() << endl;
#endif
}

/**
 * @brief Compare a PSF file against the Object file and the Image file,
 * by convolving the PSF with the Object and then comparing the result
 * with the Image.
 *
 * The output is the cost of the difference (convolved object - image)
 * and their mismatch (how many pixels are different between them).
 *
 * @param psfFileName The PSF file to be validated.
 * @param objFileName The Object file name.
 * @param imgFileName The Image file name.
 * @param zernikes The complete Zernike terms.
 * @param diffraction_mask The diffraction mask.
 * @param phase_mask The phase mask.
 * @param n_zernikes The number of Zernike terms that are being used on these calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 */
void validate_psf_file(char *psfFileName, char *objFileName, char *imgFileName, double* zernikes, int* diffraction_mask, int *phase_mask,
		int n_zernikes, int image_size, int phase_size, double range){
	int img_area = image_size * image_size;
	double *psf = new double[img_area];
	double *obj = new double[img_area];
	double *img = new double[img_area];

	// Read the files.
	read_fits_file(psfFileName, psf, img_area);
	read_fits_file(objFileName, obj, img_area);
	read_fits_file(imgFileName, img, img_area);

	// Initialize the OpenCl environment.
	clFactory::startup();

	// Load the data to memory.
	CPSO *cpso = new CPSO(zernikes, phase_mask, diffraction_mask, phase_size, image_size, n_zernikes, range, 0, 0, 0, 0, 1, 1);
	cpso->set_images(obj, img);

	// Execute the validation.
	FFT_TYPE ret = cpso->validatePsf(psf);
	cout << "Final Image Difference: = " << ret.x << " distance with " << ret.y << " different points." << endl;

	// Release the OpenCl data.
	cpso->release();
	cpso->clear_static_data();
	cpso->finalize_cl();

	// Finalize the OpenCl environment.
	clFactory::shutdown();

	delete psf;
	delete obj;
	delete img;
	delete cpso;
}

void nollTest(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, vector<int> *_devices){

	cout << "Starting Noll testing..." << endl;

	// CPSO Startup.
	PsfEstimator::startup(1, zernikes, phase_mask, diffraction_mask, phase_size,  image_size, n_zernikes, range, 1.25, 1, 0.5, 0.25, 1, 1, _devices);

	int image_area = image_size * image_size;
	ostringstream *oss = new ostringstream();
	CPSO *cpso = NULL;
	WORD *coefs = new WORD[n_zernikes];
	WORD *artificial_image1 = new WORD[image_area];
	WORD *artificial_image2 = new WORD[image_area];
	WORD *psfNoll1 = new WORD[image_area];
	WORD *psfNoll2 = new WORD[image_area];
	WORD *psfNollInv1 = new WORD[image_area];
	WORD *psfNollInv2 = new WORD[image_area];
	WORD *phase1 = new WORD[image_area];
	WORD *phase2 = new WORD[image_area];

	for(int i=0; i < n_zernikes; i++) coefs[i] = 0;
	coefs[(int)(rand() / (WORD)RAND_MAX) * (n_zernikes-1)] = CPSO::randNormalDistribution(0, range / 4);

	// Generate a random image.
	cpso = PsfEstimator::getCPSO(object, object, coefs);
	cpso->runPsf(NULL, 1);
	cpso->saveFirstResult();
	cpso->getBestCoefs(coefs);
	cpso->getBestConvolvedObject(artificial_image1);
	cpso->getBestPsfe(psfNoll1);
	cpso->getBestPhase(phase1);
	*oss << object_filename << "-noll-image-artificial.fits";
	write_fits_file((char*)oss->str().c_str(), artificial_image1, image_size, image_size);
	oss->str("");
	CPSO::invertPsf(psfNoll1, psfNollInv1, image_size);
	*oss << object_filename << "-noll-psfinv.fits";
	write_fits_file((char*)oss->str().c_str(), psfNollInv1, image_size, image_size);
	oss->str("");
	cpso->release();

	// Inverse the coefficients.
	for(int i=0; i < n_zernikes; i++) coefs[i] *= -1;

	// Generate a new artificial image with the coefficients inverted.
	cpso = PsfEstimator::getCPSO(object, object, coefs);
	cpso->runPsf(NULL, 1);
	cpso->saveFirstResult();
	cpso->getBestConvolvedObject(artificial_image2);
	cpso->getBestPsfe(psfNoll2);
	cpso->getBestPhase(phase2);
	*oss << object_filename << "-nollinv-image-artificial.fits";
	write_fits_file((char*)oss->str().c_str(), artificial_image2, image_size, image_size);
	oss->str("");
	CPSO::invertPsf(psfNoll2, psfNollInv2, image_size);
	*oss << object_filename << "-nollinv-psfinv.fits";
	write_fits_file((char*)oss->str().c_str(), psfNollInv2, image_size, image_size);
	cpso->release();

	double sum = 0;
	for(int i=0; i < image_area; i++) sum += abs(psfNollInv1[i] - psfNollInv2[i]);
	printf("Difference: %lf", sum);

	delete cpso;
	delete artificial_image1;
	delete artificial_image2;
	delete coefs;
	delete psfNoll1;
	delete psfNoll2;
	delete psfNollInv1;
	delete psfNollInv2;
	delete phase1;
	delete phase2;
	delete oss;

	cout << "Noll testing done." << endl;
}

/**
 * @brief Validate the PSF Estimation through CPSO technique.
 *
 * This method validate the technique by running 2 verifications:
 * <ul>
 * <li>Generating random PSFs / artificial Images and then setting the same coefficients to the PSF Estimation code.
 * This verifies if the code is able to duplicate the same results based on the same set of coefficients.
 * No search is executed on this validation.</li>
 * <li>Generating random PSFs / artificial Images and then submitting complete CPSO searches for these artificial images.
 * The best and worst values are saved to files at the end of the execution of this method.
 * The information saved is: Images, PSFs, phases and coefficients.
 * This verifies how good the estimation technique is fine tuned.</li>
 * </ul>
 *
 * @param object The Object data that will be used for the validation.
 * @param zernikes The complete Zernike matrix.
 * @param diffraction_mask The diffraction mask.
 * @param n_zernikes The number of Zernike terms that is being used for the calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 * @param phase_mask The phase mask.
 * @param object_filename The Object file name, to be used as base to save the results.
 * @param n_validations The number of validations to be done.
 * @param rank_size The number of best and good results to be kept.
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 * @param poisson The poisson noise level to be added to the artificial image.
 * @param parallel The number of threads to be used for the initial validations.
 */
void validate_psfcode(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, int n_validations, int rank_size, vector<int> *_devices,
		int poisson, int parallel){

	TimeTracker **trackers = NULL;
#ifdef _TRACK_TIME_
	// Create the trackers.
	TimeTracker tracker_estimation("single-estimation");
	int num_trackers = 6;
	trackers = new TimeTracker*[num_trackers];
	trackers[0] = new TimeTracker("cpso");
	trackers[1] = new TimeTracker("cycles");
	trackers[2] = new TimeTracker("generatePhase");
	trackers[3] = new TimeTracker("fft");
	trackers[4] = new TimeTracker("matrix-operations");
	trackers[5] = new TimeTracker("reduction");
#endif

	TRACK(trackers[0]->resume())
	cout << "Starting validations..." << endl;

	// CPSO Startup.
	PsfEstimator::startup(parallel, zernikes, phase_mask, diffraction_mask, phase_size,  image_size, n_zernikes, range, 1.25, 1, 0.5, 0.25, 1, 1, _devices);

	int size_fft = phase_size * phase_size;
	int image_area = image_size * image_size;
	WORD total_cost = 0;
	CPSO *cpso = NULL;
	// Validate the PSF reconstruction algorithm.
	omp_set_num_threads(parallel);
#pragma omp parallel for shared(trackers, object, total_cost) \
		firstprivate(n_validations, n_zernikes, image_area, size_fft, cpso) \
		default(none)
	for(int i=0; i < n_validations; i++){
		WORD *coefs = new WORD[n_zernikes];
		WORD *conobj = new WORD[image_area];
		cpso = PsfEstimator::getCPSO(object, object);
		// Calculates 1 random PSF and the artificial corrupted image.
		cpso->runPsf(trackers, 1);
		// Save the results.
		cpso->saveFirstResult();
		cpso->getBestCoefs(coefs);
		cpso->getBestConvolvedObject(conobj); // Artificial Corrupted Image.
		cpso->release();

		// Convert to double, if necessary.
#ifdef _DOUBLE_WORD_
		double *artificial_image = conobj;
#else
		double *artificial_image = new double[size_fft];
		for(int j=0; j < image_area; j++) artificial_image[j] = conobj[j];
#endif

		// Calculates how far it goes from the artificial corrupted image
		// by using the same coefficients and running 1 single time.
		cpso = PsfEstimator::getCPSO(object, artificial_image, coefs);
		cpso->run(trackers, 1);

		// Sum the difference from the artificial image.
		#pragma omp critical (total_cost)
			total_cost += cpso->getGBestCost();

		cpso->release();

		delete coefs;
		delete conobj;
#ifndef _DOUBLE_WORD_
		delete artificial_image;
#endif
	}
	// The Total Cost must be something really small (something between 0 and 2).
	cout << "Total difference for " << n_validations << " validations: " << total_cost << endl;

	int best_pos, worst_pos;
	WORD best_costs[rank_size], worst_costs[rank_size], curr_cost;
	WORD *psf = new WORD[image_area];
	WORD *phase = new WORD[size_fft];
	WORD *image = new WORD[image_area];
	WORD *coefs = new WORD[n_zernikes];
	FFT_TYPE *object_fft = new FFT_TYPE[image_area];
	FFT_TYPE *psf_fft = new FFT_TYPE[image_area];
	FFT_TYPE *image_fft = new FFT_TYPE[image_area];

	WORD *best_artificial_psf = new WORD[image_area * rank_size];
	WORD *best_artificial_phase = new WORD[size_fft * rank_size];
	WORD *best_artificial_image = new WORD[image_area * rank_size];
	WORD *best_coefs = new WORD[n_zernikes * rank_size];
	WORD *best_psf = new WORD[image_area * rank_size];
	WORD *best_phase = new WORD[size_fft * rank_size];
	WORD *best_image = new WORD[image_area * rank_size];
	WORD *best_imgdiff = new WORD[image_area * rank_size];
	FFT_TYPE *best_artificial_psf_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *best_artificial_image_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *best_psf_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *best_image_fft = new FFT_TYPE[image_area * rank_size];

	WORD *worst_artificial_psf = new WORD[image_area * rank_size];
	WORD *worst_artificial_phase = new WORD[size_fft * rank_size];
	WORD *worst_artificial_image = new WORD[image_area * rank_size];
	WORD *worst_coefs = new WORD[n_zernikes * rank_size];
	WORD *worst_psf = new WORD[image_area * rank_size];
	WORD *worst_phase = new WORD[size_fft * rank_size];
	WORD *worst_image = new WORD[image_area * rank_size];
	WORD *worst_imgdiff = new WORD[image_area * rank_size];
	FFT_TYPE *worst_artificial_psf_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *worst_artificial_image_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *worst_psf_fft = new FFT_TYPE[image_area * rank_size];
	FFT_TYPE *worst_image_fft = new FFT_TYPE[image_area * rank_size];

#ifdef _TRACK_TIME_
	double all_costs[n_validations], all_times[n_validations], all_stable_convergence[n_validations];
#endif

	double *artificial_image = new double[size_fft];

	int i, j;
	for(i=0; i < rank_size; i++) best_costs[i] = WORD_MAX;
	memset(worst_costs, 0, rank_size * SIZEOF_WORD);

	// Calculate the Hamming window.
	double *hamming = new double[image_area];
	frac_hamming(hamming, 0.3, image_size, image_size);
	// Apply the Hamming window over the Object.
	double *object_hamming = new double[image_area];
	for(j=0; j < image_area; j++) object_hamming[j] = object[j];
	apply_window(object_hamming, hamming, image_area);

	// Validate the PSF estimation method through CPSO.
	cout << "Making a complete search..." << endl;
	for(i=0; i < n_validations; i++){
		// Calculates 1 random PSF and the artificial image.
		//--------------------------------------------------
		cout << "Processing search #" << i << endl;
		// No image is used here, so just repeat the object as parameter.
		CPSO::generateNormalDistrRandomCoefs(coefs, n_zernikes, range);
		cpso = PsfEstimator::getCPSO(object, object, coefs);
		// No search is made, only a random PSF is applied over the object.
		cpso->runPsf(trackers, 1);
		// Save the results.
		cpso->saveFirstResult();
		cpso->getObjectFFT(object_fft);
		cpso->getBestPhase(phase);
		cpso->getBestPsfe(psf);
		cpso->getBestPsfeFFT(psf_fft);
		cpso->getBestConvolvedObject(image); // Artificial Corrupted Image.
		cpso->release();

		// Add the Poisson Noise to the artificial image.
		if(poisson){
			addPoissonNoise(image, image_size, image_size, poisson, 0);
		}

		// Duplicate the values, converting to double if necessary.
		for(j=0; j < image_area; j++) artificial_image[j] = image[j];

		// Apply the Hamming window over the Image.
		apply_window(artificial_image, hamming, image_area);

		// Try to find the same artificial corrupted image by making a complete search.
		//-----------------------------------------------------------------------------
		// Set the pool configuration for a complete search.
		TRACK(tracker_estimation.start())
		PsfEstimator::resetPool(1.25, 1, 0.5, 0.25, 32, 3);
		cpso = PsfEstimator::getCPSO(object_hamming, artificial_image);
		cpso->setOriginalPsf(psf);
		cpso->run(trackers, 396);
		// Get the current final cost.
		curr_cost = cpso->getGBestCost();

#ifdef _TRACK_TIME_
		tracker_estimation.end();
		all_costs[i] = curr_cost;
		all_times[i] = tracker_estimation.elapsedTime();
		all_stable_convergence[i] = cpso->getStableCycle();
		tracker_estimation.print();
#endif

		// Print the coefs.
		cout << "Validation#" << i << endl << " Coefs used to create the artificial image: [";
		cout << coefs[0];
		for(j=1; j < n_zernikes; j++) cout << ", " << coefs[j];
		cout << "]" << endl;
		cpso->getBestCoefs(coefs);
		cout << " Best coefs found during the CPSO search :  [";
		cout << coefs[0];
		for(j=1; j < n_zernikes; j++) cout << ", " << coefs[j];
		cout << "]" << endl;
		cout << "Stabilization at cycle: " << cpso->getStableCycle() << endl;

		// Check if it's a good value.
		for(best_pos=-1, j=0; j < rank_size; j++){
			if(curr_cost < best_costs[j]){
				if(j < rank_size-1) memcpy(&best_costs[j+1], &best_costs[j], (rank_size-j-1)*SIZEOF_WORD);
				best_pos = j;
				break;
			}
		}
		// Check if it's a bad value.
		for(worst_pos=-1, j=0; j < rank_size; j++){
			if(curr_cost > worst_costs[j]){
				if(j < rank_size-1) memcpy(&worst_costs[j+1], &worst_costs[j], (rank_size-j-1)*SIZEOF_WORD);
				worst_pos = j;
				break;
			}
		}
		// Only read from device if it's a good or a bad value.
		if(best_pos != -1 || worst_pos != -1){
			if(best_pos != -1){
				best_costs[best_pos] = curr_cost;
				memcpy(&best_artificial_psf[image_area * best_pos], psf, image_area * SIZEOF_WORD);
				memcpy(&best_artificial_psf_fft[image_area * best_pos], psf_fft, image_area * SIZEOF_FFTTYPE);
				memcpy(&best_artificial_phase[size_fft * best_pos], phase, size_fft * SIZEOF_WORD);
				memcpy(&best_artificial_image[image_area * best_pos], image, image_area * SIZEOF_WORD);
				cpso->getImageFFT(&best_artificial_image_fft[image_area * best_pos]);
				cpso->getBestCoefs(&best_coefs[n_zernikes * best_pos]);
				cpso->getBestPsfe(&best_psf[image_area * best_pos]);
				cpso->getBestPsfeFFT(&best_psf_fft[image_area * best_pos]);
				cpso->getBestPhase(&best_phase[size_fft * best_pos]);
				cpso->getBestConvolvedObject(&best_image[image_area * best_pos]);
				cpso->getBestConvolvedObjectFFT(&best_image_fft[image_area * best_pos]);
			}
			// A value could be both good and bad at startup.
			if(worst_pos != -1){
				worst_costs[worst_pos] = curr_cost;
				memcpy(&worst_artificial_psf[image_area * worst_pos], psf, image_area * SIZEOF_WORD);
				memcpy(&worst_artificial_psf_fft[image_area * worst_pos], psf_fft, image_area * SIZEOF_FFTTYPE);
				memcpy(&worst_artificial_phase[size_fft * worst_pos], phase, size_fft * SIZEOF_WORD);
				memcpy(&worst_artificial_image[image_area * worst_pos], image, image_area * SIZEOF_WORD);
				cpso->getImageFFT(&worst_artificial_image_fft[image_area * worst_pos]);
				cpso->getBestCoefs(&worst_coefs[n_zernikes * worst_pos]);
				cpso->getBestPsfe(&worst_psf[image_area * worst_pos]);
				cpso->getBestPsfeFFT(&worst_psf_fft[image_area * worst_pos]);
				cpso->getBestPhase(&worst_phase[size_fft * worst_pos]);
				cpso->getBestConvolvedObject(&worst_image[image_area * worst_pos]);
				cpso->getBestConvolvedObjectFFT(&worst_image_fft[image_area * worst_pos]);
			}
			// The calculation of the image difference MUST be made here because it will overwrite some device data.
			if(best_pos != -1){
				cpso->calcDifference(trackers, &best_image[image_area * best_pos], &best_imgdiff[image_area * best_pos]);
			}
			if(worst_pos != -1){
				cpso->calcDifference(trackers, &worst_image[image_area * worst_pos], &worst_imgdiff[image_area * worst_pos]);
			}
		}
		cpso->release();
		// Return the pool configuration for only 1 particle.
		PsfEstimator::resetPool(1.25, 1, 0.5, 0.25, 1, 1);
	}

	WORD psf_inv[image_area];
	FFT_TYPE fft_psf_inv[image_area];

	// Save the results.
	ostringstream *oss = new ostringstream();
	for(i=0; i < rank_size; i++){
		// Save the Best values to file.
		*oss << object_filename << "-best-" << i << "-image-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &best_artificial_image[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-psf-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &best_artificial_psf[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-phase-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &best_artificial_phase[image_area * i], phase_size, phase_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-image-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &best_image[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-image-difference.fits";
		write_fits_file((char*)oss->str().c_str(), &best_imgdiff[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-psf-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &best_psf[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-phase-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &best_phase[image_area * i], phase_size, phase_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-coefs-calculated.txt";
		save_coefs((char*)oss->str().c_str(), &best_coefs[image_area * i], n_zernikes);
		oss->str("");

		*oss << object_filename << "-best-" << i << "-psfinv-artificial.fits";
		CPSO::invertPsf(&best_artificial_psf[image_area * i], psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-psfinv-calculated.fits";
		CPSO::invertPsf(&best_psf[image_area * i], psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-image-artificial-fft.fits";
		CPSO::invertPsf(&best_artificial_image_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-image-calculated-fft.fits";
		CPSO::invertPsf(&best_image_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-psf-artificial-fft.fits";
		CPSO::invertPsf(&best_artificial_psf_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-best-" << i << "-psf-calculated-fft.fits";
		CPSO::invertPsf(&best_psf_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");

		// Save the Worst values to file.
		*oss << object_filename << "-worst-" << i << "-image-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_artificial_image[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-artificial-fft.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_artificial_image_fft[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psf-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_artificial_psf[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psf-artificial-fft.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_artificial_psf_fft[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-phase-artificial.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_artificial_phase[image_area * i], phase_size, phase_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_image[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-calculated-fft.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_image_fft[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-difference.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_imgdiff[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psf-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_psf[image_area * i], image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-phase-calculated.fits";
		write_fits_file((char*)oss->str().c_str(), &worst_phase[image_area * i], phase_size, phase_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-coefs-calculated.txt";
		save_coefs((char*)oss->str().c_str(), &worst_coefs[image_area * i], n_zernikes);
		oss->str("");

		*oss << object_filename << "-worst-" << i << "-psfinv-artificial.fits";
		CPSO::invertPsf(&worst_artificial_psf[image_area * i], psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psfinv-calculated.fits";
		CPSO::invertPsf(&worst_psf[image_area * i], psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-artificial-fft.fits";
		CPSO::invertPsf(&worst_artificial_image_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-image-calculated-fft.fits";
		CPSO::invertPsf(&worst_image_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psf-artificial-fft.fits";
		CPSO::invertPsf(&worst_artificial_psf_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
		*oss << object_filename << "-worst-" << i << "-psf-calculated-fft.fits";
		CPSO::invertPsf(&worst_psf_fft[image_area * i], fft_psf_inv, image_size);
		write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
		oss->str("");
	}
	// Save the Object's FFT.
	*oss << object_filename << "-object-fft.fits";
	CPSO::invertPsf(object_fft, fft_psf_inv, image_size);
	write_fits_file((char*)oss->str().c_str(), fft_psf_inv, image_size, image_size);
	oss->str("");

	delete object_fft;
	delete artificial_image;
	delete oss;
	delete psf;
	delete psf_fft;
	delete phase;
	delete image;
	delete image_fft;
	delete object_hamming;
	delete coefs;
	delete best_artificial_psf;
	delete best_artificial_psf_fft;
	delete best_artificial_phase;
	delete best_artificial_image;
	delete best_artificial_image_fft;
	delete best_coefs;
	delete best_psf;
	delete best_psf_fft;
	delete best_phase;
	delete best_image;
	delete best_image_fft;
	delete best_imgdiff;
	delete worst_artificial_psf;
	delete worst_artificial_psf_fft;
	delete worst_artificial_phase;
	delete worst_artificial_image;
	delete worst_artificial_image_fft;
	delete worst_coefs;
	delete worst_psf;
	delete worst_psf_fft;
	delete worst_phase;
	delete worst_image;
	delete worst_image_fft;
	delete worst_imgdiff;

	TRACK(trackers[0]->pause())

#ifdef _TRACK_TIME_
	// Print tracking information.
	cout << "Time spent per piece of code:" << endl;
	for (int i = 0; i < num_trackers; i++) {
		trackers[i]->end();
		trackers[i]->print();
		delete trackers[i];
	}
	delete[] trackers;

	cout << "* Statistics for Estimations:" << endl;
	cout << "- Time: avg.time=" << CPSO::calc_mean(all_times, n_validations) <<
			", std.dev.time=" << CPSO::calc_stddev(all_times, n_validations) <<
			", variance.time=" << CPSO::calc_variance(all_times, n_validations) << endl;
	cout << "- Fitness: avg.fitness=" << CPSO::calc_mean(all_costs, n_validations) <<
			", std.dev.fitness=" << CPSO::calc_stddev(all_costs, n_validations) <<
			", variance.fitness=" << CPSO::calc_variance(all_costs, n_validations) << endl;
	cout << "- Stabilization: avg.cycle=" << CPSO::calc_mean(all_stable_convergence, n_validations) <<
			", std.dev.cycle=" << CPSO::calc_stddev(all_stable_convergence, n_validations) <<
			", variance.cycle=" << CPSO::calc_variance(all_stable_convergence, n_validations) << endl;
#endif
}

/**
 * @brief Generate artificial random images.
 *
 * @param object The Object data that will be used for the validation.
 * @param zernikes The complete Zernike matrix.
 * @param diffraction_mask The diffraction mask.
 * @param n_zernikes The number of Zernike terms that is being used for the calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 * @param phase_mask The phase mask.
 * @param object_filename The Object file name, to be used as base to save the results.
 * @param n_images The number of artificial images to be generated.
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 * @param poisson The poisson noise level to be added to the artificial image.
 * @param parallel The number of threads to be used for the initial validations.
 */
void generate_random_images(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, int n_images, vector<int> *_devices, int poisson, int parallel){

	TimeTracker **trackers = NULL;
#ifdef _TRACK_TIME_
	// Create the trackers.
	int num_trackers = 6;
	trackers = new TimeTracker*[num_trackers];
	trackers[0] = new TimeTracker("cpso");
	trackers[1] = new TimeTracker("cycles");
	trackers[2] = new TimeTracker("generatePhase");
	trackers[3] = new TimeTracker("fft");
	trackers[4] = new TimeTracker("matrix-operations");
	trackers[5] = new TimeTracker("reduction");
#endif

	TRACK(trackers[0]->resume())
	cout << "Starting generation..." << endl;

	// CPSO Startup.
	PsfEstimator::startup(parallel, zernikes, phase_mask, diffraction_mask, phase_size,  image_size, n_zernikes, range, 1.25, 1, 0.5, 0.25, 1, 1, _devices);

	int size_fft = phase_size * phase_size;
	int image_area = image_size * image_size;
	double *artificial_image = new double[image_area * n_images];
	WORD *coefs = new WORD[n_zernikes * n_images];
	CPSO *cpso = NULL;
	// Validate the PSF reconstruction algorithm.
	omp_set_num_threads(parallel);
#pragma omp parallel for shared(trackers, object, artificial_image, coefs) \
		firstprivate(n_images, n_zernikes, image_size, image_area, size_fft, range, cpso, poisson) \
		default(none)
	for(int i=0; i < n_images; i++){
		WORD *_coefs = new WORD[n_zernikes];
		WORD *conobj = new WORD[image_area];

		CPSO::generateNormalDistrRandomCoefs(_coefs, n_zernikes, range);
		cpso = PsfEstimator::getCPSO(object, object, _coefs);
		// Calculates 1 random PSF and the artificial corrupted image.
		cpso->runPsf(trackers, 1);
		// Save the results.
		cpso->saveFirstResult();
		cpso->getBestConvolvedObject(conobj); // Artificial Corrupted Image.
		cpso->release();

		// Add the Poisson Noise to the artificial image.
		if(poisson){
			addPoissonNoise(conobj, image_size, image_size, poisson, 0);
		}

		for(int j=0; j < size_fft; j++) artificial_image[(i*image_area) + j] = conobj[j];
		for(int j=0; j < n_zernikes; j++) coefs[(i*n_zernikes) + j] = _coefs[j];

		delete _coefs;
		delete conobj;
	}

	ostringstream oss;
	oss << object_filename << "-generated-" << n_images << "-artificial-images.fits";
	writeFitsFile((char*)oss.str().c_str(), artificial_image, image_size, image_size, n_images);
	oss.str("");
	oss << object_filename << "-generated-" << n_images << "-coefs-calculated.txt";
	save_coefs((char*)oss.str().c_str(), coefs, n_zernikes, n_images);

	// The Total Cost must be something really small (something between 0 and 2).
	cout << "Generation finished." << endl;

	delete artificial_image;
	delete coefs;
}

/**
 * @brief Transform Cartesian coordinates x and y to polar theta.
 *
 * @param x The x position.
 * @param y The y position.
 * @return The theta value.
 */
double rect_to_theta(double x, double y) {
	double theta;

	if (x == 0) {
		if (y == 0) {
			theta = 0;
		}
		else if (y > 0) {
			theta = pi * 0.5;
		}
		else {
			theta = 1.5 * pi;
		}
	}
	else if(x < 0){
		if (y == 0) {
			theta = pi;
		}
		else if (y > 0) {
			theta = pi - atan(-y / x);
		}
		else {
			theta = pi + atan(y / x);
		}
	}
	else if (x > 0 && y > 0) {
		theta = atan(y / x);
	}
	else {
		theta = 2.0 * pi - atan(-y / x);
	}

	return theta;
}

/**
 * @brief Generate a 2d array that contains normalized
 * radial values within a circle around the center.
 *
 * @param array The array.
 * @param radius The radius of the circle.
 * @param size_x The size of the array.
 */
void gen_radial_values(double* array, double radius, int size_x) {
	double rr, x, y;
	double hhsize = size_x * 0.5;
	for (int i, j = 0; j < size_x; j++) {
		y = j - hhsize;
		for (i = 0; i < size_x; i++) {
			x = i - hhsize;
			rr = rect_to_r(x, y) / radius;
			if (rr <= 1) {
				array[j * size_x + i] = rr;
			}
			else {
				array[j * size_x + i] = 0;
			}
		}
	}
}

/**
 * @brief Generate a 2d array that contains polar theta values.
 *
 * @param array The array.
 * @param size_x The size of the array.
 */
void gen_angular_values(double* array, int size_x){
	double x, y;
	double hsize = size_x * 0.5;
	for (int i, j = 0; j < size_x; j++) {
		y = j - hsize;
		for (i = 0; i < size_x; i++) {
			x = i - hsize;
			array[j * size_x + i] = rect_to_theta(x, y);
		}
	}
}

/**
 * @brief Compute factorial of a number.
 *
 * @param n The number to be factored.
 */
double factorial_to(int n) {
	int fact = 1;
	for(; n > 1; n--) fact *= n;
	return fact;
}

/**
 * @brief Compute radial part of a zernike polynomial.
 *
 * @param r The radius.
 * @param n The zernike order n.
 * @param m The zernike order m.
 */
double zernike_radial(double r, int n, int m){
	long double calc1, calc2, calc3;
	long double result = 0;
	int upper = ((n - m) / 2)  + 1;
	for (int i = 0; i < upper; i++) {
		calc1 = pow(-1, i) * factorial_to(n - i);
		calc2 = factorial_to(i) * factorial_to(((n + m) / 2) - i) * factorial_to(((n - m) / 2) - i);
		calc3 = pow(r, n - (2.0 * i));
		result += calc1 * calc3 / calc2;
	}
	return result;
}

/**
 * @brief Compute angular part of a zernike polynomial.
 *
 * @param theta The angle.
 * @param n The zernike order n.
 * @param m The zernike order m
 * @param evenOdd Even or odd zernike (if positive or negative order m).
 */
double zernike_angular(double theta, int n, int m, int evenOdd) {
	double result;
	if (m == 0) {
		result = sqrt(n + 1);
	}
	else if (evenOdd == 0) {
		result = sqrt(2.0) * sqrt(n + 1) * cos(m * theta);
	}
	else {
		result = sqrt(2.0) * sqrt(n + 1) * sin(m * theta);
	}
	return result;
}

/**
 * @brief Generates a 3d data cube with zernike polynomials.
 *
 * @param zernikes The data cube.
 * @param rarray The array containing radial values.
 * @param tarray The array containing angular (theta) values.
 * @param size_x The array size.
 * @param nz The number of Zernikes contained in the cube.
 */
void gen_zernike_values(double *zernikes, double* rarray, double* tarray, int size_x, int nz) {
	double rad, ang;
	int evenOdd, n = 0, m = 0, size_fft = size_x * size_x;
	for (int k, j, i = 0; i < nz; i++) {
		translate_noll_index(i + 4, &n, &m);
		if (m < 0) {
			m = -m;
			evenOdd = 1;
		}
		else{
			evenOdd = 0;
		}
		for (k = 0; k < size_x; k++) {
			for (j = 0; j < size_x; j++) {
				rad = zernike_radial(rarray[k * size_x + j], n, m);
				ang = zernike_angular(tarray[k * size_x + j], n, m, evenOdd);
				zernikes[(i * size_fft) + (k * size_x + j)] = rad * ang;
			}
		}
	}
}

/**
 * @brief Get Zernike orders n and m from noll-index j.
 *
 * This code was dapted from a C++ - routine written by D.Schmidt, therefore:
 * Copyright (c), Dirk Schmidt <Dirk.Schmidt@kis.uni-freiburg.de>
 *
 * @param j The noll index.
 * @param n The zernike order n.
 * @param m The zernike order m.
 */
void translate_noll_index(int j, int *n, int *m) {
	if (j < 1) {
		fprintf(stderr, "ERROR: j=%i but it must be j>=1.\n", j);
		*n = -1;
		return;
	}

	*n = 0;
	*m = 0;

	if (j == 1) {
		return;
	}

	int l, k = j; // some helpers

	// for each n, there are n+1 possible values for m.
	// so what is n for the given j?
	for (l = 0; k > 0; l++) {
		k -= l;
	}
	*n = l - 2; // -2 here and not -1 because l++
	// is also computed in the 'for' loop at the very end.

	// what is the smallest j for this n?
	k = 1;
	for (l = 1; l < *n; l++) {
		k += (l + 1);
	}

	// so how many m values must be missed out then?
	k = j - k;
	// even n start with m=0, uneven n start with m=1
	if (*n % 2 == 0) {
		*m = 0;
	}
	else {
		*m = 1;
	}

	// increase m by 2 in every second step of k-1 overall steps
	// (-1, because the first step is already
	// done with setting the starting m)
	// (every second step, because in between
	// the sign is changed, what's to be considered
	// after this loop)
	for (l = 1; l < k; l++) {
		if (l % 2 == 0 && *n % 2 != 0) {
			*m += 2;
		}
		else if (l % 2 != 0 && *n % 2 == 0) {
			*m += 2;
		}
	}

	// negative sign for uneven j
	if (j % 2 != 0) {
		*m = -*m;
	}
}

/**
 * @brief Find min and max of a 2d array.
 *
 * @param array The array.
 * @param min The minimum.
 * @param max The maximum.
 * @param size The size of the array.
 */
void min_max_array(double *array, double *min, double *max, int size) {
	*max = *array;
	*min = *array;
	for (int i, j = 0; j < size; j++) {
		for (i = 0; i < size; i++) {
			if (array[j * size + i] > *max){
				*max = array[j * size + i];
			}
			if (array[j * size + i] < *min){
				*min = array[j * size + i];
			}
		}
	}
}

/**
 * @brief Normalize Zernikes (scale to -1 < z < 1).
 *
 * @param zernikes The Zernike data cube.
 * @param size The size of a Zernike.
 * @param nz The number of Zernikes.
 */
void norm_zernikes(double *zernikes, int size, int nz) {
	double min, max, scale, sub;
	int size_fft = size * size;
	for (int k, j, i = 0; i < nz; i++) {
		min_max_array(&zernikes[i * size_fft], &min, &max, size);
		sub = (max + min) * 0.5;
		scale = (max - min) * 0.5;
		for (k = 0; k < size; k++) {
			for (j = 0; j < size; j++) {
				zernikes[(i * size_fft) + (k * size + j)] = (zernikes[(i * size_fft) + (k * size + j)] - sub) / scale;
			}
		}
	}
}

/**
 * @brief Generate a circular mask with radius specified, centered at (x, y).
 *
 * @param mask The mask.
 * @param radius The radius.
 * @param x The center in x.
 * @param y The center in y.
 * @param size The size of the mask.
 */
void gen_circ_mask(int * mask, double radius, double x, double y, int size) {
	double dist;
	for (int i, j = 0; j < size; j++) {
		for (i = 0; i < size; i++) {
			dist = sqrt( pow(x - i, 2) + pow(y - j, 2) );
			mask[j * size + i] = (dist <= radius);
		}
	}
}

/**
 * @brief Generate a mask:
 * 0 if the corresponding frequency is beyond diff.-limit;
 * 1 otherwise.
 *
 * @param mask The mask.
 * @param freq_cut The cutting frequency (pixel).
 * @param size_x The size in x (pixel).
 * @param size_y The size in y (pixel).
 */
void gen_freq_mask(int* mask, double freq_cut, int size_x, int size_y) {
	double dist, y;
	for (int x, j = 0; j < size_y; j++) {
		y = (j > size_y / 2) ? size_y - j : j;
		for (x = 0; x < size_x; x++) {
			dist = sqrt(pow(x, 2) + pow(y, 2));
			mask[j * size_x + x] = (dist <= freq_cut);
		}
	}
}

long read_fits_file(char* filename, double* image, int sx, int sy) {
	return read_fits_file(filename, image, sx * sy);
}

long read_fits_file(char* filename, double* image, int size) {
	return read_fits_file(filename, NULL, true, image, size, 1);
}

long read_fits_file(char* filename, fitsfile *fptr, double* image, int sx, int sy, long firstByte) {
	return read_fits_file(filename, fptr, image, sx * sy, firstByte);
}

long read_fits_file(char* filename, fitsfile *fptr, double* image, int size, long firstByte) {
	return read_fits_file(filename, fptr, false, image, size, firstByte);
}

long read_fits_file(char* filename, fitsfile *fptr, bool closeFits, double* image, int size, long firstByte) {
	int status = 0;
	if (fptr == NULL && fits_open_file(&fptr, filename, READONLY, &status)){
		printerror(status);
		return NULL;
	}

	int anynull;
	double nullval = 0;
	fits_read_img(fptr, TDOUBLE, firstByte, size, &nullval, image, &anynull, &status);

	if ((status || closeFits)){
		if (fits_close_file(fptr, &status)){
			printerror(status);
		}
		return -1;
	}

	return firstByte;
}

/**
 * @brief Get the number of images available in the FITS file.
 *
 * If the file is 2D, it always returns 1.
 *
 * @param filename The FITS file name.
 * @return The number of images available in the FITS file.
 */
int fitsSize(char* filename) {
	fitsfile *fptr;
	int status = 0;
	if (fits_open_file(&fptr, filename, READONLY, &status)){
		printerror(status);
		exit(1);
	}

	int size;
	fits_read_key(fptr, TLONG, "NAXIS3", &size, NULL, &status);
	// A 2D file doesn't contain the key NAXIS3.
	if(status){
		size = 1;
		status = 0;
	}

	if (fits_close_file(fptr, &status)){
		printerror(status);
		exit(1);
	}

	return size;
}

/**
 * @brief Write the image the a FITS file.
 *
 * It converts the data to double, if necessary.
 *
 * @param filename The FITS file name to save the results.
 * @param array The image data.
 * @param sx The width of the image.
 * @param sy The height of the image.
 */
void writeFitsFile(char* filename, WORD* array, int sx, int sy) {
	// Ensure that the output file data type is always double, regardless the data type currently in use.
	int size = sx * sy;
	double _array[size];
	// Convert the data from float to double, if necessary.
	for(int i=0; i < size; i++) _array[i] = array[i];

	writeFitsFile(filename, _array, sx, sy, 1);
}

/**
 * @brief Write the image the a FITS file.
 *
 * @param filename The FITS file name to save the results.
 * @param array The image data.
 * @param sx The width of the image.
 * @param sy The height of the image.
 * @param sz The number of images to be saved. If this is a 2D image, 1 must be provided.
 */
void writeFitsFile(char* filename, double* array, int sx, int sy, int sz) {
	remove(filename);

	int status = 0; // initialize status before calling fitsio routines

	// create new FITS files
	fitsfile *fptr;
	if (fits_create_file(&fptr, filename, &status)){
		printerror(status);
		exit(1);
	}

    int  bitpix   =  -64;
    long naxis = (sz > 1) ? 3 : 2; // Specify the dimension of the image.
	long naxes[naxis];
	naxes[0] = sx; naxes[1] = sy;
    if(sz > 1){
    	// Set the number of images in a 3D file.
    	naxes[2] = sz;
    }

    // Create an entry to the image in the FITS file.
    if (fits_create_img(fptr, bitpix, naxis, naxes, &status)){
		printerror(status);
	}

	// write the array to the FITS file
	long size = sx * sy * sz;
	if (fits_write_img(fptr, TDOUBLE, 1, size, array, &status)){
		printerror(status);
	}

	// close the file
	if (fits_close_file(fptr, &status)){
		printerror(status);
		exit(1);
	}
}


/**
 * @brief Print out cfitsio error messages and exit program
 */
void printerror(int status) {
	if (status) {
		fits_report_error(stderr, status); /* print error report */
		exit(status); /* terminate program, return error status */
	}
}

/**
 * @brief Write a FITS file with a FFT_TYPE array.
 * The FFT_TYPE array is converted into a WORD array before saved.
 *
 * @param filename The FITS file name where to save the data.
 * @param array The FFT_TYPE array.
 * @param sx The x size.
 * @param sy The y size.
 */
void write_fits_file(char* filename, FFT_TYPE* array, int sx, int sy) {
	int area = sx*sy;
	WORD wArray[area];
	for(int i=0; i < area; i++) wArray[i] = sqrt(array[i].x * array[i].x + array[i].y * array[i].y) / area;
	write_fits_file(filename, wArray, sx, sy);
}

/**
 * @brief Write a FITS file with a WORD array.
 *
 * @param filename The FITS file name where to save the data.
 * @param array The WORD array.
 * @param sx The x size.
 * @param sy The y size.
 */
void write_fits_file(char* filename, WORD* array, int sx, int sy) {
	remove(filename); // Delete old file if it already exists

	fitsfile *fptr; // pointer to the FITS file, defined in fitsio.h
	int status = 0; // initialize status before calling fitsio routines
	if (fits_create_file(&fptr, filename, &status)){ // create new FITS file
		printerror(status); // call printerror if error occurs
	}

    int  bitpix   =  -64;
    long naxis = 2;  // 2-dimensional image
	long naxes[2] = {sx, sy};
	if (fits_create_img(fptr, bitpix, naxis, naxes, &status)){
		printerror(status);
	}

	// Ensure that the output file data type is always double, regardless the data type currently in use.
	double *_array;
#ifdef _DOUBLE_WORD_
	_array = array;
#else
	int sz = sx * sy;
	_array = new double[sz];
	for(int i=0; i < sz; i++) _array[i] = array[i];
#endif

	// write the array to the FITS file
	long fpixel = 1; // first pixel to write
	long nelements = naxes[0] * naxes[1]; // number of pixels to write
	if (fits_write_img(fptr, TDOUBLE, fpixel, nelements, _array, &status)){
		printerror(status);
	}

#ifndef _DOUBLE_WORD_
	delete _array;
#endif

	// close the file
	if (fits_close_file(fptr, &status)){
		printerror(status);
	}
}

/**
 * @brief Sum a 3D FITS file into a 2D result.
 *
 * @param fitsManager The FitsManager pointing to the 3D file.
 * @param sum The double array which will contain the resulting sum.
 * @param img_area The image area.
 */
void sum_3dfits_file(FitsManager *fitsManager, double *sum, int img_area){
	fits_result *result;
	int j, n_images = fitsManager->size();
	for(j=0; j < img_area; j++)	sum[j] = 0;
	for(int i=0; i < n_images; i++){
		if((result = fitsManager->nextImage()) == NULL){
			continue;
		}
		for(j=0; j < img_area; j++){
			sum[j] += result->image[j];
		}
	}
}

void copy_image(double *image_from, double *image_to, int img_area){
	for(int i=0; i < img_area; i++){
		image_to[i] = image_from[i];
	}
}

/*
* Return a noise value from the specified distribution.
*/
int chooseNoiseFromTable(double *prob, int size) {
	int Low = 0;
	int Mid = (size - 1) / 2;
	int High = size - 1;
	int Random = 1 + rand() % (size - 1);

	while (Low < High) {
		Mid = (Low + High) / 2;
		if (prob[Mid] * size < Random)
			Low = Mid + 1;
		else if (prob[Mid] * size > Random)
			High = Mid - 1;
		else
			Low = High = Mid;
	}

	return Mid;
}

/**
 * @brief Add noise to a WORD image.
 * source: http://infocom.cheonan.ac.kr/~nykwak/kuim/noise.c
 *
 * @param image The WORD image.
 * @param img_size The iamge size.
 * @param prob_size The size of the probability table.
 * @param variance The variance.
 * @param mean The mean.
 */
void addPoissonNoise(WORD *image, int img_size, int prob_size, double variance, double mean) {
	double prob[prob_size];
	double Prob[prob_size];
	int img_area = img_size * img_size;

	/* Initialize probability arrays */
	prob[0] = exp(-variance);
	Prob[0] = 0;
	int i;
	for (i=1; i < prob_size; i++) {
		prob[i] = prob[i - 1] * variance / i;
		Prob[i] = Prob[i - 1] + (prob[i - 1] + prob[i]) / 2;
	}

	/* Add noise to each point */
	for (i=0; i < img_area; i++){
		image[i] += chooseNoiseFromTable(Prob, prob_size) - variance + mean;
	}
}

/**
 * @brief Save the coefficients to a text file.
 *
 * @param file_name The file name where to save the coefficients.
 * @param coefs The coefficients pointer.
 * @param n_coefs The number of coefficients to be saved.
 */
void save_coefs(char *file_name, WORD *coefs, int n_coefs){
	save_coefs(file_name, coefs, n_coefs, 1);
}

/**
 * @brief Save multiple coefficient sets to a text file.
 *
 * @param file_name The file name where to save the coefficients.
 * @param coefs The coefficients pointer.
 * @param n_coefs The number of coefficients to be saved.
 * @param n_rows The number of coefficients sets to be saved.
 */
void save_coefs(char *file_name, WORD *coefs, int n_coefs, int n_rows){
	if(coefs == NULL || n_coefs < 1 || n_rows < 1 || file_name == NULL) return;

	FILE *fp = fopen(file_name, "w");
	if (!fp) {
		cerr << "Failed to save coefs [" << file_name << "]." << std::endl;
		return;
	}

	char val[100];
	val[0] = '\0';
	for(int j, i=0; i < n_rows; i++){
		ftoa(coefs[i * n_coefs], val);
		fwrite(val, 1, strlen(val), fp);
		for(j=1; j < n_coefs; j++){
			fwrite(", ", 1, 2, fp);
			ftoa(coefs[i * n_coefs + j], val);
			fwrite(val, 1, strlen(val), fp);
		}
		fwrite("\n", 1, 1, fp);
	}

	fclose(fp);
}

char *ftoa(float val, char *buffer){
	sprintf(buffer, "%f", val);
	return buffer;
}

/**
 * @brief Generate a Hamming window.
 *
 * @param image A pointer where the window will be created. It must be previously allocated.
 * @param img_area The image area.
 */
void hamming(double *image, int img_area){
	for(int i=0; i < img_area; i++){
	    image[i] = 0.53836 + 0.46164 * cos((2 * pi * i) / (img_area - 1));
	}
}

/**
 * @brief Generate a Hanning window.
 *
 * @param image A pointer where the window will be created. It must be previously allocated.
 * @param img_area The image area.
 */
void hanning(double *image, int img_area){
	for(int i=0; i < img_area; i++){
	    image[i] = 0.5 - 0.5 * cos((2 * pi * i) / (img_area));
	}
}

/**
 * @brief Apply a window over an image.
 *
 * @param img The Image data to have the window applied.
 * @param window The window data.
 * @param img_area The image area.
 */
void apply_window(double *img, double *window, int img_area){
	for(int i=0; i < img_area; i++) img[i] *= window[i];
}

/**
 * @brief Generates a fractional Hamming window for Knox-Thompson solar speckle imaging.
 *
 * This method is based on the IDL function frac_hamming(nx,ny,frac,Eqa).
 *
 * @param window A previously allocated double array that will contain the window data.
 * @param frac The Hamming parameter.
 * @param nx The x dimension of the window.
 * @param ny The y dimension of the window.
 */
void frac_hamming(double *window, float frac, int nx, int ny){
	int nxf = 2 * (int)((frac*nx+1)/2);
	if(nxf <= 0){
		for(int i=0; i < nx*ny; i++) window[i] = 1.0;
	}
	else{
		// Calculate the window's row.
		int nxfh = nxf / 2;
		double Fx1[nxf];
		hanning(Fx1, nxf); // defines Hamming window.
		double Fx[nx];
		for(int i=0; i < nx; i++) Fx[i] = 1.0;
		for(int i=0; i < nxfh; i++) Fx[i] = Fx1[i];
		for(int i=nxfh, j=nx-nxfh; i < nxf; i++, j++) Fx[j] = Fx1[i];

		// Calculate the window's column.
		int nyf = 2 * (int)((frac*ny+1)/2);
		int nyfh = nyf / 2;
		double Fy1[nyf];
		hanning(Fy1, nyf); // defines Hamming window.
		double Fy[ny];
		for(int i=0; i < ny; i++) Fy[i] = 1.0;
		for(int i=0; i < nyfh; i++) Fy[i] = Fy1[i];
		for(int i=nyfh, j=ny-nyfh; i < nyf; i++, j++) Fy[j] = Fy1[i];

		// Generate the fractional window.
		for(int x, y=0; y < ny; y++){
			for(x=0; x < nx; x++){
				window[(y*nx)+x] = Fx[x] * Fy[y];
			}
		}
	}
}

/**
 * @brief Test all GPUs available by sending a buffer of 10K WORDs to every GPU
 * and then making a reduction of the values sent.
 *
 * If the resulting reduction is the expected WORD value 1.0, the test is considered as Passed.
 * Otherwise, the test failed.
 */
void test_gpu(){
	clFactory::startup();
	char deviceName[1024];

	int i, sz = 10*1024;
	WORD chk_sum, arr[sz];
	for(i=0; i < sz; i++) arr[i] = 1.0/sz;

	cl_mem cl_values, cl_sum;
	cl_command_queue command_queue;
	cl_context context;
	kernel_set *kernels;

	int n_devices = clFactory::getNumAvailableDevices();
	clQueue *queue[n_devices];
	for(i=0; i < n_devices; i++){
		queue[i] = clFactory::getQueue();
		command_queue = queue[i]->getCommandQueue();
		context = queue[i]->getContext();
		kernels = queue[i]->kernels;

		clFactory::getDeviceName(queue[i], deviceName, 1024);
		printf("Validating Device [%s - Ptr %p]...\n", deviceName, queue[i]->getDevice());

		CREATE_BUFFER(context, CL_MEM_WRITE_ONLY | CL_MEM_COPY_HOST_PTR, sz * SIZEOF_WORD, arr, cl_values);
		CREATE_BUFFER(context, CL_MEM_READ_WRITE, REDUCTION_NBLOCKS * SIZEOF_WORD, NULL, cl_sum);

		CALL_KERNEL(command_queue, kernels->cl_reduce, REDUCTION_NBLOCKS*REDUCTION_BLOCKSZ, REDUCTION_BLOCKSZ, 4,
			{sizeof(cl_mem), (void*)&cl_values},
			{sizeof(cl_mem), (void*)&cl_sum},
			{REDUCTION_BLOCKSZ * SIZEOF_WORD, NULL},
			{sizeof(cl_int), (void*)&sz}
		);
		SYNC_QUEUE
		FINAL_REDUCTIONS(command_queue,  cl_sum, chk_sum, 1);
		chk_sum = ((int)(chk_sum*1000000))/1000000.0f; // Remove the trailing decimals.

		if(chk_sum == 1.0f){
			printf("Device Ok.\n");
		}
		else{
			printf("Validation failed on Device [calculated %f instead of 1].\n", chk_sum);
		}

		clReleaseMemObject(cl_values);
		clReleaseMemObject(cl_sum);
	}
	for(i=0; i < n_devices; i++){
		clFactory::disposeQueue(queue[i]);
	}
	clFactory::shutdown();
}

/**
 * @brief List all devices available.
 */
void list_gpu(){
	clFactory::startup();

	char deviceName[1024];
	int n_devices = clFactory::getNumAvailableDevices();
	clQueue *queue[n_devices];
	int i;
	for(i=0; i < n_devices; i++){
		queue[i] = clFactory::getQueue();
		clFactory::getDeviceName(queue[i], deviceName, 1024);
		printf("[%i] %s\n", i+1, deviceName);
	}

	for(i=0; i < n_devices; i++){
		clFactory::disposeQueue(queue[i]);
	}
	clFactory::shutdown();
}

/**
 * @brief Calculate the Root Mean Square contrast from an image.
 *
 * @param img the image to calculate the RMS contrast.
 * @param width the iamge width.
 * @return the value representing the RMS contrast.
 */
double rmsContrast(double* img, int width) {
	int i, area = width * width;
	double rms = 0.0, avg = 0;
	double max = 0;

	for(i=0; i < area; i++){
		max = (img[i]>max)?img[i]:max;
		avg += img[i];
	}
	avg /= area;
	avg /= max; // Normalize the average.

	double aux;
	for (i=0; i < area; i++) {
		aux = (img[i] / max) - avg;
		rms += aux * aux;
	}

	return sqrt(rms / area);
}
