
Copyright (c) 2013, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * kippef.hpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#ifndef _KIPPE_H_
#define _KIPPE_H_

#include <math.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <complex>
#include <complex.h>
#include <errno.h>
#include <string.h>
#include <fitsio.h>
#include <time.h>
#include <omp.h>
#include "config.hpp"
#include "psf_estimator.hpp"
#include "cpso.hpp"
#include "TimeTracker.hpp"
#include "FitsManager.hpp"


/**
 * The PI value.
 */
#define pi 3.1415926535897932

/**
 * @brief Exit the program with code specified.
 *
 * The parameter error is: the error code to exit.
 */
#define exit_error(err) exit(err)

/**
 * @brief Transform Cartesian coordinates (x, y) to polar r.
 *
 * @param x The x coordinate.
 * @param y The y coordinate.
 */
#define rect_to_r(x, y) sqrt(pow(x, 2) + pow(y, 2))

/**
 * @brief Set the value of a 2d array at position [i, j].
 *
 * @param array The array.
 * @param value The value to be set into the array.
 * @param i The first dimension of the array.
 * @param j The second dimension of the array.
 * @param size_x The size of the first dimension of the array.
 */
#define set_two_d_array_value(array, value, i, j, size_x) array[j * size_x + i] = value;

/**
 * @brief Get the value of a 2d array at position [i, j].
 *
 * @param array The array.
 * @param i The first dimension of the array.
 * @param j The second dimension of the array.
 * @param size_x The size of the first dimension of the array.
 */
#define get_two_d_array_value(array, i, j, size_x) array[j * size_x + i]

/**
 * @brief Compute cutting frequency of a telescope (in pixel).
 *
 * @param tel_dia The diameter of the aperture.
 * @param img_scl The image scale (arcs / pixel).
 * @param wavelength The wavelength (m).
 * @param img_size The image size (pixel).
 */
#define cmp_freq_cut(tel_dia, img_scl, wavelength, img_size) ((pi * tel_dia * img_size * img_scl) / (3600.0 * 180.0 * wavelength))

/**
 * @brief Computes the radius of the aperture, given the size N of the phase array.
 *
 * @param tel_dia The telescope diameter (m).
 * @param image_scale The image scale (arcs/pix).
 * @param wavelength The wavelength (m).
 * @param N The size of the phase arrays (pixel).
 */
#define cmp_size_of_phase(tel_dia, image_scale, wavelength, N) ((int) round((tel_dia * image_scale * pi * N) / (180.0 * 3600.0 * wavelength)))

/**
 * @brief Copies a 1d array into another 1d array.
 *
 * This method avoid errors when copying different data types.
 *
 * @param in The input array.
 * @param out The output array.
 * @param size The size of the copy.
 */
#define copy_par(in, out, size) for (int _i = 0; _i < size; _i++) out[_i] = in[_i];

/**
 * @brief Copies a 1d array into another 1d array.
 *
 * @param in The input array.
 * @param scale The output array.
 * @param size The size of the copy.
 */
#define scale_par(in, scale, size) for (int _i = 0; _i < size; _i++) in[_i] = in[_i] * scale;

/**
 * @brief Reset the phase, setting its values to zero.
 *
 * @param phase The phase.
 * @param size The phase width.
 */
#define reset_phase(phase, size) memset(phase, 0, size * size * sizeof(double))

/**
 * @brief Minimize a cost function by using the search method CPSO (Cooperative Particle Swarm Optimization).
 *
 * @param cpso The CPSO instance.
 * @param result A pointer where all the best values found will be stored. This pointer must be obtained through the {link FitsManager} class.
 * @param n_cycles The number of cycles to run the CPSO.
 */
void run_cpso(CPSO *cpso, fits_result *result, int n_cycles);


/**
 * @brief Compare a PSF file against the Object file and the Image file,
 * by convolving the PSF with the Object and then comparing the result
 * with the Image.
 *
 * The output is the cost of the difference (convolved object - image)
 * and their mismatch (how many pixels are different between them).
 *
 * @param psfFileName The PSF file to be validated.
 * @param objFileName The Object file name.
 * @param imgFileName The Image file name.
 * @param zernikes The complete Zernike terms.
 * @param diffraction_mask The diffraction mask.
 * @param phase_mask The phase mask.
 * @param n_zernikes The number of Zernike terms that are being used on these calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 */
void validate_psf_file(char *psfFileName, char *objFileName, char *imgFileName, double* zernikes, int* diffraction_mask, int *phase_mask,
		int n_zernikes, int image_size, int phase_size, double range);


void nollTest(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, vector<int> *_devices);

/**
 * @brief Validate the PSF Estimation through CPSO technique.
 *
 * This method validate the technique by running 2 verifications:
 * <ul>
 * <li>Generating random PSFs / artificial Images and then setting the same coefficients to the PSF Estimation code.
 * This verifies if the code is able to duplicate the same results based on the same set of coefficients.
 * No search is executed on this validation.</li>
 * <li>Generating random PSFs / artificial Images and then submitting complete CPSO searches for these artificial images.
 * The best and worst values are saved to files at the end of the execution of this method.
 * The information saved is: Images, PSFs, phases and coefficients.
 * This verifies how good the estimation technique is fine tuned.</li>
 * </ul>
 *
 * @param object The Object data that will be used for the validation.
 * @param zernikes The complete Zernike matrix.
 * @param diffraction_mask The diffraction mask.
 * @param n_zernikes The number of Zernike terms that is being used for the calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 * @param phase_mask The phase mask.
 * @param object_filename The Object file name, to be used as base to save the results.
 * @param n_validations The number of validations to be done.
 * @param rank_size The number of best and good results to be kept.
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 * @param poisson The poisson noise level to be added to the artificial image.
 * @param parallel The number of threads to be used for the initial validations.
 */
void validate_psfcode(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, int n_validations, int rank_size, vector<int> *_devices,
		int poisson, int parallel);

/**
 * @brief Generate artificial random images.
 *
 * @param object The Object data that will be used for the validation.
 * @param zernikes The complete Zernike matrix.
 * @param diffraction_mask The diffraction mask.
 * @param n_zernikes The number of Zernike terms that is being used for the calculations.
 * @param image_size The Image width.
 * @param phase_size The phase width.
 * @param range The Zernike coefficient range.
 * @param phase_mask The phase mask.
 * @param object_filename The Object file name, to be used as base to save the results.
 * @param n_images The number of artificial images to be generated.
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 * @param poisson The poisson noise level to be added to the artificial image.
 * @param parallel The number of threads to be used for the initial validations.
 */
void generate_random_images(double* object, double* zernikes, int* diffraction_mask,
		int n_zernikes, int image_size, int phase_size, double range, int *phase_mask,
		char *object_filename, int n_images, vector<int> *_devices, int poisson, int parallel);

/**
 * @brief Transform Cartesian coordinates x and y to polar theta.
 *
 * @param x The x position.
 * @param y The y position.
 * @return The theta value.
 */
double rect_to_theta (double x, double y);

/**
 * @brief Generate a 2d array that contains normalized
 * radial values within a circle around the center.
 *
 * @param array The array.
 * @param radius The radius of the circle.
 * @param size_x The size of the array.
 */
void gen_radial_values(double* array, double radius, int size_x);

/**
 * @brief Generate a 2d array that contains polar theta values.
 *
 * @param array The array.
 * @param size_x The size of the array.
 */
void gen_angular_values(double* array, int size_x);

/**
 * @brief Compute factorial of a number.
 *
 * @param n The number to be factored.
 */
double factorial_to(int n);

/**
 * @brief Compute radial part of a zernike polynomial.
 *
 * @param r The radius.
 * @param n The zernike order n.
 * @param m The zernike order m.
 */
double zernike_radial(double r, int n, int m);

/**
 * @brief Compute angular part of a zernike polynomial.
 *
 * @param theta The angle.
 * @param n The zernike order n.
 * @param m The zernike order m
 * @param evenOdd Even or odd zernike (if positive or negative order m).
 */
double zernike_angular(double theta, int n, int m, int evenOdd);

/**
 * @brief Generates a 3d data cube with zernike polynomials.
 *
 * @param zernikes The data cube.
 * @param rarray The array containing radial values.
 * @param tarray The array containing angular (theta) values.
 * @param size_x The array size.
 * @param nz The number of Zernikes contained in the cube.
 */
void gen_zernike_values(double *zernikes, double* rarray, double* tarray, int size_x, int nz);

/**
 * @brief Get Zernike orders n and m from noll-index j.
 *
 * This code was dapted from a C++ - routine written by D.Schmidt, therefore:
 * Copyright (c), Dirk Schmidt <Dirk.Schmidt@kis.uni-freiburg.de>
 *
 * @param j The noll index.
 * @param n The zernike order n.
 * @param m The zernike order m.
 */
void translate_noll_index(int j, int *n, int *m);

/**
 * @brief Find min and max of a 2d array.
 *
 * @param array The array.
 * @param min The minimum.
 * @param max The maximum.
 * @param size The size of the array.
 */
void min_max_array(double *array, double *min, double *max, int size);

/**
 * @brief Normalize Zernikes (scale to -1 < z < 1).
 *
 * @param zernikes The Zernike data cube.
 * @param size The size of a Zernike.
 * @param nz The number of Zernikes.
 */
void norm_zernikes(double *zernikes, int size, int nz);

/**
 * @brief Generate a circular mask with radius specified, centered at (x, y).
 *
 * @param mask The mask.
 * @param radius The radius.
 * @param x The center in x.
 * @param y The center in y.
 * @param size The size of the mask.
 */
void gen_circ_mask(int * mask, double radius, double x, double y, int size);

/**
 * @brief Generate a mask:
 * 0 if the corresponding frequency is beyond diff.-limit;
 * 1 otherwise.
 *
 * @param mask The mask.
 * @param freq_cut The cutting frequency (pixel).
 * @param size_x The size in x (pixel).
 * @param size_y The size in y (pixel).
 */
void gen_freq_mask(int* mask, double freq_cut, int size_x, int size_y);

long read_fits_file(char* filename, double* image, int sx, int sy);
long read_fits_file(char* filename, double* image, int size);
long read_fits_file(char* filename, fitsfile *fptr, double* image, int sx, int sy, long firstByte);
long read_fits_file(char* filename, fitsfile *fptr, double* image, int size, long firstByte);
long read_fits_file(char* filename, fitsfile *fptr, bool closeFits, double* image, int size, long firstByte);

/**
 * @brief Get the number of images available in the FITS file.
 *
 * If the file is 2D, it always returns 1.
 *
 * @param filename The FITS file name.
 * @return The number of images available in the FITS file.
 */
int fitsSize(char* filename);

/**
 * @brief Print out cfitsio error messages and exit program
 */
void printerror(int status);

void writeFitsFile(char* filename, WORD* array, int sx, int sy);

void writeFitsFile(char* filename, double* array, int sx, int sy, int sz);

/**
 * Write a FITS file with a FFT_TYPE array.
 * The FFT_TYPE array is converted into a WORD array before saved.
 *
 * @param filename The FITS file name where to save the data.
 * @param array The FFT_TYPE array.
 * @param sx The x size.
 * @param sy The y size.
 */
void write_fits_file(char* filename, FFT_TYPE* array, int sx, int sy);

/**
 * @brief Write a FITS file with a WORD array.
 *
 * @param filename The FITS file name where to save the data.
 * @param array The WORD array.
 * @param sx The x size.
 * @param sy The y size.
 */
void write_fits_file(char* filename, WORD* array, int sx, int sy);

/**
 * @brief Sum a 3D FITS file into a 2D result.
 *
 * @param fitsManager The FitsManager pointing to the 3D file.
 * @param sum The double array which will contain the resulting sum.
 * @param img_area The image area.
 */
void sum_3dfits_file(FitsManager *fitsManager, double *sum, int img_area);

void copy_image(double *image_from, double *image_to, int img_area);

double calc_mean(double *image, int img_area);

double calc_variance(double *image, int img_area);

double calc_stddev(double *image, int img_area);

int chooseNoiseFromTable(double *prob, int size);

/**
 * @brief Add noise to a WORD image.
 * source: http://infocom.cheonan.ac.kr/~nykwak/kuim/noise.c
 *
 * @param image The WORD image.
 * @param img_size The iamge size.
 * @param prob_size The size of the probability table.
 * @param variance The variance.
 * @param mean The mean.
 */
void addPoissonNoise(WORD *image, int img_size, int prob_size, double variance, double mean);

/**
 * @brief Save the coefficients to a text file.
 *
 * @param file_name The file name where to save the coefficients.
 * @param coefs The coefficients pointer.
 * @param n_coefs The number of coefficients to be saved.
 */
void save_coefs(char *file_name, WORD *coefs, int n_coefs);

/**
 * @brief Save multiple coefficient sets to a text file.
 *
 * @param file_name The file name where to save the coefficients.
 * @param coefs The coefficients pointer.
 * @param n_coefs The number of coefficients to be saved.
 * @param n_rows The number of coefficients sets to be saved.
 */
void save_coefs(char *file_name, WORD *coefs, int n_coefs, int n_rows);

char *ftoa(float val, char *buffer);

/**
 * @brief Generate a Hamming window.
 *
 * @param image A pointer where the window will be created. It must be previously allocated.
 * @param img_area The image area.
 */
void hamming(double *image, int img_area);

/**
 * @brief Generate a Hanning window.
 *
 * @param image A pointer where the window will be created. It must be previously allocated.
 * @param img_area The image area.
 */
void hanning(double *image, int img_area);

/**
 * @brief Generates a fractional Hamming window for Knox-Thompson solar speckle imaging.
 *
 * This method is based on the IDL function frac_hamming(nx,ny,frac,Eqa).
 *
 * @param window A previously allocated double array that will contain the window data.
 * @param frac The Hamming parameter.
 * @param nx The x dimension of the window.
 * @param ny The y dimension of the window.
 */
void frac_hamming(double *window, float frac, int nx, int ny);

/**
 * @brief Apply a window over an image.
 *
 * @param img The Image data to have the window applied.
 * @param window The window data.
 * @param img_area The image area.
 */
void apply_window(double *img, double *window, int img_area);

/**
 * @brief Test all GPUs available by sending a buffer of 10K WORDs to every GPU
 * and then making a reduction of the values sent.
 *
 * If the resulting reduction is the expected WORD value 1.0, the test is considered as Passed.
 * Otherwise, the test failed.
 */
void test_gpu();

/**
 * @brief List all devices available.
 */
void list_gpu();

/**
 * @brief Calculate the Root Mean Square contrast from an image.
 *
 * @param img the image to calculate the RMS contrast.
 * @param width the iamge width.
 * @return the value representing the RMS contrast.
 */
double rmsContrast(double* img, int width);

#endif /* _KIPPE_H_ */
