
Copyright (c) 2013, Torsten Waldmann, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * PSF.cpp
 *
 *  Created on: 27/05/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "psf.hpp"

using namespace std;

/**
 * @brief PSF Constructor.
 *
 * @param context the OpenCL context instance to be used for this PSF.
 * @param _command_queue the OpenCL command queue instance to submit the commands.
 * @param img_size The image width.
 * @param phase_size The phase width.
 * @param _n_zernikes The number of Zernike terms that is being used for the calculations.
 */
PSF::PSF(cl_context context, cl_command_queue _command_queue, int img_size, int phase_size, int _n_zernikes){
	command_queue = _command_queue;
	img_area = img_size * img_size;
	size_fft = phase_size * phase_size;
	n_zernikes = _n_zernikes;

	CREATE_BUFFER(context, CL_MEM_READ_WRITE, size_fft * SIZEOF_WORD, NULL, cl_psf);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, size_fft * SIZEOF_FFTTYPE, NULL, cl_pupil);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, size_fft * SIZEOF_FFTTYPE, NULL, cl_focus);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_WORD, NULL, cl_psfe);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, REDUCTION_NBLOCKS * SIZEOF_WORD, NULL, cl_sum);
	CREATE_BUFFER(context, CL_MEM_READ_WRITE, img_area * SIZEOF_FFTTYPE, NULL, cl_fft_psfe);

	vn_pupil = new viennacl::vector<WORD>(cl_pupil, size_fft*2);
	vn_focus = new viennacl::vector<WORD>(cl_focus, size_fft*2);
	vn_fft_psfe = new viennacl::vector<WORD>(cl_fft_psfe, img_area*2);

	coefs = new WORD[n_zernikes];
	has_coefs = false;
	has_pupil = false;
}

/**
 * @brief The PSF destructor.
 */
PSF::~PSF() {
	delete vn_pupil;
	delete vn_focus;
	delete vn_fft_psfe;

	clReleaseMemObject(cl_pupil);
	clReleaseMemObject(cl_focus);
	clReleaseMemObject(cl_psf);
	clReleaseMemObject(cl_sum);
	clReleaseMemObject(cl_psfe);
	clReleaseMemObject(cl_fft_psfe);
	if(has_coefs) clReleaseMemObject(cl_coefs);
	if(has_pupil) clReleaseMemObject(cl_external_pupil);

	delete coefs;
}

/**
 * @brief Mark the position where the coefficients will be copied from an external matrix.
 *
 * Call the {@link getCoefs()} method to have the coefficients refreshed.
 * These coefficients are not used internally for this class, thus it's only informational data.
 *
 * @param _cl_coefs a reference to the OpenCl instance containing the source data.
 * @param offset the array position from _cl_coefs where the first value to be copied is located.
 */
void PSF::setCoefsPosition(cl_mem _cl_coefs, int offset){
	if(has_coefs) clReleaseMemObject(cl_coefs);
	clRetainMemObject(_cl_coefs);
	cl_coefs = _cl_coefs;
	coefs_offset = offset;
	has_coefs = true;
}

/**
 * @brief Mark the position where the pupil will be copied from an external matrix.
 *
 * Call the {@link refreshPupil} method to have the pupil refreshed.
 * The pupil is used internally for this class, thus it's important
 * to refresh the pupil before using it.
 *
 * @param _cl_pupil a reference for the openCl object containing the source data.
 * @param offset the array position from _cl_pupil where the first value to be copied is located.
 */
void PSF::setPupilPosition(cl_mem _cl_pupil, int offset){
	if(has_pupil) clReleaseMemObject(cl_external_pupil);
	clRetainMemObject(_cl_pupil);
	cl_external_pupil = _cl_pupil;
	pupil_offset = offset;
	has_pupil = true;
}

/**
 * @brief Refresh the internal pupil information with the external pupil matrix.
 *
 * @param block If this is a blocking copy.
 */
void PSF::refreshPupil(bool block){
	clMemcpyDeviceToDeviceOffset(command_queue, cl_pupil, 0, cl_external_pupil, pupil_offset * SIZEOF_FFTTYPE, size_fft * SIZEOF_FFTTYPE, block);
}

/**
 * @brief Calculate the FFT of the pupil.
 */
void PSF::pupilToFocusFft(){
	viennacl::fft(*vn_pupil, *vn_focus);
}

/**
 * @brief Calculate the FFT of the extracted (final) PSF.
 */
void PSF::psfeFft(){
	viennacl::inplace_fft(*vn_fft_psfe);
}

/**
 * @brief Refresh the internal coefficients with the external coefficient matrix.
 *
 * @return The coefficients related to this PSF.
 */
WORD* PSF::getCoefs(){
	clMemcpyDeviceToHostOffset(command_queue, coefs, cl_coefs, coefs_offset * SIZEOF_WORD, n_zernikes * SIZEOF_WORD);
	return coefs;
}
