
Copyright (c) 2013, Torsten Waldmann, Peter Frank Perroni

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
/*
 * psf_estimator.cpp
 *
 *  Created on: 16/07/2012
 *  Author: Peter Frank Perroni (pfperroni@inf.ufpr.br)
 */

#include "psf_estimator.hpp"

CPSO **PsfEstimator::cpso_pool;
bool *PsfEstimator::cpso_in_use;
int PsfEstimator::cpso_pool_size = 0;
omp_lock_t PsfEstimator::mutex;
double *PsfEstimator::zernikes;
int *PsfEstimator::phase_mask;
int *PsfEstimator::diffraction_mask;
int PsfEstimator::phase_size, PsfEstimator::image_size, PsfEstimator::n_zernikes, PsfEstimator::psf_range, PsfEstimator::n_particles, PsfEstimator::n_swarms;
WORD PsfEstimator::w, PsfEstimator::c1, PsfEstimator::c2, PsfEstimator::reset_at;

/**
 * @brief Startup the PSF estimation by making all the necessary initializations.
 *
 * This method must always be called only once before requesting any search instance.
 * The OpenCl environment and the search methods are started up by here.
 *
 * @param _cpso_pool_size The pool size for the CPSO search method. This is lazy load.
 * @param _zernikes The complete Zernike matrix.
 * @param _phase_mask The phase mask.
 * @param _diffraction_mask The diffraction mask.
 * @param _phase_size The phase width.
 * @param _image_size The image width.
 * @param _n_zernikes The number of Zernike terms used for the calculations.
 * @param _psf_range The range for the Zernike coefficients.
 * @param _w The parameter w for the CPSO search.
 * @param _c1 The parameter c1 for the CPSO search.
 * @param _c2 The parameter c2 for the CPSO search.
 * @param _reset_at The particle reset parameter for the CPSO search.
 * @param _n_particles The number of particles for the CPSO search.
 * @param _n_swarms The number of swarms for the CPSO search.
 * @param _devices The list of devices to be used. If Null, all available devices will be used.
 */
void PsfEstimator::startup(int _cpso_pool_size, double* _zernikes, int *_phase_mask, int *_diffraction_mask,
		int _phase_size, int _image_size, int _n_zernikes, int _psf_range, WORD _w, WORD _c1, WORD _c2,
		WORD _reset_at, int _n_particles, int _n_swarms, vector<int> *_devices){
	// Allows one 1 startup.
	if(cpso_pool_size > 0 || _cpso_pool_size < 1){
		return;
	}

	omp_init_lock(&mutex);
	clFactory::startup(_devices);

#ifdef _PROFILING_
	// Reset the profiling counters and initialize the locks.
	Profiling::startup();
#endif

	zernikes = _zernikes;
	phase_mask = _phase_mask;
	diffraction_mask = _diffraction_mask;
	phase_size = _phase_size;
	image_size = _image_size;
	n_zernikes = _n_zernikes;
	psf_range = _psf_range;
	w = _w;
	c1 = _c1;
	c2 = _c2;
	reset_at = _reset_at;
	n_particles = _n_particles;
	n_swarms = _n_swarms;

	cpso_pool_size = _cpso_pool_size;
	cpso_in_use = new bool[cpso_pool_size];
	cpso_pool = new CPSO*[cpso_pool_size];
	for(int i=0; i < cpso_pool_size; i++){
		// Lazy load pool.
		cpso_pool[i] = NULL;
		cpso_in_use[i] = false;
	}
}

/**
 * @brief Shutdown the OpenCl and the search methods.
 *
 * This method release all the remaining OpenCl data and
 * frees up the data allocated by the search methods.
 */
void PsfEstimator::shutdown(){
	for(int i=0; i < cpso_pool_size; i++){
		if(cpso_pool[i] != NULL){
			cpso_pool[i]->finalize_cl();
			delete cpso_pool[i];
		}
	}
	delete[] cpso_pool;
	delete cpso_in_use;

	omp_destroy_lock(&mutex);

#ifdef _PROFILING_
	Profiling::shutdown();
#endif

	CPSO::clear_static_data();
	clFactory::shutdown();
}

/**
 * @brief Get a CPSO instance from the pool.
 *
 * It returns the first instance available from the pool.
 *
 * This method initializes the pool of CPSO as needed (lazy load).
 * If there's no instance available on the pool, it creates a new instance.
 * If the pool is completely in use, wait for one instance to be released.
 *
 * Once the instance is obtained, the instance data is reset
 * and the new Object and Image are transfered to the device.
 *
 * @param object The object data.
 * @param image The image data.
 * @param startp_coefs The Zernike coefficients for the startup (not required).
 * 			If provided, it will be used for initializing the search.
 * 			If not provided, random coefficients will be created.
 * @return The CPSO instance obtained.
 */
CPSO* PsfEstimator::getCPSO_internal(double *object, double *image, WORD *startp_coefs){
	CPSO *cpso = NULL;
	int i;
	while(cpso == NULL){
		omp_set_lock(&mutex);
		for(i=0; i < cpso_pool_size; i++){
			// If there's no instance into this position of the pool, creates one.
			if(cpso_pool[i] == NULL){
				cpso_pool[i] = new CPSO(zernikes, phase_mask, diffraction_mask, phase_size, image_size, n_zernikes,
						psf_range, w, c1, c2, reset_at, n_particles, n_swarms);
			}
			// Search for one instance in the pool that is not in use.
			if(!cpso_pool[i]->isInUse()){
				cpso = cpso_pool[i];
				// Lock the CPSO instance to inform that it's being used.
				cpso->lock();
				omp_unset_lock(&mutex);
				// If provided, set the coefficients for the startup of the search.
				if(startp_coefs != NULL){
					cpso->setStartupCoefs(startp_coefs);
				}
				// Set the new Object and Image for the search.
				cpso->set_images(object, image);
				return cpso;
			}
		}
		omp_unset_lock(&mutex);
		// If all instances are in use, wait a moment before checking again.
		usleep(10000);
	}
	return cpso;
}

/**
 * @brief Get a CPSO instance from the pool.
 *
 * @param object The object data.
 * @param image The image data.
 * @param startp_coefs The Zernike coefficients for the startup (not required).
 * 			If provided, it will be used for initializing the search.
 * 			If not provided, random coefficients will be created.
 * @return The CPSO instance obtained.
 */
CPSO* PsfEstimator::getCPSO(double *object, double *image, WORD *startp_coefs){
	return getCPSO_internal(object, image, startp_coefs);
}

/**
 * @brief Get a CPSO instance from the pool.
 *
 * @param object The object data.
 * @param image The image data.
 * @return The CPSO instance obtained.
 */
CPSO* PsfEstimator::getCPSO(double *object, double *image){
	return getCPSO_internal(object, image, NULL);
}

/**
 * @brief Reset the pool of CPSO Objects.
 *
 * The reset consists of finalizing all instances already created.
 * If any instance is in use, wait for its release.
 *
 * @param _w The new parameter w for the CPSO search.
 * @param _c1 The new parameter c1 for the CPSO search.
 * @param _c2 The new parameter c2 for the CPSO search.
 * @param _reset_at The new particle reset parameter for the CPSO search.
 * @param _n_particles The new number of particles for the CPSO search.
 * @param _n_swarms The new number of swarms for the CPSO search.
 */
void PsfEstimator::resetPool(WORD _w, WORD _c1, WORD _c2, WORD _reset_at, int _n_particles, int _n_swarms){
	omp_set_lock(&mutex);
	w = _w;
	c1 = _c1;
	c2 = _c2;
	reset_at = _reset_at;
	n_particles = _n_particles;
	n_swarms = _n_swarms;

	for(int i=0; i < cpso_pool_size; i++){
		if(cpso_pool[i] != NULL){
			while(cpso_pool[i]->isInUse()) usleep(10000);
			cpso_pool[i]->finalize_cl();
			delete cpso_pool[i];
			cpso_pool[i] = NULL;
			cpso_in_use[i] = false;
		}
	}
	omp_unset_lock(&mutex);
}
